/*********************************************************************
 *  File : flgrf.c
 *
 *  Construct the table of fermion lines.
 *
 *  Only spin=1/2 is checked.
 *
 *                                                 by Kaneko Toshiaki
 *
 * 1996/08/07
 *     An elmemnt "fclass" is added to the struct tfline.
 *     This is a bit string of the type of fermion (M_DIRAC,
 *     M_GHOST or M_MAJORANA) included in the f-line.
 */

#include <stdlib.h>
#include <stdio.h>
#include "grdata.h"

#define MAJORANA

/*--------------------------------------------------------------------
 * local functions
 */
Private Void    flinit   ARG((NoArg));
Private int     flfind   ARG((int*));
Private int     flstart  ARG((int*));
Private int     flstart0 ARG((int*));
Private Void    fldmpt   ARG((int,int));
Private int     flnextl  ARG((node, int, node));
Private bool    flinl    ARG((node, int));
Private int     floutl   ARG((node));
Private bool    flundef  ARG((NoArg));
Private Void    fldmpw   ARG((NoArg));
Private Void    flchk1   ARG((int));

#define flmaj(p)    (ptable[p]->ptype == P_MAJORANA)
#define flin(p)     ((p) < 0 && (pvdfermi(-(p))))
#define flout(p)    ((p) > 0 && (pvdfermi(p)))
#define flinout(p)  flmaj(abs(p))

/*--------------------------------------------------------------------
 * local variables
 */
Local bool floccupied = False;
Local bool majorana;
Local node flnodes[MAXNODES]; /* temporary list of nodes */
Local int  fllegs[2][MAXNODES]; /* temporary list of legs */

/* the number of incoming fermion legs
 * the number of legs if majorana particle appears in the graph
 */
Local node flflegs[3][MAXNODES];

#define fltnam(typ)  (((typ)==FLOPEN)?"open  ": \
                      ((typ)==FLCLOSED)?"closed":"undef ")
/*==================================================================*/
Public  Void    flget(grf)
tgraph *grf;
/*--------------------------------------------------------------------
 *    Construct the table of f-lines.
 */
{
    int f, nn, n, ntyp, i;

    cgraph = grf;

    if(flundef()) {
        return;
    }

    /* free used area */
    if(floccupied) {
        for(f = nflines-1; f >= 0; f--) {
            if(fltable[f].length > 0) {
                if(fltable[f].ndlist != NULL) {
                    free((AnyP)fltable[f].ndlist);
                    for(i = 0; i < 2; i++) {
                        free((AnyP)fltable[f].lglist[i]);
                    }
                }
            }
        }
    } else {
        floccupied = True;
    }

    flinit();

    nflines = 0;
    while((nn = flfind(&ntyp)) > 0) {
#if DBG
        if(dbg >= fldbg) {
            printf("flget: flfind = %d, ntype= %s\n",nn,fltnam(ntyp));
        }
#endif
/*        printf("flget: flfind = %d, ntype= %s\n",nn,fltnam(ntyp));*/

        fltable[nflines].ndlist = (node *) malloc(nn*sizeof(node));
        if(fltable[nflines].ndlist == NULL) {
            error("flget","Memory allocation failed.");
            erfail();
        }
        for(i = 0; i < 2; i++) {
            fltable[nflines].lglist[i] = (int *) malloc(nn*sizeof(node));
            if(fltable[nflines].lglist[i] == NULL) {
                error("flget","Memory allocation failed.");
                erfail();
            }
        }
        fltable[nflines].type   = ntyp;
        fltable[nflines].length = nn;
        fltable[nflines].fclass = 0;
        for(n = 0; n < nn; n++) {
            fltable[nflines].ndlist[n] = flnodes[n];
            for(i = 0; i < 2; i++) {
                fltable[nflines].lglist[i][n] = fllegs[i][n];
            }
            if(fllegs[FL_OUT][n] >= 0) {
                fltable[nflines].fclass |=
                    pmask[ptable[abs(
                          ndparticle(flnodes[n], fllegs[FL_OUT][n]))]->ptype];
            }
        }

        /* determine direction of i-lines of majorana */

        nflines++;
    }

/*    printf("flget:nflines =%d\n",nflines);*/

#if DBG
    if(dbg >= fldbg) {
        fldump();
    }
#endif
}

/*==================================================================*/
Private bool    flundef(NoArg)
/*--------------------------------------------------------------------
 *    Check whether if Majorana particle appears in the graph or not.
 */
{
    node    n;
    int     l, nlg;

    for(n = 0; n < ndnodes(); n++) {
        nlg = ndlegs(n);
        for(l = 0; l < nlg; l++) {
            if(ndparticle(n, l) == 0) {
/*                printf("flundef:True\n");*/
                return True;
            }
        }
    }
/*    printf("flundef:False\n");*/
    return False;
}

/*==================================================================*/
Public  bool    flmajo(NoArg)
/*--------------------------------------------------------------------
 *    Check whether if Majorana particles are defined in the model
 */
{
/*  We assume that fermion number viorating vertices appears
 *  only in such models that Majorana particles are defined.
 */
    return majop;
}

/*==================================================================*/
Public  bool    flmajog(NoArg)
/*--------------------------------------------------------------------
 *    Check whether if Majorana particle appears in the graph or not.
 */
{
    node    n;
    int     l, nlg;
    pindex  p;

    for(n = 0; n < ndnodes(); n++) {
        nlg = ndlegs(n);
        for(l = 0; l < nlg; l++) {
            p = abs(ndparticle(n, l));
            if(flmaj(p)) {
                return True;
            }
        }
    }
    return False;
}

/*==================================================================*/
Private Void    flinit(NoArg)
/*--------------------------------------------------------------------
 *    Initialization
 */
{
    node    n;
    int     l, nlg;
    pindex  p;

    majorana = flmajo();

    for(n = 0; n < ndnodes(); n++) {
        for(l = 0; l < 3; l++) {
            flflegs[l][n] = 0;
        }
        nlg = ndlegs(n);
        for(l = 0; l < nlg; l++) {
            p = ndparticle(n,l);
            if(flin(p)) {
                flflegs[FL_IN][n]++;
                ndilwk(n,l) = FL_IN;
            } else if(flout(p)) {
                flflegs[FL_OUT][n]++;
                ndilwk(n,l) = FL_OUT;
            } else if(flinout(p)) {
                flflegs[FL_BOTH][n]++;
                ndilwk(n,l) = FL_BOTH;
            } else {
                ndilwk(n,l) = FL_UNDEF;
            }
        }
    }
}

/*==================================================================*/
Private int     flstart0(ntyp)
int    *ntyp;
/*--------------------------------------------------------------------
 *    Finds a starting node of new f-line : older version
 */
{
    int start, n, l, p, pr, m=0;

    start = -1;
    pr    = 0;
    for(n = 0; n < ndnodes(); n++) {
        if(majorana) {
            if(ndtype(n) == NDEXTERN) {
                if(ndilwk(n, 0) != FL_UNDEF) {
                    start = n;
                    *ntyp = FLOPEN;
                    p  = ndparticle(n,0);
                    pr = 3;
                    printf("flstart0:n = %d, p = %d\n", n, p);
                    if(p > 0 && pvfermi(p)) {
                        m = 1;
                        break;
                    }
                }
            } else if(pr < 3) {
                for(l = 0; l < ndlegs(n); l++) {
                    if(ndilwk(n, l) != FL_UNDEF) {
                        break;
                    }
                }
                if(l < ndlegs(n)) {
                    if(ndtype(n) == NDONSHELL && pr < 2) {
                        pr    = 2;
                        start = n;
                        *ntyp = FLCLOSED;
                        m = 2;
                    } else if(pr < 1) {
                        pr    = 1;
                        start = n;
                        *ntyp = FLCLOSED;
                        m = 3;
                    }
                }
            }
        } else {
            if(ndtype(n) == NDEXTERN) {
                if(ndilwk(n, 0) == FL_OUT) {
                    flflegs[FL_OUT][n] = 0;
                    start = n;
                    *ntyp = FLOPEN;
                        m = 4;
                    break;
                }
            } else if(flflegs[FL_OUT][n] > 0) {
                start = n;
                *ntyp = FLCLOSED;
                if(ndtype(n) == NDONSHELL) {
                    m = 5;
                    break;
                }
            }
        }
    }

#if DBG
    if(dbg >= fldbg) {
        printf("flstart0:1: start=%d, type=%d, m=%d\n", start, *ntyp, m);
    }
#endif

    return start;
}

/*==================================================================*/
Private int     flstart(ntyp)
int    *ntyp;
/*--------------------------------------------------------------------
 *    Finds a starting node of new f-line
 */
{
    node    n, start;
    int     nnodes, l, m=0;

    *ntyp  = FLUNDEF;
    start  = -1;
    nnodes = ndnodes();

    /* find the starting point of open g-line
     *     1. external outgoing Dirac
     */
    for(n = 0; n < nnodes; n++) {
        if(ndtype(n) == NDEXTERN) {
            if(ndilwk(n, 0) != FL_UNDEF) {
                if(flflegs[FL_OUT][n] > 0) {
                    start = n;
                    *ntyp = FLOPEN;
                    m = 1;
                    break;
                }
            }
        }
    }

    /* find the starting point of open g-line 
     *     2. external Majorana
     */
    if(start < 0 && majorana) {
        for(n = 0; n < nnodes; n++) {
            if(ndtype(n) == NDEXTERN) {
                if(ndilwk(n, 0) != FL_UNDEF) {
                    if(flflegs[FL_BOTH][n] > 0) {
                        start = n;
                        *ntyp = FLOPEN;
                        m = 2;
                        /* break; */
                    }
                }
            }
        }
    }

    /* find the starting point of open g-line 
     *     3. external incoming Dirac
     */
    if(start < 0 && majorana) {
        for(n = 0; n < nnodes; n++) {
            if(ndtype(n) == NDEXTERN) {
                if(ndilwk(n, 0) != FL_UNDEF) {
                    if(flflegs[FL_IN][n] > 0) {
                        start = n;
                        *ntyp = FLOPEN;
                        m = 3;
                    }
                }
            }
        }
    }

    /* find the starting point of closed g-line
     *     4. vertex with outgoing Dirac
     */
    if(start < 0) {
        for(n = 0; n < nnodes; n++) {
            if(ndtype(n) == NDEXTERN) {
                continue;
            }
            for(l = 0; l < ndlegs(n); l++) {
                if(ndilwk(n, l) != FL_UNDEF) {
                    break;
                }
            }
            if(l < ndlegs(n)) {
                if(flflegs[FL_OUT][n] > 0) {
                    start = n;
                    *ntyp = FLCLOSED;
                    m = 4;
                    break;
                }
            }
        }
    }

    /* find the starting point of closed g-line
     *     5. vertex with Majorana
     */
    if(start < 0 && majorana) {
        for(n = 0; n < nnodes; n++) {
            if(ndtype(n) == NDEXTERN) {
                continue;
            }
            for(l = 0; l < ndlegs(n); l++) {
                if(ndilwk(n, l) != FL_UNDEF) {
                    break;
                }
            }
            if(l < ndlegs(n)) {
                if(flflegs[FL_BOTH][n] > 0) {
                    start = n;
                    *ntyp = FLCLOSED;
                    m = 5;
                    break;
                }
            }
        }
    }

    /* find the starting point of closed g-line
     *     6. vertex with incoming Dirac
     */
    if(start < 0 && majorana) {
        for(n = 0; n < nnodes; n++) {
            if(ndtype(n) == NDEXTERN) {
                continue;
            }
            for(l = 0; l < ndlegs(n); l++) {
                if(ndilwk(n, l) != FL_UNDEF) {
                    break;
                }
            }
            if(l < ndlegs(n)) {
                if(flflegs[FL_IN][n] > 0) {
                    start = n;
                    m = 6;
                    *ntyp = FLCLOSED;
                }
            }
        }
    }

#if DBG
    if(dbg >= fldbg) {
        printf("flstart:1: start=%d, type=%d, m=%d\n", start, *ntyp, m);
    }
#endif

    return start;
}

/*==================================================================*/
Private int     flfind(ntyp)
int    *ntyp;
/*--------------------------------------------------------------------
 *    Finds a fermion line and returns the length.
 *
 *    For loop, the starting vertex is one which has the maximum
 *    index of nodes not counted.
 */
{
    node    n, nv, no, start;
    int     nl, l, lv, inout;
#if DBG
    node    nn;
    int     ll;
#endif

#if DBG
    if(dbg >= fldbg) {
        printf("flfind: majorana = %d\n", majorana);
        fldmpw();
    }
#endif

    if(False) {
        if((start = flstart0(ntyp)) < 0) {   /* compatible to oder version */
            return 0;
        }
    } else {
        if((start = flstart(ntyp)) < 0) {
            return 0;
        }
    }
    nv = start;
    nl = 0;
    lv = -1;

    /* search nodes */
    for(n = 0; n <= ndnodes(); n++) { /* maximum number of repeat */

#if DBG
        if(dbg >= fldbg) {
            printf("flfind:4: %d: nv = %d\n", nl, nv);
            printf("node = %d, leg = %d\n", nv, lv);
            for(nn = 0; nn < ndnodes(); nn++) {
                printf("%4d : ", nn);
                for(ll = 0; ll < ndlegs(nn); ll++) {
                    printf(" %d:%d", ll, ndilwk(nn, ll));
                }
                printf("\n");
            }
        }
#endif

        /* end of closed g-line */
        if(nl > 0 && start == nv) {
            if(*ntyp != FLCLOSED) {
                error("flfind", "end of loop: inconsistent type");
                fldmpw();
                fprintf(Stderr, "node = %d, leg = %d\n", nv, lv);
                erfail();
            }
            inout = ndilwk(nv, lv);
            ndilwk(nv, lv) = FL_UNDEF;
            --flflegs[inout][nv];
            fllegs[FL_IN][0] = lv;

#if DBG
            if(dbg >= fldbg) {
                printf("flfind:5: %d: nv = %d\n", nl, nv);
                printf("node = %d, leg = %d\n", nv, lv);
                for(nn = 0; nn < ndnodes(); nn++) {
                    printf("%4d : ", nn);
                    for(ll = 0; ll < ndlegs(nn); ll++) {
                        printf(" %d:%d", ll, ndilwk(nn, ll));
                    }
                    printf("\n");
                }
            }
#endif
            return nl;
        }


        /* enter to the node "nv", from the leg "lv" */
        if(lv >= 0) {
            inout = ndilwk(nv, lv);
            if(flflegs[inout][nv] < 1) {
                error("flfind", "on the path: cannot enter to the node");
                fprintf(Stderr, "node = %d, leg = %d\n", nv, lv);
                fldmpw();
                erfail();
            }
            ndilwk(nv, lv) = FL_UNDEF;
            --flflegs[inout][nv];
        }

        /* save the node 'nv' */
        flnodes[nl] = nv;
        fllegs[FL_IN][nl] = lv;

        /* search ndlegs(nv)-th outgoing fermion leg */
        l = floutl(nv);
        fllegs[FL_OUT][nl] = l;
        nl++;

        /* end of open g-line */
        if(l < 0) {
            if(ndtype(nv) == NDEXTERN && *ntyp == FLOPEN) {
                if(*ntyp != FLOPEN) {
                    error("flfind", "end of open-l: inconsistent type");
                    fldmpw();
                    fprintf(Stderr, "node = %d, leg = %d\n", nv, lv);
                    erfail();
                }
                return nl;

            /* illegal case */
            } else {
                error("flfind", "inconsistent g-line.");
                fprintf(Stderr, "vertex = %d\n", nv);
                fldmpt(nl,*ntyp);
                fldmpw();
                fprintf(Stderr, "vertex = %d\n", nv);
                erfail();
            }
        }
        inout = ndilwk(nv, l);
        ndilwk(nv, l) = FL_UNDEF;
        --flflegs[inout][nv];

        no = nv;
        nv = ndnextn(no, l);
        lv = flnextl(no, l, nv);
        if(lv < 0 || !flinl(nv, lv)) {
            error("flfind", "inconsistent g-line.");
            fprintf(Stderr, "%d: (%d, %d) --> (%d, %d)\n", nl, no, l, nv, lv);
            fldmpt(nl,*ntyp);
            fldmpw();
            fprintf(Stderr, "%d: (%d, %d) --> (%d, %d)\n", nl, no, l, nv, lv);
            erfail();
        }

#if DBG
        if(dbg >= fldbg) {
            printf("flfind:8: %d: (%d, %d) --> (%d, %d)\n", nl, no, l, nv, lv);
        }
#endif
    }
    error("flfind", "inconsistent search");
    fldmpt(nl,*ntyp);
    fldmpw();
    erfail();
    return -1;
}

/*==================================================================*/
Private  int    floutl(nv)
node nv;
/*--------------------------------------------------------------------
 *   Find a leg of 'nv' from which g-line can go out.
 */
{
    int ni, l;

    ni = 0;
    for(l = 0; l < ndlegs(nv); l++) {
        if(ndilwk(nv, l) == FL_OUT) {
            return l;
        } else if(ndilwk(nv, l) == FL_BOTH) {
            return l;
        }
    }
    if(majorana) {
        for(l = 0; l < ndlegs(nv); l++) {
            if(ndilwk(nv, l) != FL_UNDEF) {
                return l;
            }
        }
    }
    return -1;
}

/*==================================================================*/
Private  int    flnextl(nv, lg, nn)
node nv;
int  lg;
node nn;
/*--------------------------------------------------------------------
 */
{
    int il, l;

    il = ndiline(nv, lg);
    for(l = 0; l < ndlegs(nn); l++) {
        if(il == ndiline(nn, l)) {
            if(nn == nv && l == lg) {
                continue;
            }
            return l;
        }
    }
    return -1;
}

/*==================================================================*/
Private  bool   flinl(nv, lg)
node nv;
int  lg;
/*--------------------------------------------------------------------
 *   Find a leg of 'nv' from which g-line can come in.
 */
{
    if(ndilwk(nv, lg) == FL_IN) {
        return True;
    } else if(ndilwk(nv, lg) == FL_BOTH) {
        return True;
    }
    if(majorana) {
        if(ndilwk(nv, lg) != FL_UNDEF) {
            return True;
        }
    }
    return False;
}

/*==================================================================*/
Public  bool    flfurry(NoArg)
/*--------------------------------------------------------------------
 *    test Furry's theorem.
 *    This function should be called after "flget".
 */
{
    int     f, n, l, nf=0;
    node    nd;
    pindex  p;
    int     vi;

/*
    printf(" test Furry's theorem \n"); return;
*/
    /* looped fermion lines with odd number of lengths */
    for(f = 0; f < nflines; f++) {
        if(fltable[f].type == FLOPEN) {
            continue;
        }
        if(fltable[f].length % 2 == 0) {
            continue;
        }
        vi = True;
        for(n = 0; n < fltable[f].length; n++) {
            nd = fltable[f].ndlist[n];
            for(l = 0; l < ndlegs(nd); l++) {
                p = abs(ndparticle(nd, l));
                if(ptable[p]->ptype == P_DIRAC) {
                    continue;
                }
                if(!ptable[p]->furry) {
                    vi = False;
                    break;
                }
            }
            if(vi) {
#if DBG
                if(dbg >= fldbg) {
                    printf("delete by Furry : f=%d\n", f);
                }
#endif
                return False;
            }
        }
    }

    /* looped vertex */
    for(n = 0; n < ndnodes(); n++) {
        if(ndtype(n) != NDLOOPEDV) {
            continue;
        }
        if(ndlegs(n) % 2 == 0) {
            continue;
        }

        vi = True;
        nf = 0;
        for(l = 0; l < ndlegs(n); l++) {
            p = abs(ndparticle(n, l));
            if(ptable[p]->furry) {
                nf++;
            } else {
                vi = False;
                break;
            }
        }

        if(vi && nf%2 != 0) {
#if DBG
            if(dbg >= fldbg) {
                printf("delete by Furry : v=%d\n", n);
            }
#endif
            return False;
        }
    }
    return True;
}

/*==================================================================*/
Public  Void    fldump(NoArg)
/*--------------------------------------------------------------------
 *   Dump the table of f-lines.
 */
{
    int f, n;

    printf("\n%% +++ %d fermion lines. (graph %ld)\n",
           nflines, cgraph->grid);
    for(f = 0; f < nflines; f++) {
        printf("%% type=%s, length=%4d, class=%4d : ",
               fltnam(fltable[f].type), fltable[f].length,
               fltable[f].fclass);
        for(n = 0; n < fltable[f].length; n++) {
            printf("[%d,v:%d,%d] ",
                   fltable[f].lglist[FL_IN][n],
                   tovert(fltable[f].ndlist[n]),
                   fltable[f].lglist[FL_OUT][n]);
        }
        putchar(EOL);
    }
}

/*==================================================================*/
Public  int     flfact(NoArg)
/*--------------------------------------------------------------------
 *   Calculate sign of the graph for exchange of fermions,
 *   and (-1) for a fermion loop.
 */
{
    int  f, n, i1, i2, sgn, t;
    bool cont;
    int  *fsort;

/*  GENKYO 980701 */
/*  sgn = -1;*/
    sgn =  1;

    if(nflines < 1) {
        return sgn;
    }

    if((fsort = (int*) malloc(2*nflines*sizeof(int))) == NULL) {
        error("flfact", "memory allocation failed");
        erfail();
    }

    n = 0;
    for(f = 0; f < nflines; f++) {
        if(fltable[f].type == FLOPEN) {
            fsort[n++] = fltable[f].ndlist[0];
            fsort[n++] = fltable[f].ndlist[fltable[f].length-1];
        } else {  /* loop */
            sgn = - sgn;
        }
    }

    /* sort and count sign */
    for(i1 = 0; i1 < n-1; i1++) {
        cont = True;
        while (cont) {
            cont = False;
            for(i2 = i1+1; i2 < n; i2++) {
                if(fsort[i1] > fsort[i2]) {
                    t         = fsort[i1];
                    fsort[i1] = fsort[i2];
                    fsort[i2] = t;
                    sgn *= -1;
                    cont = True;
                    break;
                }
            }
        }
    }

    free(fsort);

    return sgn;
}

#ifdef MAJORANA
/*==================================================================*/
Public  int     flmfact(NoArg)
/*--------------------------------------------------------------------
 *   Calculate sign of the graph for exchange of fermions,
 *   and (-1) for a fermion loop.
 */
{
    int  f, n, i1, i2, sgn, t;
    bool cont;
    int  *fsort;
/*  int  np, fn, nd, mj = 0;*/
    int  np, fn, nd;

/* printf("flmfact:begin\n");*/
    sgn = 1;

    if(nflines < 1) {
        return sgn;
    }

    if((fsort = (int*) malloc(2*nflines*sizeof(int))) == NULL) {
        error("flfact", "memory allocation failed");
        erfail();
    }

    n = 0;
    for(f = 0; f < nflines; f++) {
        if(fltable[f].type == FLOPEN) {
            fsort[n++] = fltable[f].ndlist[0];
            fsort[n++] = fltable[f].ndlist[fltable[f].length-1];
        } else {  /* loop */
            np = 0;
            for(fn = 0; fn < fltable[f].length; fn++) {
                nd = fltable[f].ndlist[fn];
                if(ndtype(nd) == NDONSHELL) {
                    np++;
                }
            }
            if(np == 0) {    /* loop */
                sgn = - sgn;
            } else if(np % 2 != 0) {
                error("flfact", "illegal number of projection operators");
                fprintf(Stderr, "f = %d/%d\n", f, nflines);
                erfail();
            } else if((((int) np/2))%2 == 0) {
/*              mj = 1;*/
                sgn = - sgn;
            } else {
/*              mj = 1;*/
            }
/*printf("fl = %d/%d, np = %d, sgn = %d\n", f, nflines, np, sgn);*/
        }
    }

    /* sort and count sign */
    for(i1 = 0; i1 < n-1; i1++) {
        cont = True;
        while (cont) {
            cont = False;
            for(i2 = i1+1; i2 < n; i2++) {
                if(fsort[i1] > fsort[i2]) {
                    t         = fsort[i1];
                    fsort[i1] = fsort[i2];
                    fsort[i2] = t;
                    sgn *= -1;
                    cont = True;
                    break;
                }
            }
        }
    }

    free(fsort);

/* GENKYO */
/*  if(!mj) {*/ /* change global sign */
/*      sgn = - sgn;*/
/*  }*/
/*printf("flmfact:end\n");*/
    return sgn;
}
#endif

/*==================================================================*/
Private Void    fldmpt(nl, typ)
int     nl, typ;
/*--------------------------------------------------------------------
 *   Dump temporary f-line.
 */
{
    int n;

    printf("\n%% +++ current fermion lines. length=%d, type=%s\n",
           nl, fltnam(typ));
    printf("%% ");
    for(n = 0; n < nl; n++) {
        printf(" [%d, v:%d, %d] ",fllegs[FL_IN][n], flnodes[n],
                                  fllegs[FL_OUT][n]);
    }
    putchar(EOL);
}


/*==================================================================*/
Private Void    fldmpw(NoArg)
/*--------------------------------------------------------------------
 *   Dump temporary f-line.
 */
{
    int nn, ll;

    for(nn = 0; nn < ndnodes(); nn++) {
        printf("%4d : flflegs", nn);
        for(ll = 0; ll < 3; ll++) {
            printf(" %d:%2d", ll, flflegs[ll][nn]);
        }
        printf(" : ndilw");
        for(ll = 0; ll < ndlegs(nn); ll++) {
            printf(" %d:%2d", ll, ndilwk(nn, ll));
        }
        printf("\n");
    }
}

/*==================================================================*/
Public  Void    flcheck(NoArg)
/*--------------------------------------------------------------------
 *   Check the table of f-lines.
 */
{
    int f;

    for(f = 0; f < nflines; f++) {
        flchk1(f);
    }
}

/*==================================================================*/
Private Void    flchk1(f)
int     f;
/*------------------------------------------------------------------*
 *  Check a fermion line
 *------------------------------------------------------------------*/
{
    int  n, nn, il, len, lg=0, lgn, ndl, ndnl;
    int  fail = 0;
    node nd, ndn;
    
    nd  = fltable[f].ndlist[0];
    len = fltable[f].length; 
    
#ifdef DEBUG1
    printf("flcheck : (iltable=%p) : f=%d: V(-:%d:", iltable, f, nd);
#endif
    for(n = 0; n < fltable[f].length; n++) {
        /*
         *  (nd:ndl) -- (il) -- (ndn:ndnl)
         */
        ndl  = fltable[f].lglist[FL_OUT][n];

        if(ndl < 0) {
            break;
        }
        il   = ndiline(nd, ndl);
        nn   = (n+1) % len;                     /* next node */
        ndn  = fltable[f].ndlist[nn];
        ndnl = fltable[f].lglist[FL_IN][nn];

#ifdef DEBUG1
        printf("%d)=", ndl);
#endif
        if( il < 1 ) {
            continue;
        }     
        if(iltable[il].ilnodes[0] == nd && iltable[il].illegs[0] == ndl) {
            lg = 0;
        } else if(iltable[il].ilnodes[1] == nd &&
                  iltable[il].illegs[1] == ndl) {
            lg = 1;
        } else {
            fail = 1;
        }
        lgn = 1 - lg;
#ifdef DEBUG1
        printf("=I(%d:%d:%d)=>V(%d:%d:", lg, il, lgn, ndnl, ndn);
#endif
        if(fail == 0) {
            if(iltable[il].illegs[lg] != ndl) {
                fprintf(Stderr,"\n*** flchk1 iltbl[%d].illegs[%d] = %d != %d\n",
                        il, lg, iltable[il].illegs[lg], ndl);
                fail = 2;
            } else if(iltable[il].ilnodes[lgn] != ndn) {
                fprintf(Stderr,"\n*** flchk1 iltbl[%d].illegs[%d] = %d != %d\n",
                        il, lgn, iltable[il].illegs[lgn], ndn);
                fail = 3;
            } else if(iltable[il].illegs[lgn] != ndnl) {
                fprintf(Stderr,"\n*** flchk1 iltbl[%d].illegs[%d] = %d != %d\n",
                        il, lgn, iltable[il].illegs[lgn], ndnl);
                fail = 4;
            }
        }
        
        if(fail != 0) {
#ifdef DEBUG1
            printf("FAIL)\n");
#endif
            fprintf(Stderr, "*** flchk1(f=%d,n=%d) : illegal connection ",
                    f, n);
            fprintf(Stderr, "(nd=%d,il=%d,fail=%d)\n", nd, il, fail);
            prdump();
            ildump();
            exit(FAIL);
        }

        nd = ndn;
    } 
#ifdef DEBUG1
    printf("-)\n");
#endif
}
