/*********************************************************************
 *    File :  grdata.h
 *
 *    data of graph file.
 *
 *                                               by Kaneko Toshiaki
 */


/*-------------------------------------------------------------------
 *    Data types and variable of model definition.
 */
#include "common.h"
#include "dbg.h"

/*
#define SMALL
*/

/*-------------------------------------------------------------------
 *    Size of tables
 */

/* The maximum number of processes
 */
#ifdef SMALL
#define MAXPROCS   20
#else
#ifdef HUGE
#define MAXPROCS   10000
#else
#define MAXPROCS   100
#endif
#endif

#ifdef MAXPROCS
#undef MAXPROCS
#endif
#define MAXPROCS   10000

/* The maximum number of external lines
 */
#define MAXELINE  16

/* The maximum number of loops
 */
/* In case of "grcred -a", this is corresponding to MAXPROCS */
#define MAXLOOP   10

/* The maximum number of nodes (external particles or vertices)
 */
#define MAXNODES  (2*MAXLOOP + MAXELINE - 2)

/* The maximum total number of edges of nodes
 */
#define MAXLEGS  MAXELINE*3

/* The maximum number of fermion lines in a graph.
 */
#ifdef SMALL
#define MAXFLINES 10
#else
#define MAXFLINES 20
#endif

/* size of vertex table
 */
#ifdef SMALL
#define LVTABLE 5000
#else
#define LVTABLE 10000
#endif

/*  Table of particle definitions.
 */
#define LPTABLE 150              /* size of particle table */

#define MAXDIR  LVTABLE

/*===================================================================
 */
#include "ptable.h"

/*===================================================================
 *    Data types
 *
 *    tproc  : type of physical process
 *    tgraph : type of Feynman graphs
 *    tnode  : type of nodes (external particles and vertices)
 *    tleg   : type of legs of nodes
 *    tiline : type of external lines
 *    tfline : type of fermion lines
 */

/*-------------------------------------------------------------------
 *    tleg   : type of legs of nodes
 */
typedef struct {
    /* the next node connected to the node of this leg.
     */
    int     lgnext;

    /* the incoming particle to the node.
     */
    int     lgpart;

    /* the internal line number connected to this leg.
     */
    int     lgiline;

    /* work space */
    int     lgwork;
} tleg;

/*-------------------------------------------------------------------
 *    tnode   : type of nodes
 *-------------------------------------------------------------------
 * The ordering of legs is
 *
 * 1. step of graph generation
 *    no ordering condition
 *
 * 2. after reading 'out.grf'
 *    legs are match to the definition of vertices in the model file.
 *    When the node is on a fline, the legs are rearranged
 *    according to the direction of fline.
 *
 *       m  --->---x--->--- n
 *               fline            m < n
 */
typedef struct {
    /* type of the node : see below
     */
    int nvtype;

    /* iorf of e-line or kind of vertex
     */
    int nvkind;

    /* the number of loops in the node
     */
    int nvloop;

    /* whether the vertex touches to a loop assumed to be inside the node
     */
    int nvltype;

    /* the number of legs
     */
    int nvlegs;

    /* index of 4-gluon vertex.
     * In some code generation, 4-gluon vertex is considered as a set of
     * three different connection of two 3-gluon vertices.
     * These different connection are distinguised by this index.
     *
     *                      1             2           3
     *
     *      0     2     0        3   0        3   0        1
     *       \   /       \      /     \      /     \      /
     *        \ /         \    /       \    /       \    /
     *         *     =     +--+    +    +--+    +    +--+
     *        / \         /    \       /    \       /    \
     *       /   \       /      \     /      \     /      \
     *      1     3     1        3   2        1   1        2
     *
     *  value:
     *     0 : not 4-gluon vertex
     *     1 : (0, 1)--(2, 3)
     *     2 : (0, 2)--(3, 1)
     *     3 : (0, 3)--(1, 2)
     *     4 : 4 gluon vertex but not expanded
     */
    int nvgl4;

    /* pointer to the table of legs
     */
    Uint basel;

} tnode;

/* the type node numbers or type of indices in the table of nodes
 */
typedef int node;

/*------------------------------*/
/* The values of "tnode.nvkind" */
/*------------------------------*/
/* When the node is a vertex, the value of this variable is
 * the vertex number defined in the model.
 * For an external particle, this variable take one of the following
 * values.
 */
/* initial particle */
#define ELINITIAL  (-1)

/* final particle */
#define ELFINAL      1

/* undefined */
#define ELUNDEF      0

/*------------------------------*/
/* The values of "tnode.nvtype" */
/* The node is deleted one.
 */
#define NDDEL       -1

/* external line
 */
#define NDEXTERN     0

/* tree vertex
 */
#define NDTREEV      1

/* looped vertex
 */
#define NDLOOPEDV    2

/* on-shell line
 */
#define NDONSHELL    3

/* projection operator
 */
#define NDPROJ       4

/* null node index ???= nptable ???
 */
#define NDNULL    9999

/*-----------------------------*/
/* The values of tnode.nvltype */
/*-----------------------------*/
/* not yet defined
 */
#define NDLUNDEF     0

/* outside of loops
 */
#define NDLTREE      1

/* touching to a loop
 * some legs are connected to propagators in loops but others are not.
 */
#define NDLTOUCH     2

/* inside of loops
 * all legs are connected to propagators in loops.
 */
#define NDLINSIDE    3

/*-------------------------------------------------------------------
 *    tgraph : type of Feynman graphs
 */
typedef struct {
    /* graph-id */
    long         grid;

    /* sub-id of the graph.
     * When 4-gluon vertices are expanded, one graph correspondes to
     * a set of sub-graphs the sequential number of graphs.
     * This variable represents the sequence number of such a set
     * of graphs, which has the same vale of "grid".
     */
    long         subid;

    /* the number of nodes */
    int          nnodes;

    /* The number of internal lines.
     * If table of i-lines are not ready, value of this variable is -1.
     */
    int          nilines;

    /* In a graph, which is obtained by multiplying the two graphs
     * corresponding to a squared amplitude, the node number is
     * less or equal to this variable is belonging to the conjugated
     * graph.
     */
    int          conjsrt;

    /* The symmetric factor of the graph.
     */
    long         sfactor;

    /* The type of the graph.
     */
    long         gtype;

    /* the number of 4-gluon vertices.
     */
    int          ngl4;

    /* bit-string corresponding to the set of external particles.
     *
     *   0-th bit = 0  (bit at the right end : added artificially)
     *   n-th bit = 1  when node (n+1) is an external particle.
     *            = 0  when node (n+1) is not an external particle.
     */
    Uint         gbextp;

    /* bit-string corresponding to the set of final particles.
     *
     *   0-th bit = 0  (bit at the right end : added artificially)
     *   n-th bit = 1  when node (n+1) is final particle.
     *            = 0  when node (n+1) is not a final particle.
     */
    Uint gbcnsv;

    /* bit-string corresponding to the set of external particles.
     *
     *   0-th bit = 0  (bit at the right end : added artificially)
     *   n-th bit = 1  when node (n+1) is final particle.
     *            = 0  when node (n+1) is not a final particle.
     */
    Uint gbelem;

    /* pointer to the table of vertices
     */
    Uint         basen;
    Uint         offsl;
} tgraph;

/*-------------------------------------------------------------------
 *    tproc  : type of physical process
 */
typedef struct {
    /* The orders of coupling constants.
     */
/*
    int     morder[MAXSORD];
*/

    /* The orders of coupling constants.
     */
    int     sorder[MAXSORD];

    /* Process-id.
     */
    int     procid;

    /* The number of generated graphs in the process.
     */
    int     graphs;

    /* The number of graphs in the process including expansion
     * of gluon vertices.
     */
    int     egraphs;

    /* The number of external particles.
     */
    int     nextern;

    /* The number of original external particles.
     * The value is valid only for squared process.
     */
    int     oextern;

    /* The number of loops.
     */
    int     loop;

    /* The total order of coupling constants.
     */
    int     order;

    /* The order of qcd coupling constants.
     */
    int     qcd;

    /* The number of colored external particles.
     */
    int     extcol;

    /* The number of graphs in the buffer.
     */
    int     bgraph;

    /* The maximum possible number of nodes.
     */
    int     maxnodes;

    /* The maximum possible number of legs.
     */
    int     maxlegs;

    /* The first process for the squared amplitudes.
     */
    int     proc1;

    /* The second process for the squared amplitudes.
     */
    int     proc2;

    /* The memory size used for one graph.
     */
    int     msize;

    /* Whether the nodes in the graphs are not blob of looped
     * sub-graph.
     */
    bool    expand;

    /* Whether the graphs are one paritcle irreducible.
     */
    bool    onepi;

    /* Whether the graphs are generated for the calculation of g-2.
     */
    bool    gm2;

    /* Whether the graphs are including undefined particle.
     */
    bool    undefp;

    /* Whether the particle assignment is skiped
     */
    bool    skipasp;

    /* Whether the graphs are mltiplied ones.
     */
    int     mult;

    /* The path-name of kinematics library.
     */
    string  kinem;

    /* The prefix of generated code of amplitudes
     */
    string  pprefix;

} tproc;

/*
 * maxnodes = nextern +   loop; (without Majorana)
 * maxlegs  = nextern + 3*loop; (without Majorana)
 * msize    = (sizeof tgraph) + (sizeof tnode)*maxnodes
 *            + (sizeof tleg)*maxlegs
 */

/* The values of "tproc.mult". */
#define PC_NORMAL  0
#define PC_MULT    1
#define PC_PROJ    2
#define PC_CONI    3
#define PC_CONF    4
#define PC_CONIF   5
#define PC_CUTL    6
#define PC_CUTR    7  /* PC_CUTR = PC_CUTL + 1 */

/*-------------------------------------------------------------------
 *    tfline : type of fermion lines
 *-------------------------------------------------------------------
 *  The direction of f-line is defined opposit to the flow of fermion
 *  number, similer to the definition of vertex {anti-particle, particle}.
 *
 *  For normal QED case, it corresponts to
 *
 *      0--->---1--->---2
 *     ~u     gamma     u
 *
 */
typedef struct {
    int    type;      /* type of the f-line */
    int    length;    /* length of the f-line */
    int    fclass;    /* bit string of particle types on the f-line,
                         M_MAJORANA, M_DIRAC or M_GHOST defined in
                         "ptable.h" */
    node  *ndlist;    /* list of the nodes */
    int   *lglist[2]; /* list of the legs of the nodes 0:in, 1:out */
} tfline;

#define FL_UNDEF  -1
#define FL_IN     0        /* g-line comes into the vertex */
#define FL_OUT    1        /* g-line goes out from the vertex */
#define FL_BOTH   2        /* both is possible (Majorana) */

/*-------------------------*/
/* Values of "tfline.type" */
/*-------------------------*/
/* The f-line is not defined
 */
#define FLUNDEF  0

/* The f-line has external partices at its endpoints.
 */
#define FLOPEN   1

/* The f-line is a loop or a closed circle in a squared graph.
 */
#define FLCLOSED 2

/*-------------------------------------------------------------------
 *    tiline : type of internal lines
 *-------------------------------------------------------------------
 * The momentum flows from leg 0 to 1.
 *
 *        1 ---<--- 0
 *
 * The direction of particle number flow is independent of the direction
 * of i-line.  This direction is determined from the information
 * of legs of nodes.
 */

/* The type of the elements of the table of momentum carried by i-lines
 */
typedef int tmom;

typedef struct {
    /* The type of the internal line.
     */
    int    iltype;

    /* The identifier of the internal line.
     */
    int    ilname;

    /* The block-id, in which the internal line belongs.
     */
    int    ilblock;

    /* The nodes at both end-points of the i-line
     */
    int    ilnodes[2];

    /* The legs of nodes connect to both end-points
     */
    int    illegs[2];

    /* The f-line number (-1 when the i-line is not on a fline)
     */
    int    ilfline;

    /* The direction of i-line along f-line
     *            particle
     *        1 -----<----- 0
     *   fline n   -->-- n+1  ilfldir = 1
     *   fline n+1 --<-- n    ilfldir = -1
     *   not on a fline       ilfldir = 0
     */
    int    ilfldir;

    /* momentum conservation
     */
    tmom  *ilmlist;

    /* The momentum of internal line
     *       = s1*p1 + ... + se*pe + t1*k1 + ... tl*kl
     *
     *    e  : the number of external particles
     *    l  : the number of loops
     *    si : 1, -1 or 0
     *    pi : external momentum.
     *    tj : 1, -1 or 0
     *    kj : loop momentum.
     *
     * The information of si, tj is kept in two bit strings
     *
     *   ilmabs = bit-string[fkl, ..., fk1, fpe, ..., fp1, gs]
     *   ilmsgn = bit-string[ftl, ..., ft1, fse, ..., fs1,  0]
     *
     *     si = (1 - 2*fsi)*fpi
     *     tj = (1 - 2*ftj)*fkj
     *
     * When the graph is a tree (l = 0, no fkj nor ftj)
     *
     *     si = + gs      for initial particle.
     *     si = - gs      for final   particle.
     *
     * Additional flags are saved as
     *                                k, ..., 1,   e, ...,   1
     *     tgraph.gbextp = bit-string[0, ..., 0,   1, ...,   1, 0]
     *     tgraph.gbcnsv = bit-string[0, ..., 0, fme, ..., fm1, 0]
     *
     *         fmi = 0 for initial particle
     *         fmi = 1 for final   particle
     *
     *     With them, we can calculate
     *
     *         bit-string[0, ..., 0, fpe, ..., fp1, 0]
     *             = ilmabs & tgraph.gbextp;
     *
     *         bit-string[0, ..., 0, fse, ..., fs1, 0]
     *             = ilmabs & (
     *                  (ilmabs & 1) ? (~(tgraph.gbcnsv) & tgraph.gbextp)
     *                               : tgraph.gbcnsv;
     *               )
     */
    Ulong         ilmabs;   /* coded bit-string of abs(mom) */
    Ulong         ilmsgn;   /* coded bit-string of sign(mom) */

} tiline;

/*--------------------------*/
/* value of "tiline.iltype" */
/*--------------------------*/

/* The i-line is deleted
 */
#define ILDEL    -1

/* The i-line is connected to an external particle
 */
#define ILEXTERN  0

/* The i-line correspondes to an independent loop momentum.
 */
#define ILLOOP    1

/* The i-line carrays loop momenta.
 */
#define ILDLOOP   2

/* The i-line carrays a conjugated external momentum.
 */
#define ILECONJG  3

/* The i-line is a propagator outside of loop.
 */
#define ILTREE    4

/*-------------------------------------------------------------------
 * type of output function called from fofproc in outf.c
 */
typedef Void (*toptpr) ARG((FILE *));

/*-------------------------------------------------------------------
 *    The name of model.
 */
Global string   fnmdl;

/*-------------------------------------------------------------------
 * The current process
 */
Global  tproc    *cprocess;

/* The current output process number.
 */
Global  int       nprocess;

/*-------------------------------------------------------------------
 *  The current graph
 */
Global  tgraph   *cgraph;

/* The current graph number
 */
Global  int       ngraph;

/* The type of graph : "tgraph.gtype"
 */
Global  int       gntype;

/*-------------------------------------------------------------------
 *  The table of fermion lines
 */
Global  tfline  fltable[MAXFLINES];

/* the number of fermion lines in the table
 */
Global  int     nflines;


/*-------------------------------------------------------------------
 *  The table of fermion lines
 */
Global  tiline  *iltable;

/*-------------------------------------------------------------------
 *  Data representing one-loop structure.
 *  This version is not usable for 2 or more loops.
 *
 * ilxtable[0] -- ndxtable[0] -- ilxtable[1] -- ndxtable[1] -- ...
 * ilxtable[ilxnumb-1] -- ndxtable[ilxnumb-1] -- ilxtable[0]
 */
Global  int     *ilxtable; /* list of i-lines on a loop (one-loop) */
Global  int     *ndxtable; /* list of nodes on a loop (one-loop) */

/* The number of edges on the loop.
 */
Global  int     ilxnumb;

/* Flag of composit 4 gluon vertex
 */
#ifdef DEFGLOBAL
Global  bool    compgl4 = False;
#else
Global  bool    compgl4;
#endif

/*-------------------------------------------------------------------
 * Flag for generation of counter terms
 */
#ifdef DEFGLOBAL
Global  bool    genct = False;
#else
Global  bool    genct;
#endif

/*-------------------------------------------------------------------
 * Graph information (for 1 loop graphs)
 */
typedef struct {
    int   gid;          /* graph id */
    int   gilil;        /* the number of i-lines on the loop */
    int   gillegs;      /* the number of legs attached to the loop */
    int   gilpow;       /* power of loop momentum */
    int   gidiv;        /* divergent or not */
} GrInfo;
    
Global long    prcgraphs[MAXPROCS];
Global GrInfo *grinfo[MAXPROCS];

/*-------------------------------------------------------------------
 *  Macros defined for accessing data of a node or an edge.
 */
#define eliorf(x) (((x) == ELINITIAL) ? "initial" : "final")

#define nenodes(g)         (g)->nnodes
#define nenode(g,n)        ((tnode *)(((char*)(g))+(g)->basen)+(n))
#define netype(g,n)        nenode(g,n)->nvtype
#define nelegs(g,n)        nenode(g,n)->nvlegs
#define neloop(g,n)        nenode(g,n)->nvloop
#define neltype(g,n)       nenode(g,n)->nvltype
#define nevertex(g,n)      nenode(g,n)->nvkind
#define neiorf(g,n)        nenode(g,n)->nvkind
#define nebasel(g,n)       nenode(g,n)->basel
#define neleg(g,n,p)       ((tleg *)(((char*)(g))+nebasel(g,n))+(p))
#define neparticle(g,n,p)  neleg(g,n,p)->lgpart
#define nenextn(g,n,p)     neleg(g,n,p)->lgnext
#define neiline(g,n,p)     neleg(g,n,p)->lgiline
#define neilwk(g,n,p)      neleg(g,n,p)->lgwork
#define nename(g,e)        neparticle(g,e,0)

#define ndnodes()          nenodes(cgraph)
#define ndnode(n)          nenode(cgraph, n)
#define ndtype(n)          netype(cgraph, n)
#define ndlegs(n)          nelegs(cgraph, n)
#define ndloop(n)          neloop(cgraph, n)
#define ndltype(n)         neltype(cgraph, n)
#define ndvertex(n)        nevertex(cgraph, n)
#define ndiorf(n)          neiorf(cgraph, n)
#define ndbasel(n)         nebasel(cgraph, n)
#define ndleg(n,p)         neleg(cgraph, n, p)
#define ndparticle(n,p)    neparticle(cgraph, n, p)
#define ndnextn(n,p)       nenextn(cgraph, n, p)
#define ndiline(n,p)       neiline(cgraph, n, p)
#define ndilwk(n,p)        neilwk(cgraph, n, p)
#define ndname(e)          nename(cgraph, e)

#define nengl4(g)          (g)->ngl4
#define negl4(g,n)         nenode(g,n)->nvgl4

#define ndngl4()           nengl4(cgraph)
#define ndgl4(n)           negl4(cgraph, n)

/* direction of internal line.
 *      +1 : outgoing from the node (leg=1)
 *      -1 : ingoing to the node    (leg=0)
 */
#define ildir(lg)          (2*(lg) - 1)

/*==================================================================*/
/* Standard macros
 */
/* Macros to convert node number.
 *
 *  1. tonode : macro function to covert e-line or vertex number
 *              to node number.
 *  2. tovert : macro function to convert node number to
 *              e-line or vertex number.
 */
/*
 * #define tonode(x)   (((x)<0) ? (-(x)-1) : ((x)-1+cprocess->nextern))
 * #define tovert(x)   (((x) < cprocess->nextern) ? \
 *                      (-(x)-1) : ((x)+1-cprocess->nextern))
*/
#define tonode(x)   (x)
#define tovert(x)   (x)

#define boolstr(x)  ((x) ? "Yes" : "No")

#ifndef min
#define min(x, y)   (((x) < (y)) ?  (x) : (y))
#define max(x, y)   (((x) > (y)) ?  (x) : (y))
#endif

#ifndef abs
#define abs(x)      (((x) <  0 ) ? -(x) : (x))
#endif

#ifndef sign
#define sign(x)      (((x) <  0 ) ? (-1) : (1))
#endif

#include "ergrf.h"
#include "lprocs.h"
#include "lib.h"
