/********************************************************************
 *  File : path.c
 *
 *  Pathname
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "common.h"
#include "path.h"

#ifndef MAXPATHLEN
#define MAXPATHLEN 256
#endif

#define DEBUG   0

/*------------------------------------------------------------------
 * Private functions
 */
Private string pnfirst       ARG((string));
Private string pnset         ARG((string));
Private string pnnext        ARG((NoArg));
Private string pntest        ARG((string,string,string));
Private string pnwpath       ARG((string,string,bool*));

/*------------------------------------------------------------------
 * Local variable
 */
Local string envpath = NULL;


/*==================================================================*/
Public  string pnfname(env, def, subdir, fname)
string env;
string def;
string subdir;
string fname;
/*--------------------------------------------------------------------
 * Check whether a file named 'fname' is in the specified path or not.
 * returns the filename with path.
 */
{
    string pnam, fullnam;
    bool   withp;

    if((fullnam = pnwpath(subdir, fname, &withp)) != NULL) {
        return fullnam;
    }

    if(withp) {
        return NULL;
    }

    if(env != NULL){
#if DEBUG
        fprintf(Stderr, "%s\n", env);
#endif
        for(pnam = pnfirst(env); pnam != NULL; pnam = pnnext()) {
            if((fullnam = pntest(pnam, subdir, fname)) != NULL) {
                return fullnam;
            }
        }
    }
    if(def != NULL) {
#if DEBUG
        fprintf(Stderr, "def\n");
#endif
        for(pnam = pnset(def); pnam != NULL; pnam = pnnext()) {
            if((fullnam = pntest(pnam, subdir, fname)) != NULL) {
                return fullnam;
            }
        }
    }
    return NULL;
}

/*==================================================================*/
Private string pnwpath(subdir, fname, withp)
string subdir;
string fname;
bool   *withp;
/*--------------------------------------------------------------------
 */
{
    static char fullname[MAXPATHLEN];

    if(fname[0] == '/' || fname[0] == '.') {
        *withp = True;
        return pntest("", "", fname);

    } else if(strncmp(fname, "~/", 2) == 0) {
        Discard strncpy(fullname, getenv("HOME"), MAXPATHLEN);
        fullname[MAXPATHLEN-1] = '\0';
        *withp = True;
        return pntest(fullname, "", fname+2);

    } else if(subdir != NULL) {
        if(subdir[0] == '/' || fname[0] == '.') {
            *withp = True;
            return pntest(subdir, "", fname);
    
        } else if(subdir != NULL && strncmp(subdir, "~/", 2) == 0) {
            Discard strncpy(fullname, getenv("HOME"), MAXPATHLEN);
            fullname[MAXPATHLEN-1] = '\0';
            *withp = True;
            return pntest(fullname, subdir+2, fname);
        }

    }
    *withp = False;
    return NULL;
}

/*==================================================================*/
Public  bool   pnexpath(expath, opath)
string expath;
string opath;
/*--------------------------------------------------------------------
 * Expand "~/" located at the top of the path name.
 */
{
    if(strncmp(opath, "~/", 2) == 0) {
        Discard strncpy(expath, getenv("HOME"), MAXPATHLEN);
        expath[MAXPATHLEN-1] = '\0';
        Discard strncat(expath, opath+1, MAXPATHLEN);
        expath[MAXPATHLEN-1] = '\0';
        return True;

    } else if(strchr(opath, '/') != NULL) {
        Discard strcpy(expath, opath);
        return True;
    } else {
        return False;
    }
}

/*==================================================================*/
Public  bool   pnaddext(fall, fname, ext)
string fall;
string fname;
string ext;
/*--------------------------------------------------------------------
 * Add extension "ext" to the fname.
 */
{
    string pos;

    Discard strcpy(fall, fname);

    if((pos = strrchr(fname, '.')) == NULL) {
        strcat(fall, ext);
        return True;
    } else {
        return False;
    }
}

/*==================================================================*/
Private string pnfirst(env)
string env;
/*--------------------------------------------------------------------
 * Take out first path in the environment variable 'env'
 */
{
    if((envpath = getenv(env)) == NULL) {
        return NULL;
    }

    return pnnext();
}

/*==================================================================*/
Private string pnset(env)
string env;
/*--------------------------------------------------------------------
 * Take out first path in the environment variable 'env'
 */
{
    envpath = env;

    return pnnext();
}

/*==================================================================*/
Private string pnnext(NoArg)
/*--------------------------------------------------------------------
 * Take out next path in the environment variable 'env'
 */
{
    static char pathn[MAXPATHLEN+1];
    string  p = envpath;
    int     i = 0;

    if(envpath == NULL || *envpath == '\0') {
        return NULL;
    }

    for( ; *envpath != '\0' && *envpath != ':'; envpath++) {
        if(i >= MAXPATHLEN-1) {
            Discard fprintf(Stderr, "*** pnnext : too long path name %s\n", p);
            exit(1);
        }
        pathn[i++] = *envpath;
    }
    pathn[i] = '\0';
    if(*envpath == '\0') {
        envpath = NULL;
    } else {
        envpath++;
    }

    return pathn;
}

/*==================================================================*/
Private string pntest(pathn, subdir, fname)
string  pathn;
string  subdir;
string  fname;
/*--------------------------------------------------------------------
 * Check if the file fname in the path.
 */
{
    static char fullnam[MAXPATHLEN+1];
    int    l;

    l = strlen(pathn);

    if(l >= MAXPATHLEN - 2) {
        Discard fprintf(Stderr, "*** pntest : too long path name %s \n",
                pathn);
        exit(1);
    }
    Discard strcpy(fullnam, pathn);
    if(l > 0 && fullnam[l-1] != '/') {
        fullnam[l]   = '/';
        fullnam[l+1] = '\0';
        l++;
    }

    if(subdir != NULL) {
        if(l + strlen(subdir) >= MAXPATHLEN - 2) {
            Discard fprintf(Stderr, "*** pntest : too long path name %s %s\n",
                    pathn, subdir);
            exit(1);
        }
        l += strlen(subdir);
        Discard strcat(fullnam, subdir);
        if(l > 0 && fullnam[l-1] != '/') {
            fullnam[l]   = '/';
            fullnam[l+1] = '\0';
            l++;
        }
    }

    if(l + strlen(fname) >= MAXPATHLEN - 1) {
        Discard fprintf(Stderr,
                "*** pntest : too long path and file name %s%s \n",
                fullnam,fname);
        exit(1);
    }
    Discard strcat(fullnam, fname);

#if DEBUG
    fprintf(Stderr, "test  : l=%2d |%s|%s|==>|%s|\n", l, pathn, fname, fullnam);
#endif

    if(access(fullnam, R_OK) == 0) {
        fprintf(Stderr, "+++ \"%s\" is found\n", fullnam);
        return fullnam;
    } else {
        fprintf(Stderr, "*** \"%s\" is not found\n", fullnam);
        return NULL;
    }
}

/********************* main for test ********************************
int main(argc, argv)
int     argc;
string *argv;
{
    if(argc != 2) {
        fprintf(Stderr, "%s <filename>\n", argv[0]);
        exit(0);
    }

    fprintf(Stderr, "found : |%s|\n", pnfname("PATH", ".:../elwk", "", argv[1]));
    return 0;
}
*/
