/*******************************************************************
 *    File : rdgrf.c
 *
 *    Read graph
 *
 *    Calling sequence:
 *
 *      1. Initialization
 *
 *              if(!rginit(<filename>)) {
 *                  exit(FAIL);
 *              }
 *
 *      2. Read loop
 *
 *              while(rgproc(tproc*)) {
 *                  while(rgraph(tgraph*)) {
 *                    ...
 *                  }
 *              }
 *
 *      3. Close
 *
 *              rgclose();
 *
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "grdata.h"
#include "version.h"
#include "token.h"

/*--------------------------------------------------------------------
 * Private functions
 */
Private bool    rgexternal   ARG((int));
Private bool    rgnode       ARG((node, int, int*));
Private pindex  rgene        ARG((int, string, int));
Private Void    rgndopt      ARG((int, int*, int*, int*));
Private bool    rgenv        ARG((int,int,int,int*,string*,int*,int,int*));
Private Void    rgerror      ARG((string, string));
Private int     rgcnext0     ARG((int, int, int*, int*));
Private Void    rgccon0      ARG((int, int, int, int, int));
Private Void    rgccon       ARG((int, int, int, int, int, bool));

/*--------------------------------------------------------------------
 * Local variable
 */
#define ilpartl(l,lg)     \
        ndparticle(iltable[l].ilnodes[lg],iltable[l].illegs[lg])
#define ilpart(l)     ilpartl(l, 1)

#define ELPART 0
#define ELIORF 1
Local   char    rgmdl[MAXPATHLEN];
Local   int     eltbl[MAXELINE][ELIORF+1];
Local   int     edgetbl[MAXNODES][MAXNODES];
Local   FILE    *rgfp;
Local   int     nvertex;
Local   int     rgind;
Local   int     extcol = 0;
Local   bool    setver = False;

/*==================================================================*/
Private Void    rgerror(fncn, msg)
string    fncn;
string    msg;
/*--------------------------------------------------------------------
 */
{
    fprintf(Stderr, "*err* %s : %s\n", fncn, msg);
}

/*==================================================================*/
Public  bool    rginit(fn)
string    fn;
/*--------------------------------------------------------------------
 *  initializes read package.
 *
 *   1. open grf file.
 *   2. read mdl file.
 */
{
    static  int rdmdl=0;
    FILE*   echo;
    int     ind, i;
    trdsava sarea;
    int     grfver[MAXVER];

#if DBG
    if(dbg >= rgdbg) {
        printf("rginit(%s)\n", fn);
    }
#endif

    /* echo : file pointer for echo of input. */
    echo = NULL;        /* no echo */

    /* open input file */
    rgfp = fopen(fn, "r");
    if(rgfp == NULL) {
        fprintf(Stderr, "*err* rginit : open err fname = %s\n", fn);
        return False;
    }

    /* initialize token package */
    rdinit(rgfp, RDNOCHANGE, echo);

    /* version number */
    rgind = rdstat();
    if(rdchkid(rgind, "version")) {

        /* do not redefine */
        for(ind = rdarg(rgind), i = 0; rdtype(ind) == RDINT;
             ind = rdarg(ind), i++) {
            if(i < MAXVER) {
                grfver[i] = (int)rdint(ind);
            } else {
                fprintf(Stderr, "\n*err* pvread : ");
                fprintf(Stderr, "too many sub-version level\n");
                exit(FAIL);
            }
        } /* for ind */
        if(grfver[0] != GRCVER0 || grfver[1] != GRCVER1 || 
           grfver[2] != GRCVER2) {
            rgerror("rginit", "Illegal version");
            fprintf(Stderr,"    grc  %d.%d.%d\n",
                    GRCVER0, GRCVER1, GRCVER2);
            fprintf(Stderr,"    file %d.%d.%d\n",
                    grfver[0], grfver[1], grfver[2]);
            exit(FAIL);
        }
        setver = True;
    } else {
        rgerror("rginit", "version = {int, ...}; is expected.");
        setver = False;
    }

    /* input : model="<model file>"; */
    if(setver) {
        rgind = rdstat();
    }
    ind   = rdarg(rgind);
    fnmdl = rgmdl;

    if(rdchkid(rgind, "model") && rdtype(ind) == RDSTRING) {
        if(rdmdl == 0) {
            fnmdl = rgmdl;
            pvfname(fnmdl, rdid(ind), MAXPATHLEN);
            rdsave(&sarea);
            pvread(fnmdl);       /* read particle table */
            rdrestore(&sarea);   /* recover fp in the token package */
        } else if(strcmp(fnmdl, rdid(ind)) == 0) {
        } else {
            pvfname(fnmdl, rdid(ind), MAXPATHLEN);
            rgerror("rginit", "different model");
            fprintf(Stderr,"    previous : %s\n", fnmdl);
            fprintf(Stderr,"    new      : %s\n", rdid(ind));
            exit(1);
        }
        rdmdl++;
    } else {
        rgerror("rginit", "model = \"<filename>\"; is expected.");
        rdprint(rgind);
        ercont();
        return False;
    }

#if DBG
    if(dbg >= rgdbg) {
        printf("end of rginit\n");
    }
#endif
    return True;
}

/*==================================================================*/
Public  bool    rgproc(proc)
tproc *proc;
/*--------------------------------------------------------------------
 *    read header of a process
 *
 *    1. process number
 *    2. table of external particles
 *    3. order of coupling constants
 *    4. the number of loops
 *    5. flags.
 */
{
    int     ind, io, arg, j, dummy;
    bool    st;
    int     nextern, order, loop, onepi, expand, msize, gm2, skipasp;
    int     assign;
    string  kinem;
    string  prefix;

    cprocess = proc;
    nprocess  = -1;
    nextern   = -1;
    order     = 0;
    loop      = -1;
    onepi     = True;
    expand    = True;
    gm2       = False;
    skipasp   = False;
    kinem     = NULL;
    prefix    = "";

    for(io = 0; io < nsordtbl; io++) {
        cprocess->sorder[io] = 0;
    }

    st = True;
    while (True) {
        /* read : process=<int>; */
        rgind = rdstat();
        if(rdtype(rgind) != RDID) {
            return False;
        }

        ind = rdarg(rgind);

        /* read : process=<int>; */
        if(rdstrcmp(rdid(rgind), "end") == 0) {
            return False;
        } else if(rdstrcmp(rdid(rgind), "process") == 0) {
            if(rdtype(ind) == RDINT) {
                nprocess = rdint(ind);
            } else {
                rgerror("rginit", "process = <int>");
                rdprint(rgind);
                st = False;
            }

        /* read : external=<int>; */
        } else if(rdstrcmp(rdid(rgind), "external") == 0) {
            if(rdtype(ind) == RDINT) {
                nextern = rdint(ind);
            } else {
                rgerror("rginit", "external = <int>");
                rdprint(rgind);
                st = False;
            }
            /* read table of external particles */
            st = rgexternal(nextern);

        /* read : order=<int>; */
        /*
        } else if(rdstrcmp(rdid(rgind), "order") == 0) {
            if(rdtype(ind) == RDINT) {
                order = rdint(ind);
            } else {
                rgerror("rginit", "order = <int>");
                fprintf(Stderr,"*err* rginit : order = <int>\n");
                rdprint(rgind);
                st = False;
            }
        */

        /* read : loop=<int>; */
        } else if(rdstrcmp(rdid(rgind), "loop") == 0) {
            if(rdtype(ind) == RDINT) {
                loop = rdint(ind);
            } else {
                rgerror("rginit", "loop = <int>");
                rdprint(rgind);
                st = False;
            }

        /* read : opi=Yes|No; */
        } else if(rdstrcmp(rdid(rgind), "opi") == 0) {
            if(!rdbool(ind, &onepi)) {
                rgerror("rginit", "opi = <Yes/No>");
                rdprint(rgind);
                st = False;
            }

        /* read : expand=Yes|No; */
        } else if(rdstrcmp(rdid(rgind), "expand") == 0) {
            if(!rdbool(ind, &expand)) {
                rgerror("rginit", "expand = <Yes/No>");
                rdprint(rgind);
                st = False;
            }

        /* read : gm2=Yes|No; */
        } else if(rdstrcmp(rdid(rgind), "gm2") == 0) {
            if(!rdbool(ind, &gm2)) {
                rgerror("rginit", "gm2 = <Yes/No>");
                rdprint(rgind);
                st = False;
            }

        /* read : assign=Yes|No; */
        } else if(rdstrcmp(rdid(rgind), "assign") == 0) {
            if(!rdbool(ind, &assign)) {
                rgerror("rginit", "Assign = <Yes/No>");
                rdprint(rgind);
                st = False;
            } else {
                skipasp = ! assign;
            }

        /* read : kinem=string; */
        } else if(rdstrcmp(rdid(rgind), "kinem") == 0) {
            arg = rdarg(rgind);
            if(rdtype(arg) == RDSTRING){
                kinem = (string) malloc((strlen(rdid(arg))+1)*sizeof(char));
                if(kinem == NULL) {
                    fprintf(Stderr, "\n*err* rpget: memory allocation error\n");
                    exit(1);
                }
                strcpy(kinem, rdid(arg));
            } else {
                fprintf(Stderr, "\n*err* rpget: cptest should integer\n");
                rdprint(rgind);
                st = False;
            }

        /* read : prefix=string; */
        } else if(rdstrcmp(rdid(rgind), "prefix") == 0) {
            arg = rdarg(rgind);
            if(rdtype(arg) == RDSTRING){
                prefix = (string) malloc((strlen(rdid(arg))+1)*sizeof(char));
                if(prefix == NULL) {
                    fprintf(Stderr, "\n*err* rpget: memory allocation error\n");
                    exit(1);
                }
                strcpy(prefix, rdid(arg));
            } else {
                fprintf(Stderr, "\n*err* rpget: cptest should integer\n");
                rdprint(rgind);
                st = False;
            }

        /* read : expmin=int; not used */
        } else if(rdstrcmp(rdid(rgind), "expmin") == 0) {
            if(rdtype(ind) != RDINT) {
                rgerror("rginit", "expmin = int");
                rdprint(rgind);
                st = False;
            }

        /* read : expmax=int; not used */
        } else if(rdstrcmp(rdid(rgind), "expmax") == 0) {
            if(rdtype(ind) != RDINT) {
                rgerror("rginit", "expmax = int");
                rdprint(rgind);
                st = False;
            }

        /* read : block=Yes|No; not used */
        } else if(rdstrcmp(rdid(rgind), "block") == 0) {
            if(!rdbool(ind, &dummy)) {
                rgerror("rginit", "block = <Yes/No>");
                rdprint(rgind);
                st = False;
            }

        /* read : tadpole=Yes|No; not used */
        } else if(rdstrcmp(rdid(rgind), "tadpole") == 0) {
            if(!rdbool(ind, &dummy)) {
                rgerror("rginit", "tadpole = <Yes/No>");
                rdprint(rgind);
                st = False;
            }

        /* read : extself=Yes|No; not used */
        } else if(rdstrcmp(rdid(rgind), "extself") == 0) {
            if(!rdbool(ind, &dummy)) {
                rgerror("rginit", "extself = <Yes/No>");
                rdprint(rgind);
                st = False;
            }

        /* read : selfe=Yes|No; not used */
        } else if(rdstrcmp(rdid(rgind), "selfe") == 0) {
            if(!rdbool(ind, &dummy)) {
                rgerror("rginit", "selfe = <Yes/No>");
                rdprint(rgind);
                st = False;
            }

        /* read : countert=Yes|No; not used */
        } else if(rdstrcmp(rdid(rgind), "countert") == 0) {
            if(!rdbool(ind, &dummy)) {
                rgerror("rginit", "countert = <Yes/No>");
                rdprint(rgind);
                st = False;
            }

        /* read : anyct=Yes|No; not used */
        } else if(rdstrcmp(rdid(rgind), "anyct") == 0) {
            if(!rdbool(ind, &dummy)) {
                rgerror("rginit", "anyct = <Yes/No>");
                rdprint(rgind);
                st = False;
            }

        /* read : undefp=Yes|No; not used */
        } else if(rdstrcmp(rdid(rgind), "undefp") == 0) {
            if(!rdbool(ind, &dummy)) {
                rgerror("rginit", "undefp = <Yes/No>");
                rdprint(rgind);
                st = False;
            }

        /* read : hep=Yes|No; not used */
        } else if(rdstrcmp(rdid(rgind), "hep") == 0) {
            if(!rdbool(ind, &dummy)) {
                rgerror("rginit", "hep = <Yes/No>");
                rdprint(rgind);
                st = False;
            }

        } else {
            for(io = 0; io < nsordtbl; io++) {
                if(rdstrcmp(rdid(rgind), sordtbl[io]) == 0) {
                    cprocess->sorder[io] = rdint(ind);
                    order += cprocess->sorder[io];
                    break;
                }
            }
            if(io >= nsordtbl) {
                break;
            }
        }
    }
    if(nprocess < 0) {
        rgerror("rginit", "process = <int>");
        st = False;
    }
    if(nextern < 0) {
        rgerror("rginit", "external = <int>");
        st = False;
    }
    if(order <= 0) {
        rgerror("rginit", "order = <int>");
        fprintf(Stderr,"*err* rginit : \"order\"=<int>\n");
        st = False;
    }
    if(loop < 0) {
        rgerror("rginit", "loop = <int>");
    }

    /* save the process */
#if DBG
    if(dbg >= rgdbg) {
        printf("st = %d\n", st);
    }
#endif
    if(st) {
        if(proc == NULL) {
            rgerror("rgproc", "null proc pointer");
            exit(1);
        }
        cprocess->procid    = nprocess;
        cprocess->graphs    = 0;
        cprocess->nextern   = nextern;
        cprocess->oextern   = nextern;
        cprocess->order     = order;
        cprocess->loop      = loop;
        cprocess->onepi     = onepi;
        cprocess->expand    = expand;
        cprocess->gm2       = gm2;
        cprocess->skipasp   = skipasp;
        cprocess->kinem     = kinem;
        cprocess->pprefix   = prefix;
        cprocess->maxnodes  = nextern +   order;
        cprocess->maxlegs   = nextern + 3*order;
        msize =   sizeof(tgraph)
                + sizeof(tnode) * cprocess->maxnodes
                + sizeof(tleg)  * cprocess->maxlegs;
        cprocess->msize     = msize;
        cprocess->extcol    = extcol;   /* see rgexternal() */

        /* order of QCD coupling constant */
        cprocess->qcd = 0;
        for(io = 0; io < nmordtbl; io++) {
            if((mordopt[io] & MOPT_QCD) == 0) {
                continue;
            }
            for(j = 0; j < nsordtbl; j++) {
                if(sordset[j][io] != 0) {
                    continue;
                }
                cprocess->qcd += cprocess->sorder[j];
            }
        }
    }


#if DBG
    if(dbg >= rgdbg) {
        printf("end of rgproc\n");
    }
#endif
    return st;
}

/*==================================================================*/
Public  bool    rgclose(NoArg)
/*--------------------------------------------------------------------
 *  closes read packages.
 */
{
#if DBG
    if(dbg >= rgdbg) {
        printf("rgclose()\n");
    }
#endif
    if(rgfp != NULL) {
        rdend();
        fclose(rgfp);
        rgfp = NULL;
    }

    return True;
}

/*==================================================================*/
Private bool    rgexternal(nextern)
int nextern;
/*--------------------------------------------------------------------
 *  read the table of external particles.
 */
{
    int i, a, iorf, n;
    bool st;
    pindex pc;

#if DBG
    erdbg("rgexternal", "begin", rgdbg);
    if(dbg >= rgdbg) {
        printf("rgexternal(%d)\n", nextern);
    }
#endif

    st = True;
    extcol = 0;
    for(i = 0; i < nextern; i++) {
        rgind = rdstat();

        /* external particle number */
        n = tonode(rdint(rgind));
        if(i != n) {
            fprintf(Stderr, "*err* rgexternal : illegal order\n");
            rdprint(rgind);
            st = False;
            break;
        }

        /* initial / final */
        a = rdarg(rgind);
        if(rdchkid(a, "initial")) {
            iorf = ELINITIAL;
        } else if(rdchkid(a, "final")) {
            iorf = ELFINAL;
        } else {
            fprintf(Stderr, "*err* rgexternal : ");
            fprintf(Stderr, "not initial nore final\n");
            rdprint(rgind);
            st = False;
            break;
        }

        /* particle name */
        a = rdarg(a);
        if(rdtype(a) != RDID) {
            fprintf(Stderr, "*err* rgexternal : ");
            fprintf(Stderr, "illegal particle name.\n");
            rdprint(rgind);
            st = False;
            break;
        }

        /* save */
        pc = rgene(n, rdid(a), iorf);    /* when error stop */
        if(ptable[abs(pc)]->color != 1) {
            extcol++;
        }
#if DBG
        if(dbg >= rgdbg) {
            rdprint(rgind);
        }
#endif
    }

    if(!st) {
        while (!rdchkid(rgind, "eend"))
            rgind = rdstat();
        exit(1);
    }
    rgind = rdstat();
    st = rdchkid(rgind, "eend");

#if DBG
    erdbg("rgexternal", "end", rgdbg);
#endif

    return st;
}

/*==================================================================*/
Public  bool    rgraph(grf)
tgraph  *grf;
/*--------------------------------------------------------------------
 *  returns False if EOF or error is found.
 */
{
    int i, j, lp, lg;
    bool st;
    int  legs;
    long grn;
    int     nlgtbl;  /* counter of allocated legs */
    node nd;

#if DBG
    erdbg("rgraph", "begin", rgdbg);
#endif
    st = True;

    if(grf == NULL) {
        rgerror("rgraph", "null graph\n");
        exit(1);
    }

    rgigraph(cprocess, grf);

    nlgtbl = 0;  /* the number of allocated legs */

    cgraph = grf;
    ndnodes() = cprocess->nextern;  /* necessary ??? */

    /* "graph=<int>;" */
    if(rdkval(rgind, "pend", &grn)) {
        return False;
    } else if(!rdkval(rgind, "graph", &grn)) {
        fprintf(Stderr, "*err* rgraph : graph=<int>; expected\n");
        rdprint(rgind);
        rdskpid("gend");
        rgind = rdstat();
        exit(1);
    }
    cgraph->grid  = grn;
    cgraph->subid = 0;

    /* "gtype=<int>;" */
    rgind = rdstat();
    if(!rdkval(rgind, "gtype", &grn)) {
        fprintf(Stderr, "*err* rgraph : gtype=<int>; expected\n");
        rdprint(rgind);
        rdskpid("gend");
        rgind = rdstat();
        exit(1);
    }
    cgraph->gtype = grn;

    /* "sfactor=<int>;" */
    rgind = rdstat();
    if(!rdkval(rgind, "sfactor", &grn)) {
        fprintf(Stderr, "*err* rgraph : sfactor=<int>; expected\n");
        rdprint(rgind);
        rdskpid("gend");
        rgind = rdstat();
        exit(1);
    }
    cgraph->sfactor = grn;

    /* "vertex=<int>;" */
    rgind = rdstat();
    if(!rdkval(rgind, "vertex", &grn)) {
        fprintf(Stderr, "*err* rgraph : vertex=<int>; expected\n");
        rdprint(rgind);
        rdskpid("gend");
        rgind = rdstat();
        exit(1);
    }
    nvertex = grn;

    ndnodes() = cgraph->conjsrt = cprocess->nextern + nvertex;

    /* read nodes line */
    for(i = 0; i < ndnodes(); i++) {
        rgind = rdstat();
        if(!rgnode(i, ndnodes(), &nlgtbl)) {
#if DBG
            if(dbg >= rgdbg) {
                printf("st of rgnode = False\n");
            }
#endif
            st = False;
            break;
        }
#if DBG
        if(dbg >= rgdbg) {
            printf("node = %d ",i);
            rdprint(rgind);
        }
#endif
    }

    /* "vend;" */
    rgind = rdstat();
    if(!rdchkid(rgind, "vend")) {
        fprintf(Stderr, "*err* rgraph : vend;");
        fprintf(Stderr, " nnodes = %d, nexternal = %d, nvertex = %d\n",
                ndnodes(), cprocess->nextern, nvertex);
        rdprint(rgind);
        rdskpid("gend");
        rgind = rdstat();
        return False;
    }

    /* "gend=<number>;" */
    rgind = rdstat();
    if(!rdchkid(rgind, "gend")) {
        fprintf(Stderr, "*err* rgraph : gend;\n");
        rdprint(rgind);
        rdskpid("gend");
        rgind = rdstat();
        return False;
    }

    rgind = rdstat();

    /* construct internal line table and ndnextn(*,*) */
    cgraph->nilines = ilmake(True);

    /* construct edge table */
    for(i = 0; i < ndnodes(); i++) {
        for(j = 0; j < ndnodes(); j++) {
            edgetbl[i][j] = 0;
        }
    }

    for(nd = 0; nd < ndnodes(); nd++) {
        for(lg = 0; lg < ndlegs(nd); lg++) {
            edgetbl[nd][ndnextn(nd,lg)]++;
        }
    }

    /* calculate the number of loops */
    legs = 0;
    for(i = 0; i < ndnodes(); i++) {
        for(j = 0; j < ndnodes(); j++) {
            legs += edgetbl[i][j];
        }
    }

    lp = cprocess->loop;
    for(i = 0; i < ndnodes(); i++) {
        lp -= ndloop(i);
    }

    if(st) {
        cprocess->graphs++;
    }

#if DBG
    erdbg("rgraph", "end", rgdbg);
#endif

    return st;
}

/*==================================================================*/
Public  Void    rgigraph(proc, grf)
tproc *proc;
tgraph   *grf;
/*--------------------------------------------------------------------
 *  initialize graph data
 */
{
    int  ptr;

    ptr        = sizeof(tgraph);
    grf->basen = ptr;
    ptr       += sizeof(tnode)*(proc->maxnodes);
    grf->offsl = ptr;

    grf->nnodes = 0;
    grf->ngl4   = 0;

    grf->gbextp = 0L;
    grf->gbcnsv = 0L;
    grf->gbelem = 0L;

#if DBG
    if(dbg>=rgdbg) {
        rgdump(cprocess, cgraph);
    }
#endif
}

/*==================================================================*/
Public  tgraph *rgalloc(proc, ngrf)
tproc *proc;
int       ngrf;
/*--------------------------------------------------------------------
 *  Allocate area for "ngrf" graphs to "grf".
 */
{
    int     n, msize;
    char   *ptr;
    tgraph *grf;

    msize = proc->msize;

    if((ptr = (char *) malloc(ngrf*msize)) == NULL) {
        fprintf(Stderr, "rgalloc : memory allocation failed.\n");
        exit(16);
    }
    grf = (tgraph *) ptr;
    for(n = 0; n < ngrf; n++) {
#if DBG
        if(dbg >= rgdbg) {
            printf("rgalloc : ptr[%d] = %p\n", n, ptr);
        }
#endif
        rgigraph(proc, (tgraph *) ptr);
        ptr += msize;
    }

    return grf;
}
/*==================================================================*/
Private bool    rgnode(nd, nnd, nlgtbl)
node    nd;
int     nnd;
int    *nlgtbl;
/*--------------------------------------------------------------------
 *  reads a node and its connections.
 */
{
    int    legn[MAXNODES];
    string legp[MAXNODES];
    int    mord[MAXMORD];
    int  nleg, a, b, lp, i, noprop;
    bool st;
#if DBG
    int  l;
#endif

#if DBG
    erdbg("rgnode", "begin", rgdbg);
    if(dbg >= rgdbg) {
        printf("rgnode(%d, %d)\n", nd, nnd);
    }
#endif
    /* node number */
    if(rdtype(rgind) != RDINT) {
        rgerror("rgnode", "number is expected");
        rdprint(rgind);
        rdpstat(rgind, 0);
        return False;
    }
    if(nd != tonode(rdint(rgind))) {
        rgerror("rgnode", "illegal node number.");
        rdprint(rgind);
        rdpstat(rgind, 0);
        return False;
    }

    /* options to the node */
    lp = 0;
    for(i = 0; i < nmordtbl; i++) {
        mord[i] = 0;
    }
    noprop = False;
    if((a = rdopt(rgind)) >= 0) {
        rgndopt(a, &lp, mord, &noprop);
    }

    /* legs */
    nleg = 0;
    for(a = rdarg(rgind); a >= 0; a = rdarg(a)) {
        if(rdtype(a) != RDINT) {
            rgerror("rgnode", "not i-line number for a leg.");
            rdprint(rgind);
            return False;
        }
        if((legn[nleg] = rdint(a)) < 1) {
            rgerror("rgnode", "illegal i-line number for a leg.");
            rdprint(rgind);
            return False;
        }

        b = rdopt(a);

        if(rdtype(b) == RDID) {
            legp[nleg++] = rdid(b);
        } else {
            legp[nleg++] = "undef";
        }
    }

#if DBG
    if(dbg >= rgdbg) {
        printf("nd = %d, nleg = %d\n", nd, nleg);
        for(l = 0; l < nleg; l++) {
            printf(" %d", legn[l]);
        }
        printf("\n");
    }
#endif

     /* save */
    st = rgenv(nd, nleg, lp, legn, legp, mord, noprop, nlgtbl);  /* save */

#if DBG
    if(dbg >= rgdbg) {
        printf("rgnode : st = %d\n", st);
        rdprint(rgind);
    }
#endif

#if DBG
    erdbg("rgnode", "end", rgdbg);
#endif

    return st;
 }

/*==================================================================*/
Private pindex  rgene(n, ptcl, iorf)
int n;
string ptcl;
int iorf;
/*--------------------------------------------------------------------
 *  save external particle.
 */
{
    pindex pc;

#if DBG
    if(dbg >= rgdbg) {
        printf("rgene(%d, %s, %d)\n", n, ptcl, iorf);
    }
#endif
    if(n >= MAXELINE) {
        rgerror("rgene", "too many e-lines");
        exit(1);
    }

    pc = pvfind(ptcl);
    if(iorf == ELINITIAL) {
        pc *= ptable[abs(pc)]->antip;
    }


    eltbl[n][ELIORF] = iorf;
    eltbl[n][ELPART] = pc;

    return pc;
}

/*==================================================================*/
Public  Void    rgsetv(proc, grf, v, lg, nlgtbl)
tproc  *proc;
tgraph *grf;
int    v;
int    lg;
int    *nlgtbl;
/*--------------------------------------------------------------------
 *  setup table of legs.
 */
{
    if(*nlgtbl + lg > proc->maxlegs) {
        rgerror("rgsetv", "leg table overflow");
        fprintf(Stderr, " %d+%d > %d\n", *nlgtbl, lg, proc->maxlegs);
        rgdump(proc, NULL);
        exit(1);
    }

    nebasel(grf, v) = grf->offsl + (*nlgtbl)*sizeof(tleg);
    *nlgtbl += lg;
}

/*==================================================================*/
Private bool    rgenv(v, lg, lp, legn, legp, mord, noprop, nlgtbl)
int    v;
int    lg;
int    lp;
int    legn[];
string legp[];
int    mord[];
int    noprop;
int    *nlgtbl;
/*--------------------------------------------------------------------
 *  save node
 */
{
    node n;
    int  p, i, tord, lord;
    bool pdef;

#if DBG
    if(dbg >= rgdbg) {
        printf("rgenv(%d, %d, %d, %p)\n", v, lg, lp, legn);
        printf("rgenv : %d+%d > %d\n", *nlgtbl, lg, cprocess->maxlegs);
    }
#endif

    rgsetv(cprocess, cgraph, v, lg, nlgtbl);

    if(v < cprocess->nextern) {
        ndtype(v)   = NDEXTERN;
        ndlegs(v)   = 1;
        ndloop(v)   = 0;
        ndltype(v)  = NDLUNDEF;
        ndgl4(v)    = 0;
        ndiorf(v)   = eltbl[v][ELIORF];
        if(lg != 1) {
            rgerror("rgenv", "inconsistent e-line");
            exit(1);
        }
    } else {
        tord = 0;
        for(i = 0; i < nmordtbl; i++) {
            tord += mord[i];
        }
        lord = 2*lp + lg - 2;
        if(tord == 0) {
            mord[0] = lord;
        } else if(tord != lord) {
            for(i = 0; i < nmordtbl; i++) {
                fprintf(Stderr, "mord[%d] = %d\n", i, mord[i]);
            }
            fprintf(Stderr, "lp=%d, lg=%d, lord=%d, tord=%d\n",
                    lp, lg, lord, tord);
            rgerror("rgenv", "illegal ord. of coupling const.");
            prdump();
            exit(1);
        }
        ndtype(v)   = (lp == 0) ? NDTREEV :NDLOOPEDV;
        ndlegs(v)   = lg;
        ndloop(v)   = lp;
        ndltype(v)  = NDLUNDEF;
        ndgl4(v)    = 0;
        ndvertex(v) = VNULL;
        ndgl4(v)    = 0;
    }

    for(n = 0; n < ndnodes(); n++) {
        edgetbl[v][n] = 0;
    }
    pdef = True;
    for(p = 0; p < lg; p++) {
        ndiline(v, p)    = legn[p];
        ndparticle(v, p) = pvfind(legp[p]);
        if(ndparticle(v, p) == PNULL) {
            pdef = False;
        }
    }

    if(v >= cprocess->nextern) {
        if(cprocess->skipasp) {
            pdef = False;
        }
        if(pdef) {
            if((ndvertex(v) = mvmatc(v, mord)) != VNULL) {
                pvreord(v);
            }
            if((ndgl4(v) = pvgl4(v)) == 4) {
                ndngl4()++;
            }
        } else {
            ndvertex(v) = VNULL;
            ndgl4(v) = 0;
        }
    }

    return True;
}

/*==================================================================*/
Private Void rgndopt(opt, lp, mord, noprop)
int opt;
int *lp;
int *mord;
int *noprop;
/*--------------------------------------------------------------------
 *  read option for the node
 *
 *    *lp   : loop  <= "loop=<int>;"
 *    *mord : order <= "order={<int>,...};"
 */
{
    int  arg, i;

#if DBG
    if(dbg >= rgdbg) {
        printf("rgndopt(%d, %p)\n", opt, lp);
    }
#endif

    for(i = 0; i < nmordtbl; i++) {
        mord[i] = 0;
    }
    *lp = 0;
    *noprop = False;
    for( ; opt >= 0; opt = rdbrothr(opt)) {
        if(rdtype(opt) == RDID) {
            if(rdstrcmp(rdid(opt),"loop") == 0) {
                arg = rdarg(opt);
                if(rdtype(arg) != RDINT) {
                    fprintf(Stderr, "*err* rgndopt: illegal loop\n");
                } else {
                    *lp = (int)rdint(arg);
                }
            } else if(rdstrcmp(rdid(opt), "order") == 0) {
                /* do not redefine */
                arg = rdarg(opt);
                for(i = 0; i < nmordtbl; i++) {
                    if(rdtype(arg) != RDINT) {
                        fprintf(Stderr, "*err* rgndopt: illegal order\n");
                    } else {
                        mord[i] = (int)rdint(arg);
                    }
                    if(rdtype(arg) != RDEOF) {
                        arg = rdarg(arg);
                    }
                }
            } else if(rdstrcmp(rdid(opt),"noprop") == 0) {
                *noprop = True;
            }
        }
    }
}

/*==================================================================*/
Public  int     rgcomposit0(proc, grf)
tproc *proc;
tgraph   *grf;
/*--------------------------------------------------------------------
 * replace vertices with composit ones.
 */
{
    int  v, va, vn, vnn, n, lg, lgn, nva = 0;
    bool err;

    cprocess = proc;
    cgraph   = grf;

    for(v = 0; v < nvtable; v++) {
        if(vtable[v]->composit != V_COMPOSIT) {
            continue;
        }
        va = vtable[v]->auxv[0];
        vn = vnn = VNULL;
        for(n = 0; n < ndnodes(); n++) {
            if(va == ndvertex(n)) {
                vn = n;
                if(vn != VNULL) {
                    vnn = rgcnext0(v, vn, &lg, &lgn);
                }
                if(vnn != VNULL) {
                    rgccon0(v, vn, vnn, lg, lgn);
                    nva++;
                    n--;  /* n may be deleted by connection */
                }
            }
        }
    }
    err = False;
    for(n = 0; n < ndnodes(); n++) {
        if(ndtype(n) == NDEXTERN) {
            continue;
        }
        vn = ndvertex(n);
        if(vn <= 0) {
            continue;
        }
        if(vtable[vn]->composit == V_AUX) {
            fprintf(Stderr, "*** rgcomposit0: auxiliary vertices are left\n");
            fprintf(Stderr, "n = %d, vn = %d\n", n, vn);
            err = True;
        }
    }
    if(err) {
        rgdump(proc, grf);
        exit(FAIL);
    }
    return nva;
}

/*==================================================================*/
Private int rgcnext0(v, vn, lg, lgn)
int     v;
int     vn;
int    *lg;
int    *lgn;
/*--------------------------------------------------------------------
 * check whether node 'vn' can be replaced by composit vertex 'v'.
 * This function returns the other node 'vnn' which should be replaced
 * by 'v' as well as 'vn'.
 * '*lg' is the leg of 'vn' which connects with '*lgn' of 'vnn'.
 */
{
    int i, j, auxp, vnn, il, ax1;

    auxp = vtable[v]->auxp;
    ax1  = vtable[v]->auxv[1];

    for(i = 0; i < ndlegs(vn); i++) {
        if(abs(auxp) != abs(ndparticle(vn, i))) {
            continue;
        }
        vnn = ndnextn(vn, i);
        if(ndvertex(vnn) != ax1) {
            continue;
        }
        il = ndiline(vn, i);
        for(j = 0; j < ndlegs(vnn); j++) {
            if(il == ndiline(vnn, j)) {
                *lg  = i;
                *lgn = j;
                return vnn;
            }
        }
        fprintf(Stderr, "*** error : rgcnext : illegal connection\n");
        fprintf(Stderr, "v = %d, vn = %d, vnn = %d, lg = %d\n",
                v, vn, vnn, i);
        prdump();
        exit(FAIL);
    }
    return VNULL;
}

/*==================================================================*/
Private Void rgccon0(v, vn, vnn, lg, lgn)
int     v;
int     vn;
int     vnn;
int     lg;
int     lgn;
/*--------------------------------------------------------------------
 */
{
    int v0, v1, il, n, l, nl0, nl1, nln, lp;
    tleg *tmpl, *tl, *tl0, *tl1;

    cgraph->nilines = ilmake(True);
    if(vn < vnn) {
        v0 = vn;
        v1 = vnn;
    } else {
        v1 = vn;
        v0 = vnn;
    }
    il = ndiline(vn, lg);

    for(n = 0; n < ndnodes(); n++) {
        for(l = 0; l < ndlegs(n); l++) {
            if(ndiline(n, l) > il) {
                ndiline(n, l)--;
            } else if(ndiline(n, l) == il) {
                ndiline(n, l) = -1;
            }
        }
    }
    cgraph->nilines--;

    /* information of v0 and v1 */
    nl0 = ndlegs(v0);
    nl1 = ndlegs(v1);
    nln = nl0 + nl1 - 2;

    if((tmpl = (tleg*) malloc(sizeof(tleg)*nl1)) == NULL) {
        fprintf(Stderr, "*** rgccon : Memory\n");
        exit(FAIL);
    }

    for(l = 0; l < nl1; l++) {
        tmpl[l].lgnext  = ndnextn(v1, l);
        tmpl[l].lgpart  = ndparticle(v1, l);
        tmpl[l].lgiline = ndiline(v1, l);
    }

    /* move v1+1, ..., ndnode()-1 ==> v1, ... , ndnode()-2 */
    tl = tl1 = ndleg(v1, nl1 - 2);
    for(n = v1; n < ndnodes()-1; n++) {
        tl0 = tl;
        for(l = 0; l < ndlegs(n+1); l++) {
            tl->lgnext  = ndnextn(n+1, l);
            tl->lgpart  = ndparticle(n+1, l);
            tl->lgiline = ndiline(n+1, l);
            tl++;
        }
        ndbasel(n)  = ((char*)tl0) - ((char*)cgraph);
        ndtype(n)   = ndtype(n+1);
        ndlegs(n)   = ndlegs(n+1);
        ndloop(n)   = ndloop(n+1);
        ndltype(n)  = ndltype(n+1);
        ndgl4(n)    = ndgl4(n+1);
        ndvertex(n) = ndvertex(n+1);
    }

    /* move v0+1, ..., v1-1 ==> v0+1, ... , v1-1 */
    tl = tl1;
    for(n = v1-1; n > v0; n--) {
        for(l = ndlegs(n)-1; l >= 0; l--) {
            tl--;
            tl->lgnext  = ndnextn(n, l);
            tl->lgpart  = ndparticle(n, l);
            tl->lgiline = ndiline(n, l);
        }
        ndbasel(n) = ((char*)tl) - ((char*)cgraph);
    }

    /* redefine v0 */
    ndlegs(v0)   = nln;
    ndvertex(v0) = v;
    lp = 0;
    for(l = 0; l < nl0; l++) {
        if(l == lg) {
            continue;
        }
        ndparticle(v0, lp) = ndparticle(v0, l);
        ndnextn(v0, lp)    = ndnextn(v0, l);
        ndiline(v0, lp)    = ndiline(v0, l);
        lp++;
    }
    for(l = 0; l < nl1; l++) {
        if(l == lgn) {
            continue;
        }
        ndparticle(v0, lp) = tmpl[l].lgpart;
        ndnextn(v0, lp)    = tmpl[l].lgnext;
        ndiline(v0, lp)    = tmpl[l].lgiline;
        lp++;
    }
    if(lp != nln) {
        fprintf(Stderr, "*** rgcconn: illegal control\n");
        prdump();
    }
    ndnodes()--;

    if(vnn < cgraph->conjsrt) {
        cgraph->conjsrt--;
    }
    cgraph->nilines = ilmake(True);
}

/*==================================================================*/
Public  int     rgcomposit(proc, grf)
tproc *proc;
tgraph   *grf;
/*--------------------------------------------------------------------
 * replace vertices with composit ones.
 */
{
    int  v, vn, vnn, vnt, vnnt, lg, lgn, nva = 0, i, j;
    int  p = -1, il;
    bool err = False, rev;
    tproc   *sproc;
    tgraph  *sgrf;

    sproc = cprocess;
    sgrf  = cgraph;

    cprocess = proc;
    cgraph   = grf;

    for(vn = 0; vn < ndnodes(); vn++) {
        if(ndtype(vn) == NDEXTERN) {
            continue;
        }
        vnt  = ndvertex(vn);
        if(vnt <= 0) {
            continue;
        }
        if(vtable[vnt]->composit != V_AUX) {
            continue;
        }
        for(i = 0; i < ndlegs(vn); i++) {
            p = abs(ndparticle(vn, i));
            if(ptable[p]->ptype != P_AUX) {
                continue;
            }
            vnn = ndnextn(vn, i);
            il  = ndiline(vn, i);
            for(j = 0; j < ndlegs(vnn); j++) {
                if(il == ndiline(vnn, j)) {
                    lg  = i;
                    lgn = j;
                    vnnt = ndvertex(vnn);
                    goto found;
                }
            }
            fprintf(Stderr, "*** error : rgcomposit : illegal connection\n");
            fprintf(Stderr, "vn = %d, vnn = %d, lg = %d\n", vn, vnn, i);
            prdump();
            exit(FAIL);
        }
        fprintf(Stderr, "*** error : rgcomposit : aux=%d not found\n", p);
        fprintf(Stderr, "vn = %d\n", vn);
        prdump();
        exit(FAIL);

        found:;

        for(v = 0; v < nvtable; v++) {
            if(vtable[v]->composit != V_COMPOSIT) {
                continue;
            }
            if(vtable[v]->auxp != p) {
                continue;
            }
            if((vnt == vtable[v]->auxv[0] && vnnt == vtable[v]->auxv[1]) ||
               (vnt == vtable[v]->auxv[1] && vnnt == vtable[v]->auxv[0])) {
                rev = vnt == vtable[v]->auxv[1];
                rgccon(v, vn, vnn, lg, lgn, rev);
                nva++;
                vn--;  /* n may be deleted by connection */
                goto connected;
            }
        }
        if(vn <= vnn) {
            printf("*** There is no composit vertex which connects \n");
            printf("vertices %d and %d with aux %d\n\n", vnt, vnnt, p);
            pvndump(vnt);    printf("\n");
            pvndump(vnnt);   printf("\n");
            pvpdump(abs(p)); printf("\n");
        }

        err = False;

        connected:;
    }

    if(err) {
        exit(0);
    }
    cprocess = sproc;
    cgraph   = sgrf;
    return nva;
}

/*==================================================================*/
Private Void rgccon(v, vn, vnn, lg, lgn, rev)
int     v;
int     vn;
int     vnn;
int     lg;
int     lgn;
bool    rev;
/*--------------------------------------------------------------------
 *  replace (v0, v1) by connect(vnl, vnh)
 */
{
    int v0, v1, il, n, l, nl0, nl1, nln, lp, lg0, lg1;
    int vnl, vnh, lgl, lgh, nll, nlh;
    tleg *tmpl, *tl, *tl0, *tl1;

    cgraph->nilines = ilmake(True);

    /* v0 = lower address of (vn, vnn) */
    if(vn < vnn) {
        v0  = vn;
        v1  = vnn;
        lg0 = lg;
        lg1 = lgn;
    } else {
        v1 = vn;
        v0 = vnn;
        lg1 = lg;
        lg0 = lgn;
    }

    /* reversed order of vertices */
    if(rev) {
        vnl = vnn;
        vnh = vn;
        lgl = lgn;
        lgh = lg;
    } else {
        vnl = vn;
        vnh = vnn;
        lgl = lg;
        lgh = lgn;
    }

    /* delete iline */
    il = ndiline(v0, lg0);

    for(n = 0; n < ndnodes(); n++) {
        for(l = 0; l < ndlegs(n); l++) {
            if(ndiline(n, l) > il) {
                ndiline(n, l)--;
            } else if(ndiline(n, l) == il) {
                ndiline(n, l) = -1;
            }
        }
    }
    cgraph->nilines--;

    /* number of legs */
    nl0 = ndlegs(v0);
    nl1 = ndlegs(v1);

    nll = ndlegs(vnl);
    nlh = ndlegs(vnh);

    nln = nll + nlh - 2;

    /* construct image of the new vertex */
    if((tmpl = (tleg*) malloc(sizeof(tleg)*nln)) == NULL) {
        fprintf(Stderr, "*** rgccon : Memory\n");
        exit(FAIL);
    }

    lp = 0;
    for(l = 0; l < nll; l++) {
        if(l == lgl) {
            continue;
        }
        tmpl[lp].lgnext  = ndnextn(vnl, l);
        tmpl[lp].lgpart  = ndparticle(vnl, l);
        tmpl[lp].lgiline = ndiline(vnl, l);
        lp++;
    }
    for(l = 0; l < nlh; l++) {
        if(l == lgh) {
            continue;
        }
        tmpl[lp].lgnext  = ndnextn(vnh, l);
        tmpl[lp].lgpart  = ndparticle(vnh, l);
        tmpl[lp].lgiline = ndiline(vnh, l);
        lp++;
    }
    if(lp != nln) {
        fprintf(Stderr, "*** rgcconn: illegal control\n");
        prdump();
    }

    /* delete v1 */
    tl = tl1 = ndleg(v1, nl1 - 2);
    for(n = v1; n < ndnodes()-1; n++) {
        tl0 = tl;
        for(l = 0; l < ndlegs(n+1); l++) {
            tl->lgnext  = ndnextn(n+1, l);
            tl->lgpart  = ndparticle(n+1, l);
            tl->lgiline = ndiline(n+1, l);
            tl++;
        }
        ndbasel(n)  = ((char*)tl0) - ((char*)cgraph);
        ndtype(n)   = ndtype(n+1);
        ndlegs(n)   = ndlegs(n+1);
        ndloop(n)   = ndloop(n+1);
        ndltype(n)  = ndltype(n+1);
        ndgl4(n)    = ndgl4(n+1);
        ndvertex(n) = ndvertex(n+1);
    }

    /* move legs */
    tl = tl1;
    for(n = v1-1; n > v0; n--) {
        for(l = ndlegs(n)-1; l >= 0; l--) {
            tl--;
            tl->lgnext  = ndnextn(n, l);
            tl->lgpart  = ndparticle(n, l);
            tl->lgiline = ndiline(n, l);
        }
        ndbasel(n) = ((char*)tl) - ((char*)cgraph);
    }

    /* redefine v0 */
    ndlegs(v0)   = nln;
    ndvertex(v0) = v;
    for(l = 0; l < nln; l++) {
        ndparticle(v0, l) = tmpl[l].lgpart;
        ndnextn(v0, l)    = tmpl[l].lgnext;
        ndiline(v0, l)    = tmpl[l].lgiline;
    }

    /* decrease the number of nodes */
    ndnodes()--;

    if(v1 < cgraph->conjsrt) {
        cgraph->conjsrt--;
    }
    cgraph->nilines = ilmake(True);

    free(tmpl);
}

#if 0
/*==================================================================*/
Private Void rgccon1(v, vn, vnn, lg, lgn, rev)
int     v;
int     vn;
int     vnn;
int     lg;
int     lgn;
bool    rev;
/*--------------------------------------------------------------------
 */
{
    int v0, v1, il, n, l, nl0, nl1, nln, lp, lg0, lg1;
    tleg *tmpl, *tl, *tl0, *tl1;

    if(vn < vnn) {
        v0  = vn;
        v1  = vnn;
        lg0 = lg;
        lg1 = lgn;
    } else {
        v1 = vn;
        v0 = vnn;
        lg1 = lg;
        lg0 = lgn;
    }
    il = ndiline(v0, lg0);

    for(n = 0; n < ndnodes(); n++) {
        for(l = 0; l < ndlegs(n); l++) {
            if(ndiline(n, l) > il) {
                ndiline(n, l)--;
            } else if(ndiline(n, l) == il) {
                ndiline(n, l) = -1;
            }
        }
    }
    cgraph->nilines--;

    /* information of v0 and v1 */
    nl0 = ndlegs(v0);
    nl1 = ndlegs(v1);
    nln = nl0 + nl1 - 2;

    if((tmpl = (tleg*) malloc(sizeof(tleg)*nl1)) == NULL) {
        fprintf(Stderr, "*** rgccon : Memory\n");
        exit(FAIL);
    }

    for(l = 0; l < nl1; l++) {
        tmpl[l].lgnext  = ndnextn(v1, l);
        tmpl[l].lgpart  = ndparticle(v1, l);
        tmpl[l].lgiline = ndiline(v1, l);
    }

    /* move v1+1, ..., ndnode()-1 ==> v1, ... , ndnode()-2 */
    tl = tl1 = ndleg(v1, nl1 - 2);
    for(n = v1; n < ndnodes()-1; n++) {
        tl0 = tl;
        for(l = 0; l < ndlegs(n+1); l++) {
            tl->lgnext  = ndnextn(n+1, l);
            tl->lgpart  = ndparticle(n+1, l);
            tl->lgiline = ndiline(n+1, l);
            tl++;
        }
        ndbasel(n)  = ((char*)tl0) - ((char*)cgraph);
        ndtype(n)   = ndtype(n+1);
        ndlegs(n)   = ndlegs(n+1);
        ndloop(n)   = ndloop(n+1);
        ndltype(n)  = ndltype(n+1);
        ndgl4(n)    = ndgl4(n+1);
        ndvertex(n) = ndvertex(n+1);
    }

    /* move v0+1, ..., v1-1 ==> v0+1, ... , v1-1 */
    tl = tl1;
    for(n = v1-1; n > v0; n--) {
        for(l = ndlegs(n)-1; l >= 0; l--) {
            tl--;
            tl->lgnext  = ndnextn(n, l);
            tl->lgpart  = ndparticle(n, l);
            tl->lgiline = ndiline(n, l);
        }
        ndbasel(n) = ((char*)tl) - ((char*)cgraph);
    }

    /* redefine v0 */
    ndlegs(v0)   = nln;
    ndvertex(v0) = v;
    lp = 0;
    for(l = 0; l < nl0; l++) {
        if(l == lg0) {
            continue;
        }
        ndparticle(v0, lp) = ndparticle(v0, l);
        ndnextn(v0, lp)    = ndnextn(v0, l);
        ndiline(v0, lp)    = ndiline(v0, l);
        lp++;
    }
    for(l = 0; l < nl1; l++) {
        if(l == lg1) {
            continue;
        }
        ndparticle(v0, lp) = tmpl[l].lgpart;
        ndnextn(v0, lp)    = tmpl[l].lgnext;
        ndiline(v0, lp)    = tmpl[l].lgiline;
        lp++;
    }
    if(lp != nln) {
        fprintf(Stderr, "*** rgcconn: illegal control\n");
        prdump();
    }
    ndnodes()--;

    if(v1 < cgraph->conjsrt) {
        cgraph->conjsrt--;
    }
    cgraph->nilines = ilmake(True);
}
#endif

/*==================================================================*/
Public  Void    rgdump(proc, grf)
tproc *proc;
tgraph   *grf;
/*--------------------------------------------------------------------
 *  print the current graph
 */
{
    int n, p, io;

    printf("**** rgdump ****\n");

    printf("process : *(%p)", proc);
    if(proc == NULL) {
        printf(" NULL process\n");
        return;
    }
    printf(": id = %d\n",     proc->procid);
    printf("  graphs = %d\n", proc->graphs);
    printf("  nextern = %d",  proc->nextern);
    printf("  loop = %d",     proc->loop);
    printf("  order = %d\n",  proc->order);
    for(io = 0; io < nsordtbl; io++) {
        printf("  %s = %d,", sordtbl[io], proc->sorder[io]);
    }
    putchar(EOL);
    printf("  expand = %d",   proc->expand);
    printf("  onepi = %d",    proc->onepi);
    printf("  gm2 = %d\n",    proc->gm2);
    printf("  prefix = \"%s\"\n", proc->pprefix);
    printf("  kinem = \"%s\"\n",  proc->kinem);
    printf("  maxnodes = %d", proc->maxnodes);
    printf("  maxlegs = %d",  proc->maxlegs);
    printf("  msize = %d\n",  proc->msize);
    printf("  qcd = %d",      proc->qcd);
    printf("  extcol = %d\n", proc->extcol);

    printf("graph : *(%p)", grf);
    if(grf == NULL) {
        printf(" NULL graph\n");
        return;
    }

    cgraph   = grf;

    if(grf->subid == 0) {
        printf(": id=%ld\n",      grf->grid);
    } else {
        printf(": id=%ld(%ld)\n",  grf->grid, grf->subid);
    }
    printf("  nnodes = %d",  grf->nnodes);
    printf("  ngl4   = %d",  grf->ngl4);
    printf("  gbextp = %8.8ulx",  grf->gbextp);
    printf("  basen  = %d",   grf->basen);
    printf("  offsl  = %d\n", grf->offsl);
    printf("  gbcnsv = %8.8ulx",  grf->gbcnsv);
    printf("  gbelem = %8.8ulx\n",  grf->gbelem);

    for(n = 0; n < grf->nnodes; n++) {
        printf("  node = %d *(%p)", n, ndnode(0));
        printf("  type = %d",       ndtype(n));
        printf("  loop = %d",       ndloop(n));
        printf("  ltyp = %d",       ndltype(n));
        printf("  kind = %d",       ndvertex(n));
        printf("  gl4  = %d\n",     ndgl4(n));
        printf("  nlegs = %d",      ndlegs(n));
        printf("  legs = *(%p)\n",  ndleg(n, 0));
        if(ndleg(n, 0) == NULL) {
            break;
        }
        printf("  %d:", n);
        for(p = 0; p < ndlegs(n); p++) {
            printf(" i=%d[n=%d:p=%d]", ndiline(n, p), ndnextn(n, p),
                   ndparticle(n, p));
        }
        putchar(EOL);
    }

    printf("**** end of rgdump ****\n");
}

