/*********************************************************************
 *  File : token.c
 *
 *  Input module
 *
 * %%%%%%%%% description in tex format %%%%%%%%%%%
 * \begin{document}
 * \begin{enumerate}
 *
 * \item This package includes:
 *       \begin{enumerate}
 *       \item Get token elements.
 *       \item Analyze statements.
 *       \end{enumerate}
 *
 * \item Calling sequence of the module.\\
 *       Functions for this purpose are:
 * \begin{quote}{\tt
 * rdinit, rdend, rdsave, rdrestore
 * }\end{quote}
 *      One can switch input files by:
 * \begin{quote}\begin{verbatim}
 *         rdinit(file1, ...);
 *             (read file1)
 *         rdsave(sarea1);
 *         rdinit(file2);
 *             (read file2)
 *         rdend();
 *         rdrestore(sarea1);
 *             (read file1)
 *         rdend();
 * \end{quote}\end{verbatim}
 *    or
 * \begin{quote}\begin{verbatim}
 *         rdinit(file1, ...);
 *             (read file1)
 *         rdsave(sarea1);
 *         rdinit(file2);
 *             (read file2)
 *         rdsave(sarea2);
 *         rdrestore(sarea1);
 *             (read file1)
 *         rdsave(sarea1);
 *         rdrestore(sarea2);
 *             (read file2)
 *                 ...
 * \end{quote}\end{verbatim}
 *
 *   2. grammer for token elements:
 *
 *      2-1. Comment begins with character with flag "COMMENT".
 *           Characters between this character and EOL are neglected
 *           and characters in the next line are considered to continue
 *           just after the character, which is just before the comment
 *           character.
 *           Comment character is considered as normal one in character
 *           strings.
 *
 *      2-2. White spaces are characters with flag "SPACE", and are
 *           considered as spaces.
 *
 *      2-3. Identifier begins with a character flaged as "ALPHA".
 *           Rest characters are "ALPHA", "DIGIT" or "SPALPH".
 *
 *      2-4. Numbers here are rational number.  Their syntax is:
 *
 *                ["+" | "-"] unsigned_integer ["/"] unsigned_integer
 *
 *           Between each elements, additional white spaces
 *           can be inserted.
 *
 *      2-5. Strings are character strings sandwitched by the same
 *           character flaged "STRING".
 *
 *      2-6. Other charactes are considered as special characters and
 *            consist token elements by their-selves.
 *
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "common.h"
#include "dbg.h"
#include "token.h"
#include "escchar.h"

#define ADVANCED 1

/*--------------------------------------------------------------------
 * Private functions
 */
Private Void    rderrl          ARG((NoArg));
Private int     rdch0           ARG((NoArg));
Private Void    rdunch0         ARG((int));
Private int     rdch            ARG((bool, bool, bool));
Private int     rdnwsp          ARG((bool));
Private Void    rdskip          ARG((int));

Private int     rdstlist        ARG((NoArg));
Private int     rdstt           ARG((NoArg));
Private int     rdeidnuml       ARG((int));
Private int     rdeidnum        ARG((NoArg));
Private int     rdidnum         ARG((NoArg));
Private bool    rdnext          ARG((bool));
Private bool    rdcheck         ARG((int, bool));
Private bool    rdeof           ARG((NoArg));
Private bool    rdisidnum       ARG((NoArg));
Private Void    rdprin          ARG((int, int));
Private Void    rdprt           ARG((trdtoken));
Private Void    rddmp           ARG((trdtoken));

/*--------------------------------------------------------------------
 * Local variable
 */
Local   int     rdctype[256];   /* table of character type */
Local   char    rdesch[256];    /* escape char table*/
Local   int     rdhex[256];     /* value of hex char */
Local   int     rdcup;          /* = 'A' - 'a' */
Local   FILE    *rdin = (FILE*) NULL;   /* input file */
Local   FILE    *rdecho;        /* echo file */
Local   int     rdtol;          /* mode flag to convert case */

#define RDNOLEFT        -99
Local   char    rdlbuff[255];   /* line buffer for rderrl */
Local   int     rdcc = 0;       /* index in rdlbuff */
Local   int     rdln;           /* current line number in the file */
Local   int     rdc = RDNOLEFT;
        /* current character in statement analysis */

#ifdef ADVANCED
Local   bool    advanced = True;
#endif

/* type of characters */
#define SPACE   0x01
#define UPPER   0x02
#define LOWER   0x04
#define ALPHA   0x08
#define DIGIT   0x10
#define XDIGIT  0x20
#define STRING  0x40
#define ESCAPE  0x80
#define SPALPH  0x100
#define COMMENT 0x200
#define ALNUM   (ALPHA | DIGIT  | SPALPH)

/* test macro for character type */
#define rdupper(x)      (rdctype[CHAR(x)] & UPPER)
#define rdlower(x)      (rdctype[CHAR(x)] & LOWER)
#define rdalpha(x)      (rdctype[CHAR(x)] & ALPHA)
#define rdalnum(x)      (rdctype[CHAR(x)] & ALNUM)
#define rddigit(x)      (rdctype[CHAR(x)] & DIGIT)
#define rdspace(x)      (rdctype[CHAR(x)] & SPACE)
#define rdcmmnt(x)      (rdctype[CHAR(x)] & COMMENT)
#define rdesc(x)        (rdctype[CHAR(x)] & ESCAPE)
#define rdstr(x)        (rdctype[CHAR(x)] & STRING)

/* convert case */
#define rdtoup(x)       CHAR((rdlower(x)) ? (x) + rdcup : (x))
#define rdtolow(x)      CHAR((rdupper(x)) ? (x) - rdcup : (x))

/********************************************************************/
/*   Main program for test.
int main()
{
    int ind;

    rdinit(stdin, RDTOLOWER, NULL);
    while ((ind = rdstat()) >= 0)
        rdprint(ind);
}
*/

/*********************************************************************
1. Control of the module.

       rdinit, rdend, rdsave, rdrestore

   One can swithc files by:

        rdinit(file1, ...);
            (read file1)
        rdsave(sarea1);
        rdinit(file2);
            (read file2)
        rdend();
        rdrestore(sarea1);
            (read file1)
        rdend();

   or

        rdinit(file1, ...);
            (read file1)
        rdsave(sarea1);
        rdinit(file2);
            (read file2)
        rdsave(sarea2);
        rdrestore(sarea1);
            (read file1)
        rdsave(sarea1);
        rdrestore(sarea2);
            (read file2)
                ...

*/

/*==================================================================*/
Public  Void    rdinit(fp, tol, echo)
FILE    *fp;
int     tol;
FILE    *echo;
/*--------------------------------------------------------------------
   Initialization

   Arguments :
      fp  :  input file pointer.
      tol :  if      tol = RDTOLOWER then change to lower case.
             else if tol = RDTOUPPER then change to upper case.
             else                         not change.
      echo : if echo = NULL then no echo
             else                echo to the stream.

   Special characters.
      '-', '_' : alphabet.
      HT, EOL  : white space.
      '"'      : quote for string.
      '\'      : escape character.
                 escape sequence is defined as K & D 2nd ed., p193.
      '%'      : comment from this character to EOL.
                 Comment is ignored as if nothing is there.
                 So the next line is considered connected.
      '+', '-' : cannot use as the first character of ids.
--------------------------------------------------------------------*/
{
    int i;

    /* escape character (K&R 2nd ed., p193) */
    for(i = 0; i < 256; i++) {
        rdesch[i] = CHAR(i);
    }
    rdesch[CHAR('n')]  = CHR_NL;
    rdesch[CHAR('t')]  = CHR_HT;
    rdesch[CHAR('v')]  = CHR_VT;
    rdesch[CHAR('b')]  = CHR_BS;
    rdesch[CHAR('r')]  = CHR_CR;
    rdesch[CHAR('f')]  = CHR_FF;
    rdesch[CHAR('a')]  = CHR_BEL;
    /*          '0'    = octal number */
    /*          'x'    = hexiadecimal number */
    /*          'X'    = hexiadecimal number */

    /* value of hex character */
    for(i = 0; i < 256; i++) {
        rdhex[i] = 16;
    }
    for(i = '0'; i <= '9'; i++) {
        rdhex[i] = i - '0';
    }
    for(i = 'a'; i <= 'f'; i++) {
        rdhex[i] = i - 'a' + 10;
    }
    for(i = 'A'; i <= 'F'; i++) {
        rdhex[i] = i - 'A' + 10;
    }

    /* set character type */
    for(i = 0; i < 256; i++) {
        rdctype[i] = 0;
    }

    /* standard */
    for(i = 'a'; i <= 'i'; i++) {
        rdctype[i] = LOWER | ALPHA;
    }
    for(i = 'j'; i <= 'r'; i++) {
        rdctype[i] = LOWER | ALPHA;
    }
    for(i = 's'; i <= 'z'; i++) {
        rdctype[i] = LOWER | ALPHA;
    }

    for(i = 'A'; i <= 'I'; i++) {
        rdctype[i] = UPPER | ALPHA;
    }
    for(i = 'J'; i <= 'R'; i++) {
        rdctype[i] = UPPER | ALPHA;
    }
    for(i = 'S'; i <= 'Z'; i++) {
        rdctype[i] = UPPER | ALPHA;
    }

    for(i = '0'; i <= '9'; i++) {
        rdctype[i] = DIGIT | XDIGIT;
    }

    for(i = 'a'; i <= 'f'; i++) {
        rdctype[i] |= XDIGIT;
    }
    for(i = 'A'; i <= 'F'; i++) {
        rdctype[i] |= XDIGIT;
    }

    /* special treatment of characters */
    rdctype[CHAR('_')]  = SPALPH;
    rdctype[CHAR('-')]  = SPALPH;

    rdctype[CHAR(' ')]  = SPACE;
    rdctype[CHR_HT]     = SPACE;
    rdctype[CHAR(EOL)]  = SPACE;

    rdctype[CHAR('"')]  = STRING;

    rdctype[CHAR('%')]  = COMMENT;

    rdctype[CHAR('\\')] = ESCAPE;

    /* offset to change lower to upper case */
    rdcup  = 'A' - 'a';

    if(fp == (FILE *) NULL) {
        fprintf(Stderr, "rdinit : read file = NULL\n");
        exit(1);
    }

#ifdef ADVANCED
    /* default mode */
    advanced = True;
#endif
    rdln   = 1;
    rdin   = fp;
    rdecho = echo;
    rdtol  = tol;
    rdc    = RDNOLEFT;
}

#ifdef ADVANCED
/*==================================================================*/
Public  Void    rdsimple(NoArg)
/*--------------------------------------------------------------------
 * set simple integer mode.
 * 1. integers are always unsigned.
 * 2. no rational numbers;
 */
{    
    advanced = False;
}

/*==================================================================*/
Public  Void    rdadvanced(NoArg)
/*--------------------------------------------------------------------
 * set simple integer mode.
 * 1. integers are always unsigned.
 * 2. no rational numbers;
 */
{    
    advanced = True;
}
#endif

/*==================================================================*/
Public  Void    rdend(NoArg)
/*--------------------------------------------------------------------
   Finish to use file.
--------------------------------------------------------------------*/
{
    rdin = (FILE *) NULL;
}

/*==================================================================*/
Public  Void    rdsave(sarea)
trdsava *sarea;
/*--------------------------------------------------------------------
   Save current file.
--------------------------------------------------------------------*/
{
    if(rdin == (FILE *) NULL) {
        fprintf(Stderr, "rdsave : no file is used.\n");
        exit(16);
    }
    sarea->rdin   = rdin;
    sarea->rdecho = rdecho;
    sarea->rdtol  = rdtol;
    sarea->rdc    = rdc;
    sarea->rdln   = rdln;
    rdcc = 0;
    rdin = (FILE *) NULL;
}

/*==================================================================*/
Public  Void    rdrestore(sarea)
trdsava *sarea;
/*--------------------------------------------------------------------
   Save current file.
--------------------------------------------------------------------*/
{
    if(rdin != (FILE *) NULL) {
        fprintf(Stderr, "rdrestore : file is already used.\n");
        exit(16);
    }
    rdin   = sarea->rdin;
    rdecho = sarea->rdecho;
    rdtol  = sarea->rdtol;
    rdc    = sarea->rdc;
    rdln   = sarea->rdln;
    rdcc = 0;
}

/*********************************************************************
2. Basic private functions.

   rdch0, rdunch0, rdch, rderrl
*/
/*==================================================================*/
Private int     rdch0(NoArg)
/*--------------------------------------------------------------------
    get one character from a file.
    With echo when requested by "rdinit".
    Save line for "rderrl".
--------------------------------------------------------------------*/
{
    int c;

    /* get a char */
    if((c = getc(rdin)) == EOF) {
        return EOF;
    }

    if(rdecho != (FILE *) NULL) {
        putc(CHAR(c), rdecho);
    }

    if(rdcc < 255) {
        rdlbuff[rdcc++] = c;
    }

    if(c == EOL) {
        rdln++;
        rdcc = 0;
    }

    return c;
}

/*==================================================================*/
Private Void    rdunch0(c)
int     c;
/*--------------------------------------------------------------------
    unget one character from a file read by "rdch0".
--------------------------------------------------------------------*/
{
    ungetc(c, rdin);
    if(rdcc > 0) {
        rdcc--;
    } else if(c == '\n') {
        rdln--;
    }

    if(rdecho != (FILE *) NULL) {
        putc('\b', rdecho);
    }
}

/*==================================================================*/
Private int     rdch(conv, esc, comment)
bool    conv;
bool    esc;
bool    comment;
/*--------------------------------------------------------------------
    read one character from a file.

    Arguments:
        conv    : enable convert case of chars accoreding with
                 the request through "rdinit".
        esc     : enable escape character.
        comment : enable comments.
--------------------------------------------------------------------*/
{
    int c, s, d;

    /* get a char */
    if((c = rdch0()) == EOF) {
        return EOF;
    }

    /* skip comment line */
    if(comment) {
        while(rdcmmnt(c)) {

            /* skip until EOL */
            while((c = rdch0()) != EOL) {
                if(c == EOF) {
                    return EOF;
                }
            }

            /* next line */
            if((c = rdch0()) == EOF) {
                return EOF;
            }
        }
    }

    /* character conversion */
    if(conv) {
        if(rdtol == RDTOLOWER) {
            c = rdtolow(c);
        } else if(rdtol == RDTOUPPER) {
            c = rdtoup(c);
        }
    }

    /* escape character : convert P193, K&R 2nd ed. */
    if(esc && rdesc(c)) {

        if((c = rdch(False, False, False)) == EOF) {
            return EOF;

        } else if(c == '0') {     /* octal number */
            s = 0;
            while ((c = rdch(False, False, False)) == EOF &&
                   (d = rdhex[c]) < 8)
                s += 8*s + d;
            rdunch0(c);
            c = CHAR(s);

        } else if(tolower(c) == 'x') {   /* hexiadecial number */
            s = 0;
            while ((c = rdch(False, False, False)) == EOF &&
                   (d = rdhex[c]) < 16)
                s += 16*s + d;
            rdunch0(c);
            c = CHAR(s);

        } else
            c = rdesch[c];
    }

    return c;
}

/*==================================================================*/
Private Void    rderrl(NoArg)
/*--------------------------------------------------------------------
    Print current line.
    When just after calling "rdrestore", information about the current
    line may be lost and not printed.
--------------------------------------------------------------------*/
{
    int i;

    i = (rdcc > 255) ? 255 : rdcc;
    rdlbuff[i] = '\0';
    fprintf(Stderr, "%6d %s\n", rdln, rdlbuff);
}

/*********************************************************************
3. Read token elements.

   rdnwsp, rdtoken, rdskip
*/

#define rdnch() (rdch(True, False, True))

/*==================================================================*/
Private int     rdnwsp(next)
bool    next;
/*--------------------------------------------------------------------
    read to non-white-space character from a file.
    if(next) then begin from next character.
--------------------------------------------------------------------*/
{
    if(rdc == EOF) {
        return EOF;
    }

    if(rdc == RDNOLEFT || next) {
        rdc = rdnch();
    }

    while(rdspace(rdc)) {
        rdc = rdnch();
    }

    return rdc;
}

/*==================================================================*/
Public  int     rdtoken(tkn, str, nstr)
trdtoken *tkn;
string   str;
int      nstr;
/*--------------------------------------------------------------------
    returns the type of the next token element.
    White space is ignored.
    After a comment character, the line is ignored.

    RDEOF     ::= EOF
    RDID      ::= alpha [ alpha | number ]*
    RDINT     ::= ["+" | "-"] number+
    RDRNUM    ::= ["+" | "-"] number+ "/" number+
    RDSTRING  ::= quote char* quote
    RDSPECIAL ::= special character.
--------------------------------------------------------------------*/
{
    int cl, c, sgn, dn;

    cl = 0;
    if((rdc = rdnwsp(False)) == EOF) {
#if DBG
        if(dbg >= rddbg) {
            printf("rdtoken: 2 : EOF\n");
        }
#endif
        tkn->ttype = RDEOF;
        return 0;

    /* number */
#ifdef ADVANCED
    } else if(rddigit(rdc) || (advanced && rdc == '+') ||
                              (advanced && rdc == '-')) {
#else
    } else if(rddigit(rdc) || rdc == '+' || rdc == '-') {
#endif
        sgn = 1;
        if(rdc == '+') {
            rdc = rdnwsp(True);
        } else if(rdc == '-') {
            sgn = -1;
            rdc = rdnwsp(True);
        }

        if(rdc == EOF) {
            fprintf(Stderr, "rdtoken : ");
            fprintf(Stderr, "number is expected after +/-.\n");
            rderrl();
            return EOF;

        } else {
            tkn->dat.tnum[0] = 0;
            while(rdc != EOF && rddigit(rdc)) {
                tkn->dat.tnum[0] =  10*tkn->dat.tnum[0] + rdhex[rdc];
                rdc = rdnch();
            }
            tkn->dat.tnum[0] *= sgn;
            tkn->dat.tnum[1] = 1;
            tkn->ttype = RDINT;

#ifdef ADVANCED
            if(advanced && (rdc = rdnwsp(False)) == '/') {
#else
            if((rdc = rdnwsp(False)) == '/') {
#endif
                rdc = rdnwsp(True);
                tkn->dat.tnum[1] = 0;
                dn  = 0;
                while(rdc != EOF && rddigit(rdc)) {
                    tkn->dat.tnum[1] = 10*tkn->dat.tnum[1]
                                       + rdhex[rdc];
                    rdc = rdnch();
                    dn++;
                }
                if(dn == 0) {
                    fprintf(Stderr, "rdtoken : no denominator.  ");
                    fprintf(Stderr, "Taken to 1.\n");
                    rderrl();
                    tkn->dat.tnum[1] = 1;
                } else if(tkn->dat.tnum[1] == 0) {
                    fprintf(Stderr, "rdtoken : 0 denominator.  ");
                    fprintf(Stderr, "Taken to 1.\n");
                    rderrl();
                    tkn->dat.tnum[1] = 1;
                }
                tkn->ttype = RDRNUM;
            }
        }
        /* error for overflow ??? */

    /* string */
    } else if(rdstr(rdc)) {
        c = rdc;
        while((rdc = rdch(False, False, False)) != EOF &&
               rdc != EOL && rdc != c) {
           if(cl < nstr-1) {
               str[cl++] = rdc;
           }
        }
        str[cl++] = '\0';
        tkn->dat.tstr  = str;
        tkn->ttype = RDSTRING;
        rdc = RDNOLEFT;

    /* id ::= alpha (alnum)*. */
    } else if(rdalpha(rdc)) {
        str[cl++] = rdc;
        while((rdc = rdnch()) != EOF && rdalnum(rdc)) {
           if(cl < nstr-1) {
               str[cl++] = rdc;
           }
        }
        str[cl++] = '\0';
        tkn->dat.tstr  = str;
        tkn->ttype = RDID;

    /* special characters */
    } else {
        tkn->dat.tchar = rdc;
        tkn->ttype = RDSPECIAL;
        rdc = RDNOLEFT;
    }

#if DBG
    if(dbg >= rddbg) {
        printf("rdtoken: 3\n");
        printf("dbg: rdtype = %d\n",
               tkn->ttype);
    }
#endif

    return cl;
}

/*==================================================================*/
Private Void    rdskip(ch)
int     ch;
/*--------------------------------------------------------------------
    skip characters to the next character of "ch".
--------------------------------------------------------------------*/
{
    /* read a new character */
    if(rdc == RDNOLEFT) {
        rdc = rdnch();
    }

    /* skip characters */
    while(rdc != EOF && CHAR(rdc) != CHAR(ch)) {
        rdc = rdnch();
    }

    if(rdc == EOF) {
        return;

    /* to the next character of "ch" */
    } else {
        rdc = rdnch();
        return;
    }
}

/*********************************************************************
4. Read statements.

    rdstat
    rdstlist, rdstt, rdeidnum, rdidnum

*/
/*==================================================================*/
Public  int    rdstat(NoArg)
/*--------------------------------------------------------------------
    If error then returns RDERR
    else if eof then returns RDEOF
    else returns the index of rdtable.

    Syntax:

    statement   ::= stat ";".
    stat        ::= eidnum | eidnum "=" eidnumlist.
                           | eidnum "=" "{" eidnumlist "}".
    eidnumlist  ::= eidnum | eidnumlist ("," | " ") eidnum.
    eidnum      ::= idnum  | idnum "[" [stlist]  "]"
                           | idnum "(" [eidnumlist] ")".
    stlist      ::= stat   | stlist ";" stat.
    idnum       ::= id     | number.
--------------------------------------------------------------------*/
{
    int head;

    rdcurr  = 0;
    rdbuffp = 0;
    if(!rdnext(False)) {
        return RDEOF;
    }
    rdsln = rdln;
    head  = rdstt();
    if(!rdcheck(';', True)) {
        rderrl();
        return RDERR;
    }
#if DBG
    if(dbg >= rddbg) {
        printf("rdstat: end\n");
        printf("head=%4d, buffp=%4d,", head, rdbuffp);
        rdprint(head);
    }
#endif

    return head;
}

/*==================================================================*/
Private int     rdstlist(NoArg)
/*--------------------------------------------------------------------
    Syntax:

    stlist      ::= stat   | stlist ( ";" | "," ) stat.
--------------------------------------------------------------------*/
{
    int head, s;

    head = s = rdstt();
    while (s >= 0 && (rdcheck(';', False) || rdcheck(',', False))) {
        if(!rdnext(True)) {
            return RDERR;
        }
        rdtable[s].brother = rdstt();
        s = rdtable[s].brother;
    }
#if DBG
    if(dbg >= rddbg) {
        printf("rdstlist: end\n");
        printf("    head = %d\n", head);
    }
#endif
    return head;
}

/*==================================================================*/
Private int     rdstt(NoArg)
/*--------------------------------------------------------------------
    Syntax:

    stat        ::= eidnum | eidnum "=" eidnumlist.
                           | eidnum "=" "{" eidnumlist "}".
    eidnumlist  ::= eidnum | eidnumlist ("," | " ") eidnum.
--------------------------------------------------------------------*/
{
    int head, s;
    bool brace = False;

    if((head = rdeidnum()) < 0) {
        return head;
    }
    if(rdcheck('=', False)) {
        if(!rdnext(True)) {
            return RDERR;
        }
        if(rdcheck('{', False)) {
            brace = True;
            if(!rdnext(True)) {
                return RDERR;
            }
        }
/*
        s = head;
        while (s >= 0 && rdisidnum()) {
            rdtable[s].arg = rdeidnum();
            s = rdtable[s].arg;
            if(rdcheck(',', False)) {
                if(!rdnext(True)) {
                    return RDERR;
                }
            }
        }
*/
        s = rdeidnuml(head);
        if(s < 0) {
            return RDERR;
        }
        if(brace) {
            if(!rdcheck('}', True)) {
                rderrl();
                rdskip(';');
                head = RDERR;
            } else if(!rdnext(True)) {
                return RDERR;
            }
        }
    }
    return head;
}

/*==================================================================*/
Private int     rdeidnuml(head)
int     head;
/*--------------------------------------------------------------------
    Syntax:

    eidnumlist  ::= eidnum | eidnumlist ("," | " ") eidnum.
--------------------------------------------------------------------*/
{
    int s;

    s = head;
    while (s >= 0 && rdisidnum()) {
        rdtable[s].arg = rdeidnum();
        s = rdtable[s].arg;
        if(rdcheck(',', False)) {
            if(!rdnext(True)) {
                return RDERR;
            }
        }
    }
    if(s < 0) {
        return RDERR;
    }

    return head;
}

/*==================================================================*/
Private int     rdeidnum(NoArg)
/*--------------------------------------------------------------------
    Syntax:

    eidnum      ::= idnum  | idnum "[" [stlist]  "]"
                           | idnum "(" [eidnumlist] ")".
--------------------------------------------------------------------*/
{
    int head;

    if((head = rdidnum()) >= 0) {
        if(rdcheck('[', False)) {
            if(!rdnext(True)) {
                return RDERR;
            }
            if((rdtable[head].opt = rdstlist()) < 0) {
                return False;
            }
            if(!rdcheck(']', True)) {
                rderrl();
                rdskip(';');
                head = RDERR;
            } else if(!rdnext(True)) {
                return RDERR;
            }
        } else if(rdcheck('(', False)) {
            if(!rdnext(True)) {
                return RDERR;
            }
            if(rdeidnuml(head) < 0) {
                return False;
            }
            if(!rdcheck(')', True)) {
                rderrl();
                rdskip(';');
                head = RDERR;
            } else if(!rdnext(True)) {
                return RDERR;
            }
        }
    }
    return head;
}

/*==================================================================*/
Private int     rdidnum(NoArg)
/*--------------------------------------------------------------------
    Syntax:

    idnum       ::= id     | number.
--------------------------------------------------------------------*/
{
    if(!rdisidnum()) {
        fprintf(Stderr, "rdidnum: id or number is expected.\n");
        rderrl();
        rdskip(';');
        return RDERR;
    } else {
        rdcurr++;
        if(rdcurr >= LRDTABLE) {
            fprintf(Stderr, "rdidnum : table overflow\n");
            exit(1);
        }
        if(!rdnext(True)) {
            return RDERR;
        }
    }
    return (rdcurr-1);
}

/*********************************************************************
5. Tools for rdstatt.

   rdnext, rdcheck, rdeof, rdisidnum
*/

/*==================================================================*/
Private bool    rdnext(chk)
bool    chk;
/*--------------------------------------------------------------------
    read next token.
    It returns !EOF.
    If chk and EOF then print message.
    The result is given by "rdtable[rdcurr]".
    In order to keep this token element, set "rdcurr++".
--------------------------------------------------------------------*/
{
    int len, l;

    len = LRDBUFF - rdbuffp;
    if(len < MAXLID) {
        fprintf(Stderr, "rdnext : string buffer full.\n");
        exit(1);
    }
    l = rdtoken(&(rdtable[rdcurr]), rdbuff+rdbuffp, len);
    rdtable[rdcurr].brother = RDNULL;
    rdtable[rdcurr].opt     = RDNULL;
    rdtable[rdcurr].arg     = RDNULL;
    rdbuffp += l;

    if(rdeof()) {
        if(chk) {
            fprintf(Stderr, "rdnext : Unexpected EOF.\n");
        }
        return False;
    }

#if DBG
    if(dbg >= rddbg) {
        printf("rdnext: end\n");
        rddmp(rdtable[rdcurr]);
    }
#endif
    return True;
}

/*==================================================================*/
Private bool   rdcheck(c, chk)
int  c;
bool chk;
/*--------------------------------------------------------------------
    Check current token element is special char and value is "c".
    If chk then print message when failed.
--------------------------------------------------------------------*/
{
    bool st;

    st = rdtable[rdcurr].ttype == RDSPECIAL &&
         rdtable[rdcurr].dat.tchar == c;
    if(chk && !st) {
        fprintf(Stderr, "rdcheck : illegal char.\n");
        fprintf(Stderr,"    %c is expected.\n", c);
        rderrl();
    }
    return st;
}

/*==================================================================*/
Private bool   rdeof(NoArg)
/*--------------------------------------------------------------------
    Check current token element is eof.
--------------------------------------------------------------------*/
{
    return (rdtable[rdcurr].ttype == RDEOF);
}

/*==================================================================*/
Private bool   rdisidnum(NoArg)
/*--------------------------------------------------------------------
    Check current token element is id or number.
--------------------------------------------------------------------*/
{
    return (rdtable[rdcurr].ttype == RDID ||
            rdtable[rdcurr].ttype == RDSTRING ||
            rdtable[rdcurr].ttype == RDINT ||
            rdtable[rdcurr].ttype == RDRNUM);
}

/*********************************************************************
6. Debugging tools (dump statement)

   rdpstat
*/

/*==================================================================*/
Public  Void    rdpstat(ind, lvl)
int     ind;
int     lvl;
/*--------------------------------------------------------------------
    Dump statement "ind".
--------------------------------------------------------------------*/
{
    int s, b, b0;

    printf("%*s>>>> : \n", lvl," ");
    b0 = -1;
    for(b = ind; b > b0; b = rdtable[b].brother) {
        b0 = b;
        printf("%*s>%d ",lvl," ", b);
        rddmp(rdtable[b]);

        if((s = rdtable[b].opt) > b) {
            printf("%*s>opt : \n",lvl," ");
                rdpstat(s, lvl+1);
            printf("%*s<opt : \n",lvl," ");
        }

        if((s = rdtable[b].arg) > b) {
            printf("%*s>arg : \n",lvl," ");
                rdpstat(s, lvl+1);
            printf("%*s<arg : \n",lvl," ");
        }
    }
    printf("%*s<<<< : \n", lvl," ");
}

/*==================================================================*/
Public  Void    rddump(itkn)
int     itkn;
/*--------------------------------------------------------------------
    Dump the token element.
--------------------------------------------------------------------*/
{
    rddmp(rdtable[itkn]);
}

/*==================================================================*/
Private Void    rddmp(tkn)
trdtoken tkn;
/*--------------------------------------------------------------------
    Dump the token element.
--------------------------------------------------------------------*/
{
    printf(" token type    = %d", tkn.ttype);
    printf(" brother = %d", tkn.brother);
    printf(" opt = %d ", tkn.opt);
    printf(" arg = %d \n", tkn.arg);
    switch (tkn.ttype) {
    case RDID      : printf("id = %s\n", tkn.dat.tstr);
                     break;
    case RDINT     : printf("int = %ld\n", tkn.dat.tnum[0]);
                     break;
    case RDRNUM    : printf("num = %ld/%ld\n",
                            tkn.dat.tnum[0], tkn.dat.tnum[1]);
                     break;
    case RDSTRING  : printf("string = \"%s\"\n", tkn.dat.tstr);
                     break;
    case RDSPECIAL : printf("special = /%c/\n", tkn.dat.tchar);
                     break;
    case RDEOF     : printf("EOF\n");
                     break;
    }
}

/*********************************************************************
7. Print statements.

   rdprint, rdprin, rdprt
*/

/*==================================================================*/
Public  Void    rdprint(ind)
int     ind;
/*--------------------------------------------------------------------
    Print statement "ind".
--------------------------------------------------------------------*/
{
    if(ind == RDEOF) {
        printf("   EOF\n");
        return;
    } else if(ind == RDERR) {
        printf("   ERR\n");
        return;
    } else if(ind < 0) {
        printf("   ???\n");
        return;
    }
    printf("%6d ", rdsln);
    rdprin(ind, 0);
    putchar(';');
    putchar(EOL);
}

/*==================================================================*/
Private Void    rdprin(ind, argn)
int     ind;
int     argn;
/*--------------------------------------------------------------------
    Print the return value of rdstat
--------------------------------------------------------------------*/
{
    int s, b, b0;
    bool br = False;

    b0 = -1;
    for(b = ind; b > b0; b = rdtable[b].brother) {
        b0 = b;
        if(b != ind && argn == 0)
            putchar(';');
        rdprt(rdtable[b]);

        if((s = rdtable[b].opt) > b) {
            putchar('[');
            rdprin(s, 0);
            putchar(']');
        }

        if((s = rdtable[b].arg) > b) {
            br = argn == 0 && rdtable[s].arg > s;
            if(argn == 0) {
                putchar('=');
            } else {
                putchar(',');
            }
            if(br) {
                putchar('{');
            }
            rdprin(s, argn+1);
            if(br) {
                putchar('}');
            }
        }
    }
}

/*==================================================================*/
Private Void    rdprt(tkn)
trdtoken tkn;
/*--------------------------------------------------------------------
    Print the token element.
--------------------------------------------------------------------*/
{
    switch (tkn.ttype) {
    case RDID      : printf("%s", tkn.dat.tstr);
                     break;
    case RDINT     : printf("%ld", tkn.dat.tnum[0]);
                     break;
    case RDRNUM    : printf("%ld/%ld", tkn.dat.tnum[0],
                            tkn.dat.tnum[1]);
                     break;
    case RDSTRING  : printf("\"%s\"", tkn.dat.tstr);
                     break;
    case RDSPECIAL : putchar(tkn.dat.tchar);
                     break;
    case RDEOF     : printf("EOF");
                     break;
    }
}

/*********************************************************************
8. Utilities for statements

    rdbool, rdintarg, rdchckid, rdkval, rdskipid
*/
/*==================================================================*/
Public  bool    rdbool(ind, value)
int     ind;
int     *value;
/*--------------------------------------------------------------------
    read yes or no
--------------------------------------------------------------------*/
{
    if(rdtype(ind) == RDID) {
        if(rdstrcmp(rdid(ind), "yes") == 0) {
            *value = True;
            return True;
        } else if(rdstrcmp(rdid(ind), "no") == 0) {
            *value = False;
            return True;
        } else {
            return False;
        }
    } else {
        return False;
    }
}

/*==================================================================*/
Public  bool    rdenum(ind, value, nex, exstr, exval)
int     ind;
int     *value;
int     nex;
string  *exstr;
int     *exval;
/*--------------------------------------------------------------------
 *  read enumerate
 *     if the value == exstr[k] ==> exval[k]   (0 <= k < nex)
 *------------------------------------------------------------------*/
{
    int i;

    if(rdtype(ind) == RDID) {
        for(i = 0; i < nex; i++) {
            if(strcmp(rdid(ind), exstr[i]) == 0) {
                *value = exval[i];
                return True;
            }
        }
        return False;
    } else {
        return False;
    }
}

/*==================================================================*/
Public  bool    rdintarg(ind, value)
int     ind;
int     *value;
/*--------------------------------------------------------------------
    read yes or no
--------------------------------------------------------------------*/
{
        if(rdtype(ind) == RDINT) {
            *value = rdint(ind);
            return True;
        } else
            return False;
}

/*==================================================================*/
Public  bool    rdchkid(ind, str)
int     ind;
string  str;
/*--------------------------------------------------------------------
    check token element is str.
--------------------------------------------------------------------*/
{
    if(rdtype(ind) == RDID && rdstrcmp(rdid(ind), str) == 0) {
            return True;
    } else {
            return False;
    }
}

/*==================================================================*/
Public  bool    rdkval(ind, str, value)
int     ind;
string  str;
long    *value;
/*--------------------------------------------------------------------
    get token element str=integer_value.
--------------------------------------------------------------------*/
{
    int arg;

    if(rdtype(ind) == RDID && rdstrcmp(rdid(ind), str) == 0) {
            arg = rdarg(ind);
            if(rdtype(arg) == RDINT) {
                *value = rdint(arg);
                return True;
            } else
                return False;
    } else
            return False;
}

/*==================================================================*/
Public  Void    rdskpid(str)
string  str;
/*--------------------------------------------------------------------
    read token element to find str.
--------------------------------------------------------------------*/
{
    int id;

    while ((id = rdstat()) >=0 && !rdchkid(id, str))
        ;
}

/*==================================================================*/
Public  int     rdstrcmp(str1, str2)
string  str1, str2;
/*--------------------------------------------------------------------
    compare with ignoring case of characters
--------------------------------------------------------------------*/
{
    int ans;

    while (1) {
        if(*str1 == '\0') {
            return *str2;
        }
        if(*str2 == '\0') {
            return *str1;
        }
        ans = tolower(*str1++) - tolower(*str2++);
        if(ans != 0) {
            return ans;
        }
    }
}
