/*********************************************************************
 * $id$
 *
 *  File : assign.c (prefix = pa)
 *
 *  Assign particle to internal lines.
 *  Looped vertices are ignored for the assignment.
 *  Duplication of graphs in terms of automorphism groups of graphs
 *  is not removed.
 *
 *  Control of the program :
 *
 *     passign --> paselect --> pasetv --> paset1v --> pasetct  
 *             --> patest
 *             --> presult (exit)
 *
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include "grc.h"

/*------------------------------------------------------------------*/
/* Local functions */
Private Void    paend      ARG((NoArg));

Private Void    paselect   ARG((node,node,pindex,int));
Private node    paseln     ARG((NoArg));

Private bool    padef      ARG((node,int,pindex,int,bool));
Private bool    padfnext   ARG((node,int,pindex,int));
Private Void    parestore  ARG((node,int));
Private Void    pareset    ARG((NoArg));

Private int     paleg      ARG((node,int,node,bool));

Private Void    pasetv     ARG((NoArg));
Private Void    paset1v    ARG((node));
Private Void    pasetct    ARG((NoArg));
Private int     panasgn    ARG((vindex));

Private string  pastype    ARG((int));

/* Local variables */
Local   int    inpa = 0;   /* flag of dump assigned particle */
Local   int    mord[MAXMORD];
Local   int    mordr[MAXMORD];

/*==================================================================*/
Public  Void    passign(NoArg)
/*--------------------------------------------------------------------
 *  Assign particle to internal lines and determine vertex.
 */
{
#if DBG
    if(dbg >= padbg) {
        printf("passign: begin\n");
        prdump();
    }
#endif

    erenter("passign");

    if(painit()) {              /* initialization */

        inpa = 1;

        /* main body of particle assignment */
        paselect(NDNULL, NDNULL, PNULL, 1);

    } 

    paend();

#if DBG
    if(dbg >= padbg) {
        printf("passign: end\n");
        prdump();
    }
#endif

    erreturn("passign");
    inpa = 0;
}

/*==================================================================*/
Public  bool    painit(NoArg)
/*--------------------------------------------------------------------
 *  Initialization
 *
 *  1. clear ndasave, ndnextn, ndundef, ndparticle
 *  2. particle assignment of e-lines
 */
{
    node n;
    int  l;

#if DBG
    erdbg("painit", "begin", padbg);
#endif

    alinit();           /* initialize list */

    /* initialize ndasave, ndnextn, ndundef and ndparticle
     * storage for ndasave, pdnextn, pdparticle is allocated
     * at elem.c
     */
    for(n = 0; n < nnodes; n++) {
        ndasave(n) = ALNULL;
    }

    /* clear particles of i-lines 
     * for external particles, ndundef and ndparticle
     *     are already set.
     */
    for(n = 0; n < nnodes; n++) {
        if(ndtype(n) != NDEXTERN) {
            ndvertex(n) = VNULL;
            ndundef(n) = ndlegs(n);
            for(l = 0; l < ndlegs(n); l++) {
                ndparticle(n, l) = PNULL;
            }
        }
    }

    passg = True;

#if DBG
    if(dbg >= padbg) {
        prdump();
    }
#endif

    /* ndname : particle kind of e-line */
    for(n = 0; n < nnodes; n++) {
        if(ndtype(n) != NDEXTERN) {
            continue;
        }

        /* define particle for the i-line adjacent to the e-line.
         * Their particle is assigned in "elem.c/elgene" called
         * when the process is read.
         */
        if(!padef(n, 0, ndname(n), 0, False)) {
#if DBG
            erdbg("painit", "end False", padbg);
#endif
            return False;
        }
    }

#if DBG
    erdbg("painit", "end True", padbg);
#endif

    return True;
}

/*==================================================================*/
Private Void    paend(NoArg)
/*--------------------------------------------------------------------
 *  Reset the configuration
 */
{
    node n;

    pareset();
    for(n = 0; n < nnodes; n++) {
        ndundef(n) = 0;
    }

    passg = False;
}

/*==================================================================*/
Private Void    paselect(nd, tl, pl, level)
node nd;
node tl;
pindex pl;
int  level;
/*--------------------------------------------------------------------
 *  Assign particle to a leg of a node.
 *
 *  Loops of
 *          for nd = node {
 *              for lg = leg {
 *              }
 *          }
 *  is realized by recursive calls.
 *
 * Arguments:
 *    nd     : the node assigned last
 *    tl     : the target node at the last time
 *    pl     : the particle at the last time
 *    level  : used to save and restore 
 */
{
    pindex prtcl[LPTABLE];
    int    p, pm;
    pindex pn;
    vindex tn;
    int    lg;
    bool   same;

#if DBG
    erdbg("paselect", "begin", padbg);
#endif

    /* if (this is the first call) or ("nd" is already assigned) */
    if(nd == NDNULL || ndundef(nd) < 1) { 
        nd = paseln();  /* next node */
        same = False;
    } else {           /* the same node as in the last assignment */
        same = True;
    }

    /* assign legs which are uniquely determined and
     * check end of assignment.
     */
    if(nd == NDNULL) {

        /* assign vertices */
        pasetv();
        return;
    }

    /* assign the next node */

#if DBG
    if(dbg >= padbg) {
        printf("*dbg* paselect : nd = %d\n", nd);
        padump();
    }
#endif

    /* construct "prtcl" : the set of possible particles 
     * for assigning to one of legs of the node "nd"
     */
    if((pm = mvcandp(nd, prtcl)) < 1) {

#if DBG
        if(dbg >= padbg) {
            printf("*dbg* paselect : return by mvcandp pm=%d\n", pm);
            padump();
        }
#endif
        return;
    }

#if DBG
    if(dbg >= padbg) {
         printf("*dbg* paselect pm=%d, prtcl=(",pm);
         for(p = 0; p < pm; p++) {
             printf(" %d",prtcl[p]);
         }
         putchar(EOL);
    }
#endif

    /* find a leg which is not yet assigned */
    lg = panasgn(nd);
    tn = ndnextn(nd, lg);  /* adjacent node of "lg" */

    /* for all possible particles */
    for(p = 0; p < pm; p++) {
        pn = prtcl[p];

        /* self-loop : do not assign anti-particle */
        if(tn == nd && pn < 0) {
            continue;
        }

        /* avoid duplicated assignment for multiple connection 
         * The following condition uses the convention that the legs
         * of a multiple connection are neighboring ones.
         */
        if(same && tn == tl && pn > pl) {
            continue;
        }

        /* define particle for the i-line : 
         *     nodes nd and tn are modified 
         */
        if(padef(nd, lg, pn, level, True)) { /* def and save */
            paselect(nd, tn, pn, level+1);  /* next assignment */
        }

        /* reset the graph before the last assignment */
        parestore(nd, level);
        parestore(tn, level);
    }
}

/*==================================================================*/
Private Void    patest(NoArg)
/*--------------------------------------------------------------------
 * The last part of particle assignment.
 * This function checks duplication of graphs and calculate
 * symmetry factor.
 */
{
    long   nsym;

    if(hep) {
        hetest();

    /* discard duplicated graphs : nsym > 0 : the symmetry factor */
    } else if((nsym = lcatest()) > 0) {
        if(expand) {
#if DBG
            if(dbg >= padbg) {
                printf("patest: check f-line\n");
            }
#endif

            /* symmetry factor */
            cgraph->sfactor = nsym;

            /* output the resulting graph */
            presult();
        } else {
            cgraph->sfactor = 1;
            presult();
        }
#if DBG
    } else {
        if(dbg >= padbg) {
            printf("patest: del by lcatest\n");
        }
#endif
    }
    lcreset();
}

/*==================================================================*/
Private node    paseln(NoArg)
/*--------------------------------------------------------------------
 *  Select node to be assigned.
 *  If there is not such a node, returns NDNULL.
 *  Selected node is a tree vertex which has minimum number of
 *  undefined legs.
 *
 *  The selected node has the minimum number of undefined legs.
 *
 *  Returns:
 *      node number : when succeeded.
 *      NDNULL      : every nodes are determined
 */
{
    node   n, v;
    int    lg;

    n   = NDNULL;       /* selected node */
    lg  = MAXTL;        /* minimum number of undefined legs */

    for(v = 0; v < nnodes; v++) {

        /* assign only tree vertex and counter terms
         * self-energy part is considered as a kind of counter term.
         */
        if(ndtype(v) != NDTREEV) {
            if(countert) {
                if(ndtype(v) != NDLOOPEDV) {
                    continue;
                }
            } else {
                continue;
            }
        }

        /* find minimum */
        if(ndundef(v) > 0 && ndundef(v) < ndlegs(v) 
           && ndundef(v) < lg) {
            lg = ndundef(v);
            n  = v;
        }
        if(lg == 1) {
            return n;
        }
    }
    return n;
}

/*==================================================================*/
Private Void    pasetv(NoArg)
/*--------------------------------------------------------------------
 *  Assign vertices to the nodes.
 */
{
    int io;

    for(io = 0; io < nmordtbl; io++) {
        mord[io] = 0;
    }

    paset1v(0);
}

/*==================================================================*/
Private Void    paset1v(n)
node    n;
/*--------------------------------------------------------------------
 *  Assign vertices to the nodes : main part.
 */
{
    int lnl, l, v, io, na;
    bool cont, ac;
#ifdef TURBOC
    /* stack is too small */
    node *nl;

    if((nl = (node*) malloc(LVTABLE*sizeof(node))) == NULL) {
        fprintf(Stderr, "paset1v: memory allocation failed.\n");
        exit(1);
    }
#else
    node nl[LVTABLE];
#endif

#if DBG
    if(dbg >= padbg) {
        printf("paset1v:begin:n=%d->", n);
    }
#endif
    /* fined not assigned node */
    while (n < nnodes 
          && (ndtype(n) == NDEXTERN || ndvertex(n) != VNULL
             || (ndtype(n) == NDLOOPEDV && !countert)) ) {
        n++;
    }
#if DBG
    if(dbg >= padbg) {
        printf("%d/%d\n", n, nnodes);
    }
#endif

    /* no more nodes */
    if(n >= nnodes) {
        pasetct();
#ifdef TURBOC
        free(nl);
#endif
#if DBG
        if(dbg >= padbg) {
            printf("paset1v:end:1:n=%d", n);
        }
#endif
        return;
    }

#if DBG
    if(ndundef(n) > 0) {
        fprintf(Stderr, "paset1v: Illegal vertex assignment\n");
        prdump();
        exit(1);
    }
#endif

    /* find candidates */
    lnl = mvmlist(n, nl);
#if DBG
    if(dbg >= padbg) {
        int i;
        printf("paset1v:n=%d,lnl=%d\n", n, lnl);
        for(i = 0; i < lnl; i++) {
            printf("paset1v:candidate[%d] = %d\n", i, nl[i]);
        }
    }
#endif

    na = 0;
    ac = False;
    for(l = 0; l < lnl; l++) {
        v = nl[l];
        cont = True;
        for(io = 0; io < nmordtbl; io++) {
            mord[io] += vtable[v]->morder[io];
        }
        cont = (oscheck(sorder, mord, False) >= 0);
        if(cont) {

            /* found in the vertex table */
            if(ndloop(n) == vtable[v]->loop) {
                ndvertex(n) = v;
                na++;
#if DBG
                if(dbg >= padbg) {
                    printf("paset1v:n=%d is assigned to %d\n", n, v);
                }
#endif
                paset1v(n+1);        /* next node */
                ndvertex(n) = VNULL;

            /* counter term */
            } else if(anyct) {
                ac = True;
            }
        }
        for(io = 0; io < nmordtbl; io++) {
            mord[io] -= vtable[v]->morder[io];
        }
    }

    /* counter term of higher order of tree vertex :
     * go to next node without assigning vertex to node n
     */
    if(ac && na == 0 && countert && ndloop(n) > 0) {
        paset1v(n+1);
    }
#ifdef TURBOC
    free(nl);
#endif
#if DBG
    if(dbg >= padbg) {
        printf("paset1v:end:2:n=%d", n);
    }
#endif
}

/*==================================================================*/
Private Void    pasetct(NoArg)
/*--------------------------------------------------------------------
 *  Assign order to the counter terms
 */
{
    if(shrinkaux && !axconn()) {
        return;
    }
    if(hep) {
        if(osleft(mordr, sorder, mord) == 0) {
            hesetct(mordr);
        }
    } else {
        if(oscheck(sorder, mord, True) == 0) {
            patest();
        } else {
        }
    }
}

/*==================================================================*/
Private bool    padef(nd, lg, p, level, save)
node    nd;
int     lg;
pindex  p;
int     level;
bool    save;
/*--------------------------------------------------------------------
 * Assign ("nd", "lg") and its opposite leg as particle "p".
 * The old status is saved with information "level".
 * When the node is an external line, assignment of ("nd", "lg")
 * is skipped.
 * This is done by setting flag "save" false.
 *
 * Arguments :
 *     nd    : node
 *     lg    : leg
 *     p     : kind of the particle
 *     level : level of save
 *     save  : assign particle ("nd", "lg") or not.
 *
 * Returns :
 *     succeeded or not.
 */
{
    pindex po;
    node   no;
    int    lo;

    /* opposite side of the i-line */
    no = ndnextn(nd, lg);
    po = ptable[abs(p)]->antip * p;

#if DBG
    if(dbg >= padbg) {
        printf("*dbg* padef : ptable[%d]->antip = %d, po = %d\n",
               p, ptable[abs(p)]->antip, po);
    }
#endif

    /* the leg connecting from "no" to "nd" */
    if((lo = paleg(nd, lg, no, True)) < 0) {  
        error("padef", "inconsistent connection");
        prdump();
        fprintf(Stderr, "padef(%d, %d, %d, %d, %d)\n",
                nd, lg, p, level, save);
        erfail();
    } 

#if DBG
    if(dbg >= padbg) {
        printf("*dbg* padef: level = %d:", level);
        printf(" no: %d(%d, %d) -> nd: %d(%d, %d)\n",
                 no, lo, po, nd, lg, p);
    }
#endif

    /* assign to the leg ("nd", "lg") as "p" */
    if(save) {
        if(!padfnext(nd, lg, p, level)) {
            return False;
        }
    }

    return padfnext(no, lo, po, level);
}

/*==================================================================*/
Private bool    padfnext(nd, ld, pl, level)
node    nd;
int     ld;
pindex  pl;
int     level;
/*--------------------------------------------------------------------
 * Assign the particle of the leg ("nd", "ld") as particle "pl".
 * Leg of neighboring node is not determined.
 *
 * Arguments :
 *     nd;   : node
 *     ld;   : leg of "nd"
 *     pl    : kind of particle to be assigned
 *     level : level of saving
 */
{
#if DBG
    if(dbg >= padbg) {
        printf("*dbg* padfnext : (%d, %d, %d, %d)\n",nd,ld,pl,level);
    }

    if(pl == PNULL) {  /* null particle */
        printf("*** assign null particle\n");
        printf("padfnext(%d, %d, %d, %d)\n",nd,ld,pl,level);
        prdump();
        printf("*** assign null particle\n");
        erfail();
    }

    if(ndparticle(nd, ld) != PNULL) {
        printf("*err* padfnext : already assigned\n");
        printf("ndparticle(%d,%d) = %d <-- %d\n", 
               nd, ld, ndparticle(nd, ld), pl);
        erfail();
    }
#endif

    /* set particle to the leg */
    ndparticle(nd, ld) = pl;

    /* save old status in a new cell of the list of the node */
    ndasave(nd)    = alcons(level, ld, ndasave(nd));
    ndundef(nd)--;

    /* skip check in the case of looped vertex */
    if(ndtype(nd) != NDTREEV) {
        if(countert) {
            if(ndtype(nd) != NDLOOPEDV) {
                return True;
            }
        } else {
            return True;
        }
    }

    /* check whether this assignment is consistent one or not */
    if(mvable(nd) > 0) {
        return True;
    }

    return False;
}

/*==================================================================*/
Private Void    parestore(n, level)
node    n;
int     level;
/*--------------------------------------------------------------------
 * Restore old status of the node "n" up to "level".
 *
 * Arguments :
 *     n     : node.
 *     level : restoring level.
 */
{
    alist l, l1;
    int   lg;

    /* find acell to be restored */
    l  = ndasave(n);
    l1 = ALNULL;
    while(l != ALNULL && allevel(l) >= level) {
        l1 = l;

        /* find opposite side */
        if((lg = alleg(l1)) < 0) {
            error("parestore", "illegal connection");
            fprintf(Stderr,"     (%d->%d)\n", n, alleg(l1));
            prdump();
            erfail();
        }

        ndparticle(n, lg) = PNULL;
        ndvertex(  n)     = VNULL;
        ndundef(   n)++;

        l  = alnext(l1);
        ndasave(n) = l;
        alfree(l1);
    }
#if DBG
    if(dbg >= padbg) {
        erdbg("parestore","end", padbg);
        printf("    n = %d\n", n);
        padump();
    }
#endif
}

/*==================================================================*/
Private Void    pareset(NoArg)
/*--------------------------------------------------------------------
 * Restore all.
 */
{
    node n;

    for(n = 0; n < nnodes; n++) {
        parestore(n, 0);
    }
}
/*==================================================================*/
Private int     paleg(n1, lg, n2, mode)
node    n1;
int     lg;
node    n2;
bool    mode;
/*--------------------------------------------------------------------
 * Find such a leg of "n2" that connects to ("n1", "lg").
 * If "mode" is true then check whether particle of ("n1", "leg") is
 * not yet assigned, else check already assigned.
 *
 * Arguments :
 *    
 * Returns :
 *   the leg : when such a leg is found.
 *   -1      : when such a leg is not found.
 */
{
    int l, il, ic;

    /* the expected number of connections */
    ic = (n1 == n2) ? 2 : 1;

    il = ndiline(n1, lg);
    for(l = 0; l < ndlegs(n2); l++) {
        if(ndiline(n2, l) == il) {
            if(mode) {
                if(ndparticle(n2, l) == PNULL) {
                    if(--ic == 0) {
                        return l;
                    }
                }
            } else {
                if(ndparticle(n2, l) != PNULL) {
                    if(--ic == 0) {
                        return l;
                    }
                }
            }
        }
    }

    /* not found */
    return -1;
}
/*==================================================================*/
Public Void     padump(NoArg)
/*--------------------------------------------------------------------
 * Dump
 */
{
    node n;
    int  l;

    /* the title */
    printf("No         legs vertex undef\n");

    /* dump nodes */
    for(n = 0; n < nnodes; n++) {
        printf("%2d %s %4d %6d %5d ", n, pastype(ndtype(n)),
                       ndlegs(n), ndvertex(n), ndundef(n));

        /* with the list of particles */
        if(inpa) {
            for(l = 0; l < ndlegs(n); l++) {
#if DBG
                if(ndnextn(n,l) < 0 || ndnextn(n,l) >= nnodes) {
                    fprintf(Stderr, "*err* padump : illegal nextn");
                    fprintf(Stderr, "ndnextn(%d, %d) = %d\n",
                            n, l, ndnextn(n, l));
                    exit(1);
                }
#endif
    
                printf("%d[%s] ", ndnextn(n,l),
                       pvname(ndparticle(n,l)));
            }
        }

        /* with save list */
        if(passg) {
            alprint(ndasave(n), stdout);
        }
        putchar(EOL);
    }
}

/*==================================================================*/
Private int      panasgn(nd)
node   nd;
/*--------------------------------------------------------------------
 * Find such a leg of node "nd" that is not particle assigned.
 *
 * Returns :
 *     found leg.
 */
{
    int legs;
    int l;

    legs = ndlegs(nd);

    for(l = 0; l < legs; l++) {
        if(ndparticle(nd, l) == PNULL) {
            return l;
        }
    }
    
    error("panasgn", "particles are assigned to all legs.");
    fprintf(Stderr, "node = %d\n", nd);
    erfail();

    /* dummy return for the compiler */
    return -1;
}

/*==================================================================*/
Private string  pastype(tp)
int     tp;
/*--------------------------------------------------------------------
 */
{
    if(tp == NDEXTERN) {
        return "external line";
    } else if(tp == NDTREEV) {
        return "tree vertex  ";
    } else if(tp == NDLOOPEDV) {
        return "looped vertex";
    } else {
        return "undefined    ";
    }
}
