/*********************************************************************
 * $id$
 *
 * File : axconn.c
 *
 * Eliminate scalar graphs in checking blocks and one-particle
 * irreducible parts of the graph.
 *
 * Last modification 1999/01/09
 *                                                    Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include "grc.h"

#define DBG 1
#define DBG1 0
/*
#ifdef DBG
#undef DBG
#define DBG 1
#endif
#define DBG1 1
*/

/*--------------------------------------------------------------------
 * local functions
 */
Private Void    axinit   ARG((NoArg));
Private Void    axsearch ARG((node,node,int,int*,int*,int*,int*,int*,int*,int*));
Private node    axnode   ARG((node,node*,int*,int*,int*));
Private node    axnextn  ARG((node,node,int,int,int,int*));
Private int     axloop   ARG((node,node));
Private bool    axskipbr ARG((int,int,int,int,int,int,int,int));

/*--------------------------------------------------------------------
 * local variables
 */
Local   int     axcount;     /* count the number of node */
Local   int     axvdef[MAXNODES];
Local   int     axvlow[MAXNODES];
Local   bool    axexit = False;

/*
 * the sequential number in the depth first search
 */
#define axdef(v)   axvdef[v]
#define axdir(lg)  (2*(lg) - 1)

/*
 * the lowest value axvdef on the way coming back in the depth
 * first search
 */
#define axlow(v)   axvlow[v]
#define BINEW     -1

/*==================================================================*/
Public  bool    axconn(NoArg)
/*--------------------------------------------------------------------
 *  Entry point of analysis of structure of graph with axiliary particles.
 */
{
    node nd, w;
    int  jl, je, jb, jk, jc, jd, le, nl, la, law;

#if DBG1
    axdbg = -1;
    rpflags();
#endif
#if DBG
    erenter("axconn");
#endif

    if(loop < 1 || !aexpand) {
        return True;
    }

    /* specially simple graph with only two external particles
     * as nodes.
     */
    nd = axnode(0, &w, &nl, &la, &law);
    if(ndtype(nd) != NDEXTERN || nl != 1 || w != 0) {
        fprintf(Stderr, "axconn: root is not external\n");
        abort();
    }

    axinit();

    le = -1;
    axsearch(nd, -1, 0, &jl, &je, &jb, &jk, &jc, &jd, &le);
    /*
     * jl : the number of loops in the last 1PI component
     * je : the number of external particles in the last 1PI component
     * jb - 1(root-v): the number of bridges (excluding ext.particle - vertex)
     * jk - 1(root-v) : the number of blocks (excluding ext.particle)
     */

#if DBG
     if(dbg >= axdbg) {
        printf("*dbg* axconn: graph %ld, ", nconfig+1);
        printf("iline = %d bridge = %d, block = %d, exts=%d, tadp=%d\n",
               loop+nnodes-1, jb-1, jk-1, axexts, axtadp);
    }
    erreturn("axconn");
#endif

    /* the graph is not accepted */
    if(axexit) {
#if DBG
        if(dbg >= axdbg) {
            printf("*dbg* axconn: return False\n");
        }
#endif
        return False;
    }

#if DBG
    if(dbg >= axdbg) {
        printf("*dbg* axconn: return True\n");
    }
#endif
    return True;
}

/*==================================================================*/
Private Void    axinit(NoArg)
/*--------------------------------------------------------------------
 *  Entry point of scalar graph generation.
 */
{
    int v;

    extsgraph  = False;
    for(v = 0; v < nextern; v++) {
        extbrv[v] = -1;
        extbrw[v] = -1;
    }

    axcount    = 0;
    axexit     = False;
    axtadp     = False;
    axexts     = False;
    axcountert = False;
    ax1pi      = True;
    axblock    = True;

    for(v = 0; v < nnodes; v++) {
        axdef(v) = BINEW;
        axlow(v) = BINEW;
    }

    nopitbl = 0;
}

/*==================================================================*/
Private Void    axsearch(v0, ov, ill, iloop, iext, ibrg, iblk, icbrg, iedg, lext)
node    v0, ov;
int     ill;
int     *iloop, *iext, *ibrg, *iblk, *icbrg, *iedg, *lext;
/*--------------------------------------------------------------------
 *   Search graph to find cut-points and bridges.
 *
 *   Arguments:
 *       v0    : input  : vertex to visit (should not be an external).
 *       ov    : input  : the last node.
 *       ill   : input  : i-line number connecting ov and v
 *       iloop : output : no. of loop in the current block.
 *       iext  : output : no. of external particles at the current block.
 *       ibrg  : output : no. of bridges below.
 *       iblk  : output : no. of blocks below.
 *       icbrg : output : no. of bridges or external particles
 *                        in the current 1PI component.
 *       iedg  : output : no. of edges in the current 1PI component.
 *       lext  : output : the last external particle visited
 */
{
    node w, v, v1;
    int im, ie, ib, ik, ic, id, lg, iln, nl, la, la1, is;
    bool skip;

#if DBG1
    printf("*dbg* axsearch : begin v0= %d, ov = %d, ill = %d\n", v0, ov, ill);
#endif

    v = axnode(v0, &v1, &nl, &la, &la1);

    /* visit node "v" */
    axcount++;
    *iext     = 0;
    *ibrg     = 0;
    axdef(v)  = axcount;
    axlow(v)  = axcount;
    *iloop    = axloop(v, v1);
    *iblk     = 0;
    *icbrg    = 0;
    *iedg     = 0;

    /* visit sons */
    is = 0;
    for(lg = 0; lg < nl; lg++) {
        w = axnextn(v, v1, lg, la, la1, &iln);
        if(iln == ill) {
            continue;
        }

        ic = 0;

        /* self loop */
        if(w == v) {
            /* self loop is visited twice for each i-line */
            if(is%2 == 0) {
                (*iloop)++;
                (*iblk)++;
                (*iedg)++;
            }
            is++;

        /* external particle which may be the original root */
        } else if(ndtype(w) == NDEXTERN) {
            (*iext)++;
            *lext = w;
            ic = 1;

        /* edge (v, w) is a back edge.
         * An edge is visited twice.
         */
        } else if(axdef(w) != BINEW) {
            axlow(v)  = min(axlow(v), axdef(w));

            /* # loops are counted in one of two visits
             */
            if(axdef(v) < axdef(w)) {
                (*iloop)++;
                (*iedg)++;
            }

        /* new vertex */
        } else {
            axsearch(w, v, iln, &im, &ie, &ib, &ik, &ic, &id, lext);
            if(axexit) {
#if DBG1
                printf("*dbg* axsearch : end : axexit\n");
#endif
                return;
            }
            *ibrg  += ib;
            *iblk  += ik;

            /* cut point */
            if(axlow(w) >= axdef(v)) {
                (*iblk)++;

                /* select by the last bloc */
                if(*iblk > 1) {
                    axblock = False;
                }
                if(ablock && !axblock) {
                    axexit = True;
#if DBG1
                    printf("*dbg* axsearch : return False by block\n");
#endif
                    return;
                }
            }

            /* bridge */
            if(axlow(w) > axdef(v)) {
                (*ibrg)++;

                /* for the 1PI component of vertex w */
                ic++;

                /* select by the last 1PI part */
                skip = axskipbr(v, w, *ibrg, ie, im, ic, id, *lext);
                if(skip) {
                    axexit = True;
# if DBG1
                    printf("*dbg* axsearch : end : axexit:2\n");
# endif
                    return;
                }

                /* for the next 1PI component of vertex v */
                ic = 1;
                id = 0;
                im = 0;

            /* not bridge */
            } else {
                axlow(v) = min(axlow(v), axlow(w));
                id++;
            }
            *iloop += im;
            *iext  += ie;
            *iedg  += id;
        }
        *icbrg += ic;
#if DBG
        if(dbg >= axdbg) {
            printf("*dbg* axsearch : %d->(%d, %d) = %d",
                   ov, v, w, edgetbl[v][w]);
            printf(", im=%d,ie=%d,ib=%d,ik=%d\n",
                   *iloop,*iext,*ibrg,*iblk);
        }
#endif
    } /* of for */
#if DBG1
    printf("*dbg* axsearch : end\n");
#endif

    /* Free legs are regarded to be connected to a virtual node
     * which is considered to be placed above the root.
     */
    if(ndfree(v) > 0) {
        fprintf(Stderr, "* axsearch: vertex %d with free legs\n", v);
        abort();
    }
#if DBG
    if(dbg >= axdbg) {
        printf("*dbg* axsearch : %d->(%d)",
               ov, v);
        printf(", im=%d,ie=%d,ib=%d,ik=%d\n", *iloop,*iext,*ibrg,*iblk);
    }
#endif
}

/*==================================================================*/
Private node    axnode(v, vn, nl, ia, ian)
node    v, *vn;
int     *nl, *ia, *ian;
/*--------------------------------------------------------------------
 * w = axnode(v, &vn, &ia, &nl)
 *
 *      ia     ian
 *    w ---------- vn
 *          aux
 */
{
    int   lg, lg1, na = 0, la, la1;
    node  w, v1;
    int   p;

    na = 0;
    for(lg = 0; lg < ndlegs(v); lg++) {
        p = ndparticle(v, lg);
        if(ptable[abs(p)]->ptype == P_AUX) {
            na++;
            la = lg;
            v1 = ndnextn(v, lg);
        }
    }
    if(na == 0) {
        *vn  = v;
        *nl  = ndlegs(v);
        *ia  = *nl;
        *ian = *nl;
        return v;
    } else if(na != 1) {
        fprintf(Stderr, "*** axnode:%d: too many auxiliary particles\n", v);
        abort();
    }

    /* check # aux. of v1 */
    na  = 0;
    la1 = -1;
    for(lg1 = 0; lg1 < ndlegs(v1); lg1++) {
        p = ndparticle(v1, lg1);
        if(ptable[abs(p)]->ptype == P_AUX) {
            na++;
            la1 = lg1;
        }
    }
    if(na != 1) {
        fprintf(Stderr, "*** axnode:%d:%d: auxiliary particles\n", v, v1);
        abort();
    }
    if(v < v1) {
        w    = v;
        *ia  = la;
        *vn  = v1;
        *ian = la1;
    } else {
        w    = v1;
        *ia  = la1;
        *vn  = v;
        *ian = la;
    }
    *nl = ndlegs(v) + ndlegs(v1) - 2;
#if DBG1
    printf("axnode(%d) : w = %d, vn = %d, ia = %d, ian = %d\n",
           v, w, *vn, *ia, *ian);
#endif
    return w;
}

/*==================================================================*/
Private node    axnextn(v, v1, lg, la, la1, il)
node    v, v1;
int     lg, la, la1, *il;
/*--------------------------------------------------------------------
 */
{
    node  vn, vv;
    int   lv, lv1, lg1;
    int   vv1, nl, ia, ian;

    if(v == v1) {
        *il = ndiline(v, lg);
        vv  = ndnextn(v, lg);

        /* vv may be connected with aux. particle */
        vn  = axnode(vv, &vv1, &nl, &ia, &ian);
        return vn;
    } else if(la < 0 || la1 < 0) {
        fprintf(Stderr, "*** axnextn : la = %d, la1 = %d\n", la, la1);
        abort();
    }

    lv  = ndlegs(v);
    lv1 = ndlegs(v1);

    if(lg >= lv + lv1 - 2) {
        fprintf(Stderr, "*** axnextn : lg = %d, lv = %d, lv1 = %d\n", lg, lv, lv1);
        abort();
    }

    /* lg = lv - 1 : aux. */
    lg1 = lg + 1 - lv;
    if(lg < la) {
        *il = ndiline(v, lg);
        vv  = ndnextn(v, lg);
    } else if(lg < lv - 1) {
        *il = ndiline(v, lg + 1);
        vv  = ndnextn(v, lg + 1);
    } else if(lg1 < la1) {
        *il = ndiline(v1, lg1);
        vv  = ndnextn(v1, lg1);
    } else {
        *il = ndiline(v1, lg1 + 1);
        vv  = ndnextn(v1, lg1 + 1);
    }
    vn  = axnode(vv, &vv1, &nl, &ia, &ian);
#if DBG1
    printf("axnextn(%d, %d, %d, %d, %d) : vn = %d, il = %d, vv = %d\n",
           v, v1, lg, la, la1, vn, *il, vv);
#endif
    return vn;
}

/*==================================================================*/
Private int     axloop(v, v1)
node    v, v1;
/*--------------------------------------------------------------------
 */
{
    if(v == v1) {
        return ndloop(v);
    }
    return ndloop(v) + ndloop(v1);
}

/*==================================================================*/
Private bool    axskipbr(v, w, brg, ext, lp, cb, ed, lex)
int     v, w, brg, ext, lp, cb, ed, lex;
/*--------------------------------------------------------------------
 * Test the graph is acceptable when cut-point is detected
 * Selection rules are given by flags.
 *
 * Arguments:
 *     v    : the curent node.
 *     w    : the node at the oposit side of the bridge
 *     brg  : the total number of bridges
 *            not includes external particles.
 *     ext  : the number of external particles.
 *     lp   : the number of loops in the current 1PI part.
 *     cb   : the number of bridges of the current 1PI part
 *            External particle are included.
 *     ed   : the number of edges of the current 1PI part
 *     lex  : the last external particle visited.
 *
 * Flages set in this function
 *
 *  axtadp     : the graph includes tadpoles
 *  ax1pi      : the graph includes only one 1PI part
 *  axcountert : the 1PI part is a counter term
 *  axexts     : the graph includes external self-energy parts
 *  axextct    : the external selef-energy is one counter term. 
 *
 */
{
    static bool lbiexts = False;
    int p, exts1 = True, exts2 = True, ex1 = -1, ex2 = -1, cbiexts;
    int tadp, texts;
    int vn, vv1, nl, ia, ian;

    if(brg > 1) {
        ax1pi = False;
    } else if(brg > 0) {
        if(v == 0) {
            ax1pi = True;
        } else {
            ax1pi = False;
        }
    } else {
        ax1pi = True;
    }

    /* tadpole or VEV */
    if(ext < 1) {
        tadp   = True;
        axtadp = True;
    } else {
        tadp   = False;
    }

    /* expanded or not */
    vn  = axnode(w, &vv1, &nl, &ia, &ian);
    if(lp > 0 && nl == cb && axloop(vn, vv1) == lp) {
        axcountert = True;
    } else {
        axcountert = False;
    }

    /* self energy part at an extenal particle
     * This flag checks both sides of the bridge
     */
    texts   = False;         /* the current 1pi part is an ext-self. */
    cbiexts = lbiexts;
    lbiexts = False;
    if(ext == 1) {
        texts   = True;
        cbiexts = True;
        p   = abs(ndparticle(lex, 0));
        ex1 = lex;
        if(p == PNULL) {
            fprintf(Stderr, "*** particle for the external particle %d ", lex);
            fprintf(Stderr, "is not defined.");
        } else {
            exts1 = ptable[p]->extself;
        }
        extsgraph = True;
        extbrv[lex] = w;   /* nearer side to the external particle */
        extbrw[lex] = v;
    }
    if(v != 0 && ext == nextern - 1) {
        /* there are external self-energy part between v and root=0 */
        texts   = True;
        lbiexts = True;
        p   = abs(ndparticle(0, 0));
        ex2 = 0;
        if(p == PNULL) {
            fprintf(Stderr, "*** particle for the external particle 0 (root) ");
            fprintf(Stderr, "is not defined.");
        } else {
            exts2 = ptable[p]->extself;
        }
        extsgraph = True;
        extbrv[0] = v;   /* nearer side to the external particle */
        extbrw[0] = w;
    }
    if(texts) {
        axexts = True;
    }

#if DBG
    if(dbg>=axdbg+1) {
        printf("graphid=%ld, ",nconfig);
        printf("axskipbr: tadp=%d, ax1pi=%d, ", tadp, ax1pi);
        printf("axcountert =%d, axexts = %d, cb = %d, ed = %d\n",
                axcountert, texts , cb, ed);
        printf("selfe = %d, extself = %d, cb = %d, v = %d, w = %d",
                 aselfe, aextself, cb, v, w);
        printf("exts1 = %d (%d) , exts2 = %d (%d)\n",
                 exts1, ex1, exts2, ex2);
    }
#endif
#if DBG
    if(dbg>=axdbg) {
        printf("axskipbr: tadp=%d, ax1pi=%d, axcountert =%d, axexts = %d\n",
                          tadp   , ax1pi   , axcountert,     texts );
        printf("selfe = %d, extself = %d, cb = %d, v = %d, w = %d, ed = %d\n",
                aselfe, aextself, cb, v, w, ed);
    }
#endif

    /* tadpole */
    if(!atadpole && tadp) {
#if DBG
        if(dbg>=axdbg) {
            printf("axskipbr: return Skip by tadpole\n");
        }
#endif
        return True;
    }

    /* discard graph with expanded 2 point functions */
    if(!aselfe && cb == 2 && !axcountert) {
#if DBG
        if(dbg>=axdbg) {
            printf("axskipbr: return Skip by selfe\n");
        }
#endif
        return True;
    }

    /* aexpmin and aexpmax */
    if(cb < aexpmin || (aexpmax >= 0 && cb > aexpmax)) {
        if(ndloop(w) != lp) {     /* not counter term */
#if DBG
            if(dbg>=axdbg) {
                printf("axskipbr: return Skip by expmin/expmax\n");
            }
#endif
            return True;
        }
            
    }

    /* self-energy part at an external particle */
    if(texts ) {
        if(aextself == False) {
#if DBG
            if(dbg>=axdbg) {
                printf("axskipbr: return Skip by extself\n");
            }
#endif
            return True;
        } else if(aextself == INMODEL && (!exts1 || !exts2)) {
#if DBG
            if(dbg>=axdbg) {
                printf("axskipbr: return Skip by off-shell renorm.\n");
            }
#endif
            return True;
        }
    }

    /* more than one 1PI parts */
    if(aonepi && !ax1pi) {
#if DBG
        if(dbg>=axdbg) {
            printf("axskipbr: return Skip by onepi\n");
        }
#endif
        return True;
    }

    if(nopitbl >= MAXLOOP) {
        fprintf(Stderr, "*** axskipbr : nopitbl = %d > MAXLOOP = %d\n",
                        nopitbl, MAXLOOP);
        exit(FAIL);
    }

    opitbl[nopitbl].nlegs   = cb;
    opitbl[nopitbl].next    = ext;
    opitbl[nopitbl].loop    = lp;
    opitbl[nopitbl].nedges  = ed;
    opitbl[nopitbl].extself = cbiexts;
    nopitbl++;

#if DBG
    if(dbg>=axdbg) {
        printf("axskipbr: return OK\n");
    }
#endif
    return False;
}
