/*********************************************************************
 * $id$
 *
 * File : bfconn.c
 *
 * Classification of graphs
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include "grc.h"

/*--------------------------------------------------------------------
 * local functions
 */
Private Void    bfinit   ARG((NoArg));
Private Void    bfsearch ARG((node,node,int*,int*,int*,int*,int*));
Private node    bfnext   ARG((node,node,node));

/*--------------------------------------------------------------------
 * local variables
 */
Local   int     bfcount;     /* count the number of node */
Local   int     bfvdef[MAXNODES];
Local   int     bfvlow[MAXNODES];
Local   int     bfctbl[MAXNODES][MAXNODES];
Local   int     bfgtbl[MAXNODES][MAXNODES];

#define bfdef(v)   bfvdef[v]
#define bflow(v)   bfvlow[v]
#define BINEW     -1

/*==================================================================*/
Public  Void    bfginit(NoArg)
/*--------------------------------------------------------------------
 *  Entry point of scalar graph generation.
 */
{
    int v, w;

    for(v = 0; v < nnodes; v++) {
        for(w = 0; w < nnodes; w++) {
            bfgtbl[v][w] = 0;
        }
    }
}

/*==================================================================*/
Public  Void    bfgend(NoArg)
/*--------------------------------------------------------------------
 *  Entry point of scalar graph generation.
 */
{
    int v, w;
    long sum, n;

    printf("*** Table of graphs by largest 1PI part\n");
    sum = 0L;

    for(w = 0; w < MAXNODES; w++) {
        for(v = 0; v < MAXNODES; v++) {
            n = bfgtbl[v][w];
            if(n != 0) {
                printf("lp = %4d, ex = %4d, graphs = %12ld\n", w, v, n);
                sum += n;
            }
        }
    }
    printf("                      Total  = %12ld\n", sum);
}

/*==================================================================*/
Public  int     bfconn(NoArg)
/*--------------------------------------------------------------------
 *  Entry point of scalar graph generation.
 */
{
    int nd, jl, je, jb, jk, jc, l, n;

#if DBG
    erenter("bfconn");
#endif

    if((nd = bfnext(0, -1, -1)) == NDNULL) {
        return -1;
    }
    bfinit();
    bfsearch(nd, -1, &jl, &je, &jb, &jk, &jc);
    bfctbl[jc][jl]++;

    for(l = loop; l >= 0; l--) {
        for(n = nnodes; n >= 0; n--) {
            if(bfctbl[n][l] != 0) {
                bfgtbl[n][l]++;
                return l;
            }
        }
    }
    return -1;
}


/*==================================================================*/
Private Void    bfinit(NoArg)
/*--------------------------------------------------------------------
 *  Entry point of scalar graph generation.
 */
{
    int v, w;

    bfcount = 0;

    for(v = 0; v < nnodes; v++) {
        bfdef(v) = BINEW;
        bflow(v) = BINEW;
        for(w = 0; w < nnodes; w++) {
            bfctbl[v][w] = 0;
        }
    }
}

/*==================================================================*/
Private Void    bfsearch(v, ov, iloop, iext, ibrg, iblk, icbrg)
node    v, ov;
int     *iloop, *iext, *ibrg, *iblk, *icbrg;
/*--------------------------------------------------------------------
 *   Search graph to fined bridge.
 *
 *   Arguments:
 *       v : input  : node to visit.
 *       ov : input  : the last node.
 *       iloop : output : no. of loop below.
 *       iext  : output : no. of external particles below.
 *       ibrg  : output : no. of bridges below.
 *       icbrg : output : no. of bridges of the current opi component.
 */
{
    node w;
    int il, ie, ib, ik, ic, lp;

#if DBG
    if(dbg >= bidbg)
        printf("*dbg* bfsearch : begin v = %d, ov = %d\n", v, ov);
#endif

    /* visit node "v" */
    bfcount++;
    *iext     = 0;
    *ibrg     = 0;
    bfdef(v)  = bfcount;
    bflow(v)  = bfcount;
    *iloop    = 0;
    *iblk     = 0;
    *icbrg    = 0;

    /* visit sons */
    for(w = bfnext(v, ov, -1); w != NDNULL; w = bfnext(v, ov, w)) {
        ic = 0;
        if(w == v) {    /* self loop */
            (*iloop) += edgetbl[v][w] / 2;
            (*iblk)  += edgetbl[v][w] / 2;

        } else if(ndtype(w) == NDEXTERN) {    /* external particle */
            (*iext)++;
            ic = 1;

        } else if(bfdef(w) != BINEW) {  /* edge (v, w) is a back edge */
            bflow(v)  = min(bflow(v), bfdef(w));
            if(bfdef(v) < bfdef(w))
                (*iloop) += edgetbl[v][w];

        } else {    /* new vertex */
            bfsearch(w, v, &il, &ie, &ib, &ik, &ic);
            (*iloop) += il;
            (*iext)  += ie;
            (*ibrg)  += ib;
            (*iblk)  += ik;

            lp = ndloop(w);

            if(bflow(w) >= bfdef(v)) { /* cut point */
                (*iblk)++;
            }
            if(bflow(w) > bfdef(v)) {
                if(edgetbl[v][w] == 1) { /* bridge */
                    (*ibrg)++;
                    ic++;
                    bfctbl[ic][il]++;
                    *iloop -= il;
                    ic = 1;
                } else {                 /* multiple connection */
                    (*iloop) += edgetbl[v][w] - 1;
                }
            } else {                     /* not bridge */
                bflow(v) = min(bflow(v), bflow(w));
                (*iloop) += edgetbl[v][w] - 1;
            }
        }
        *icbrg += ic;
#if DBG
        if(dbg >= bidbg) {
            printf("*dbg* bfsearch : %d->(%d, %d) = %d",
                   ov, v, w, edgetbl[v][w]);
            printf(", il=%d,ie=%d,ib=%d,ik=%d\n",
                   *iloop,*iext,*ibrg,*iblk);
        }
#endif
    } /* of for */
    if(ndfree(v) > 0)
        bflow(v) = -1;
}

/*==================================================================*/
Private node    bfnext(v, ov, lv)
node    v, ov, lv;
/*--------------------------------------------------------------------
 *  Find sons of "v", where father of "v" is "ov".
 *  The latest found son is "lv".
 */
{
    node  w;

    for(w = lv + 1; w < nnodes; w++) {
        if(w == ov)
            continue;
        if(edgetbl[v][w] > 0)
            return w;
    }
    return NDNULL;
}

/*==================================================================*/
Public  Void    bfdump(fp)
FILE    *fp;
/*--------------------------------------------------------------------
 */
{
    int  n, l;

    fprintf(fp, "\n*** Dump for kbsconn\n");
    fprintf(fp, "+++ 1PI components\n");

    fprintf(fp, "loop legs | number\n");
    for(l = 0; l <= loop; l++) {
        for(n = 0; n <= nnodes; n++) {
            if(bfctbl[n][l] != 0) {
                fprintf(fp, "%4d %4d | %4d\n", l, n, bfctbl[n][l]);
            }
        }
    }
}

