/*********************************************************************
 * $id$
 *
 * File : biconn.c
 *
 * Eliminate scalar graphs in checking blocks and one-particle
 * irreducible parts of the graph.
 *
 * Last modification 1999/01/09
 *                                                    Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include "grc.h"

#ifdef DBG
#undef DBG
#define DBG 1
#endif
#define DBG1 0

/*--------------------------------------------------------------------
 * local functions
 */
Private Void    biinit   ARG((NoArg));
Private Void    bisearch ARG((node,node,int*,int*,int*,int*,int*,int*,int*));
Private node    binext   ARG((node,node,node));
Private bool    biskipcp ARG((int,int));
Private bool    biskipbr ARG((int,int,int,int,int,int,int,int));

/*--------------------------------------------------------------------
 * local variables
 */
Local   int     bicount;     /* count the number of node */
Local   int     bivdef[MAXNODES];
Local   int     bivlow[MAXNODES];
Local   bool    biexit = False;

/*
 * the sequential number in the depth first search
 */
#define bidef(v)   bivdef[v]

/*
 * the lowest value bivdef on the way coming back in the depth
 * first search
 */
#define bilow(v)   bivlow[v]
#define BINEW     -1

/*==================================================================*/
Public  bool    biconn(NoArg)
/*--------------------------------------------------------------------
 *  Entry point of scalar graph generation.
 */
{
    int nd, jl, je, jb, jk, jc, jd, le;

#if DBG
    erenter("biconn");
#endif

    /* not including outside loops */
/* for check
    if(loop < 1 || !sexpand) {
        return True;
    }
*/

    /* As function bisearch visits only from vertices,
     * the root is replaced by the neighbouring vertex in this
     * package.
     */
    if((nd = binext(0, -1, -1)) == NDNULL) {
        return True;
    }
    /* specially simple graph with only two external particles
     * as nodes.
     */
    if(ndtype(nd) == NDEXTERN) {
        return True;
    }
    nd = 0;

    biinit();

    le = -1;
    bisearch(nd, -1, &jl, &je, &jb, &jk, &jc, &jd, &le);
    /*  
     * jl : the number of loops in the last 1PI component
     * je : the number of external particles in the last 1PI component
     * jb - 1(root-v): the number of bridges (excluding ext.particle - vertex)
     * jk - 1(root-v) : the number of blocks (excluding ext.particle)
     */     

#if DBG
     if(dbg >= bidbg) {
        printf("*dbg* biconn: graph %ld, loop = %d(%d), ",
               nconfig, loop, jl);
        printf("extern = %d(%d)\n", nextern, je);

        printf("*dbg* biconn: ");
        printf("iline = %d bridge = %d, block = %d, type=%d\n",
               loop+nnodes-1, jb, jk, biexts*2 + bitadp);
    }
    erreturn("biconn");
#endif
#if DBG1
    printf("*dbg* biconn: graph %ld, ", nconfig);
    printf("iline = %d bridge = %d, block = %d, type=%d\n",
           loop+nnodes-1, jb-1, jk-1, biexts*2 + bitadp);
#endif

    /* the graph is not accepted */
    if(biexit) {
#if DBG
        if(dbg >= bidbg) {
            printf("*dbg* biconn: return False\n");
        }
#endif
#if DBG1
        printf("*dbg* biconn: return False\n");
#endif
        return False;
    }

#if DBG
    if(dbg >= bidbg) {
        printf("*dbg* biconn: return True\n");
    }
#endif
#if DBG1
    printf("*dbg* biconn: return True\n");
#endif
    return True;
}

/*==================================================================*/
Private Void    biinit(NoArg)
/*--------------------------------------------------------------------
 *  Entry point of scalar graph generation.
 */
{
    int v;

    extsgraph  = False;
    for(v = 0; v < nextern; v++) {
        extbrv[v] = -1;
        extbrw[v] = -1;
    }

    bicount    = 0;
    biexit     = False;
    bitadp     = False;
    biexts     = False;
    bicountert = False;
    bi1pi      = True;
    biblock    = True;

    for(v = 0; v < nnodes; v++) {
        bidef(v) = BINEW;
        bilow(v) = BINEW;
    }

    nopitbl = 0;
}

/*==================================================================*/
Private Void    bisearch(v, ov, iloop, iext, ibrg, iblk, icbrg, iedg, lext)
node    v, ov;
int     *iloop, *iext, *ibrg, *iblk, *icbrg, *iedg, *lext;
/*--------------------------------------------------------------------
 *   Search graph to find cut-points and bridges.
 *
 *   Arguments:
 *       v     : input  : vertex to visit (should not be an external).
 *       ov    : input  : the last node.
 *       iloop : output : no. of loop below including internal loops of v.
 *       iext  : output : no. of external particles below.
 *       ibrg  : output : no. of bridges below.
 *       iblk  : output : no. of blocks below.
 *       icbrg : output : no. of bridges or external particles
 *                        in the current 1PI component.
 *       iedg  : output : no. of edges in the current 1PI component.
 *       lext  : output : the last external particle visited
 */
{
    node w;
    int il, ie, ib, ik, ic, id;
    bool skip;

#if DBG
    if(dbg >= bidbg)
        printf("*dbg* bisearch : begin v = %d, ov = %d\n", v, ov);
#endif

    /* visit node "v" */
    bicount++;
    *iext     = 0;
    *ibrg     = 0;
    bidef(v)  = bicount;
    bilow(v)  = bicount;
    *iloop    = ndloop(v);
    *iblk     = 0;
    *icbrg    = 0;
    *iedg     = 0;

    /* visit sons */
    for(w = binext(v, ov, -1); w != NDNULL; w = binext(v, ov, w)) {
        ic = 0;

        /* self loop */
        if(w == v) {
            (*iloop) += edgetbl[v][w] / 2;
            (*iblk)  += edgetbl[v][w] / 2;
            (*iedg)  += edgetbl[v][w] / 2;

        /* external particle which may be the original root */
        } else if(ndtype(w) == NDEXTERN) {
            (*iext)++;
            *lext = w;
            ic = 1;

        /* edge (v, w) is a back edge.
         * An edge is visited twice.
         */
        } else if(bidef(w) != BINEW) {
            bilow(v)  = min(bilow(v), bidef(w));

            /* # loops are counted in one of two visits
             */
            if(bidef(v) < bidef(w)) {
                (*iloop) += edgetbl[v][w];
                (*iedg)  += edgetbl[v][w];
            }

        /* new vertex */
        } else {
            bisearch(w, v, &il, &ie, &ib, &ik, &ic, &id, lext);
            if(biexit) {
                return;
            }
            *ibrg  += ib;
            *iblk  += ik;

            /* cut point */
            if(bilow(w) >= bidef(v)) {
                (*iblk)++;

                /* select by the last block */
                skip = biskipcp(v, *iblk);
                if(sblock && skip) {
                    biexit = True;
                    return;
                }
            }

            /* bridge */
            if(bilow(w) > bidef(v)) {
                /* bridge is a single connection */
                if(edgetbl[v][w] == 1) {
                    (*ibrg)++;

                    /* for the 1PI component of vertex w */
                    ic++;

                    /* select by the last 1PI part */
                    skip = biskipbr(v, w, *ibrg, ie, il, ic, id, *lext);
                    if(skip) {
                        biexit = True;
                        return;
                    }

                    /* for the next 1PI component of vertex v */
                    ic = 1;
                    id = 0;
                    il = 0;

                /* more than one connection */
                } else {
                    il += edgetbl[v][w] - 1;
                    id += edgetbl[v][w];
                }

            /* not bridge */
            } else {
                bilow(v) = min(bilow(v), bilow(w));
                il += edgetbl[v][w] - 1;
                id += edgetbl[v][w];
            }
            *iloop += il;
            *iext  += ie;
            *iedg  += id;
        }
        *icbrg += ic;
#if DBG
        if(dbg >= bidbg) {
            printf("*dbg* bisearch : %d->(%d, %d) = %d",
                   ov, v, w, edgetbl[v][w]);
            printf(", il=%d,ie=%d,ib=%d,ik=%d\n",
                   *iloop,*iext,*ibrg,*iblk);
        }
#endif
    } /* of for */

    /* Free legs are regarded to be connected to a virtual node
     * which is considered to be placed above the root.
     */
    if(ndfree(v) > 0) {
        bilow(v) = -1;
    }
}

/*==================================================================*/
Private node    binext(v, ov, lv)
node    v, ov, lv;
/*--------------------------------------------------------------------
 *  Find a son of "v", where father of "v" is "ov".
 *  The son last found is "lv".
 */
{
    node  w;

    for(w = lv + 1; w < nnodes; w++) {
        if(w == ov)
            continue;
        if(edgetbl[v][w] > 0)
            return w;
    }
    return NDNULL;
}

/*==================================================================*/
Private bool    biskipcp(v, blk)
int     v;
int     blk;
/*--------------------------------------------------------------------
 * Test the graph is acceptable when cut-point is detected
 * Selection rules are given by flags.
 *
 * Arguments:
 *     v    : the current node
 *     blk  : the number of blocks
 */
{
    biblock = True;   /* the graph includes only one block */

    if(blk > 1) {
        biblock = False;
    }

#if DBG
    if(dbg>=bidbg) {
        printf("biskipcp: biblock=%d\n", biblock);
    }
#endif

    /* more than one blocks */
    if(sblock && !biblock) {
#if DBG
        if(dbg>=bidbg)
            printf("biskipcp: return True by block\n");
#endif
        return True;
    }

    /* not skip and continue to search */
#if DBG
    if(dbg>=bidbg)
        printf("biskipcp: return False\n");
#endif
    return False;
}

/*==================================================================*/
Private bool    biskipbr(v, w, brg, ext, lp, cb, ed, lex)
int     v, w, brg, ext, lp, cb, ed, lex;
/*--------------------------------------------------------------------
 * Test the graph is acceptable when cut-point is detected
 * Selection rules are given by flags.
 *
 * Arguments:
 *     v    : the curent node.
 *     w    : the node at the oposit side of the bridge
 *     brg  : the total number of bridges
 *            not includes external particles.
 *     ext  : the number of external particles.
 *     lp   : the number of loops in the current 1PI part.
 *     cb   : the number of bridges of the current 1PI part
 *            External particle are included.
 *     ed   : the number of edges of the current 1PI part
 *     lex  : the last external particle visited.
 *
 * Flages set in this function
 *
 *  bitadp     : the graph includes tadpoles
 *  bi1pi      : the graph includes only one 1PI part
 *  bicountert : the 1PI part is a counter term
 *  biexts     : the graph includes external self-energy parts
 *  biextct    : the external selef-energy is one counter term. 
 *
 */
{
    static bool lbiexts = False;
    int p, exts1 = True, exts2 = True, ex1 = -1, ex2 = -1, cbiexts;

    if(brg > 1) {
        bi1pi = False;
    } else if(brg > 0) {
        if(v == 0) {
            bi1pi = True;
        } else {
            bi1pi = False;
        }
    } else {
        bi1pi = True;
    }

    /* tadpole or VEV */
    if(ext < 1) {
        bitadp = True;
    } else {
        bitadp = False;
    }

    /* expanded or not */
    if(lp > 0 && ndlegs(w) == cb && ndloop(w) == lp) {
        bicountert = True;
    } else {
        bicountert = False;
    }

    /* self energy part at an extenal particle
     * This flag checks both sides of the bridge
     */
    biexts  = False;
    cbiexts = lbiexts;
    lbiexts = False;
    if(ext == 1) {
        biexts  = True;
        cbiexts = True;
        p   = abs(ndparticle(lex, 0));
        ex1 = lex;
        if(p == PNULL) {
            fprintf(Stderr, "*** particle for the external particle %d ", lex);
            fprintf(Stderr, "is not defined.");
        } else {
            exts1 = ptable[p]->extself;
        }
        extsgraph = True;
        extbrv[lex] = w;   /* nearer side to the external particle */
        extbrw[lex] = v;
    }
    if(v != 0 && ext == nextern - 1) {
        /* there are external self-energy part between v and root=0 */
        biexts  = True;
        lbiexts = True;
        p   = abs(ndparticle(0, 0));
        ex2 = 0;
        if(p == PNULL) {
            fprintf(Stderr, "*** particle for the external particle 0 (root) ");
            fprintf(Stderr, "is not defined.");
        } else {
            exts2 = ptable[p]->extself;
        }
        extsgraph = True;
        extbrv[0] = v;   /* nearer side to the external particle */
        extbrw[0] = w;
    }

#if DBG
    if(dbg>=bidbg+1) {
        printf("graphid=%ld, ",nconfig);
        printf("biskipbr: bitadp=%d, bi1pi=%d, ", bitadp, bi1pi);
        printf("bicountert =%d, biexts = %d, cb = %d, ed = %d\n",
                bicountert, biexts, cb, ed);
        printf("selfe = %d, extself = %d, cb = %d, v = %d, w = %d",
                 sselfe, sextself, cb, v, w);
        printf("exts1 = %d (%d) , exts2 = %d (%d)\n",
                 exts1, ex1, exts2, ex2);
    }
#endif
#if DBG
    if(dbg>=bidbg) {
        printf("biskipbr: bitadp=%d, bi1pi=%d, bicountert =%d, biexts = %d\n",
                          bitadp   , bi1pi   , bicountert,     biexts);
        printf("selfe = %d, extself = %d, cb = %d, v = %d, w = %d, ed = %d\n",
                sselfe, sextself, cb, v, w, ed);
    }
#endif

    /* tadpole */
    if(!stadpole && bitadp) {
#if DBG
        if(dbg>=bidbg) {
            printf("biskipbr: return True by tadpole\n");
        }
#endif
        return True;
    }

    /* discard graph with expanded 2 point functions */
    if(!sselfe && cb == 2 && !bicountert) {
#if DBG
        if(dbg>=bidbg) {
            printf("biskipbr: return True by selfe\n");
        }
#endif
        return True;
    }

    /* sexpmin and sexpmax */
    if(cb < sexpmin || (sexpmax >= 0 && cb > sexpmax)) {
        if(ndloop(w) != lp) {     /* not counter term */
#if DBG
            if(dbg>=bidbg) {
                printf("biskipbr: return True by sexpmin/sexpmax\n");
            }
#endif
            return True;
        }
            
    }

    /* self-energy part at an external particle */
    if(biexts) {
        if(sextself == False) {
#if DBG
            if(dbg>=bidbg)
                printf("biskipbr: return True by extself\n");
#endif
            return True;
        } else if(sextself == INMODEL && (!exts1 || !exts2)) {
#if DBG
            if(dbg>=bidbg)
                printf("biskipbr: return True by off-shell renorm.\n");
#endif
            return True;
        }
    }

    /* more than one 1PI parts */
    if(sonepi && !bi1pi) {
#if DBG
        if(dbg>=bidbg)
            printf("biskipbr: return True by onepi\n");
#endif
        return True;
    }

#if DBG
    if(dbg>=bidbg)
        printf("biskipbr: return False\n");
#endif

    /* not skip and continue to search */

    if(nopitbl >= MAXLOOP) {
        fprintf(Stderr, "*** biskipbr : nopitbl = %d > MAXLOOP = %d\n",
                        nopitbl, MAXLOOP);
        exit(FAIL);
    }

    opitbl[nopitbl].nlegs   = cb;
    opitbl[nopitbl].next    = ext;
    opitbl[nopitbl].loop    = lp;
    opitbl[nopitbl].nedges  = ed;
    opitbl[nopitbl].extself = cbiexts;
    nopitbl++;

    return False;
}
