/********************************************************************
 * $id$
 *
 *  File : elem.c
 *
 *  Operations on graph elements
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include "grc.h"

/*------------------------------------------------------------------*/
/* Local variable */
Local   bool    elalloc = False;

/*==================================================================*/
Public Void   elinit(NoArg)
/*--------------------------------------------------------------------
 *  Initialize the module of elements.
 *  This function may be called several times.
 */
{
    int i, j, msize, bptr;
    char *ptr;

    passg    = False;
    nnodes   = 0;
    nlgtbl   = 0;
    nconfig  = 0;

    if(!elalloc) {      /* for the first call */
        elalloc = True;
        msize =   sizeof(tgraph)
                + sizeof(tnode) * MAXNODES
                + sizeof(tleg)  * MAXLEGS  * MAXNODES;

        if((ptr = (char*) malloc(msize)) == NULL) {
            error("elinit", "memory allocation failed(1).");
            erfail();
        }
        cgraph        = (tgraph*) ptr;
        bptr          = sizeof(tgraph);
        cgraph->basen = bptr;
        bptr         += sizeof(tnode)*MAXNODES;
        cgraph->offsl = bptr;


        for(i = 0; i < MAXNODES; i++) {
            if((nodetbl[i] = (tdnode*)malloc(sizeof(tdnode)))
               == NULL) {
                error("elinit", "memory allocation failed(1).");
                erfail();
            }
            if((pntbl[i] = (tpntbl*)malloc(sizeof(tpntbl))) == NULL) {
                error("elinit", "memory allocation failed(2).");
                erfail();
            }

            ndtype(i)   = NDNULL;
        }
    } else {
        for(i = 0; i < MAXNODES; i++) {
            ndtype(i)   = NDNULL;
        }
    }
#if DBG
    for(i = 0; i < MAXNODES; i++) {
        for(j = 0; j < MAXNODES; j++) {
            nddeg(i, j) = 0;
        }
        ndtype(i) = NDNULL;
    }
#endif
    for(i = 0; i < MAXNODES; i++) {
        ndlevel(i) = NDINF;
        for(j = 0; j < MAXNODES; j++) {
            edgetbl[i][j] = 0;
        }
    }
}

/*==================================================================*/
Public Void   elgene(ptcl, infi)
pindex  ptcl;
int     infi;
/*--------------------------------------------------------------------
 *  Create external particle
 *
 *  Argument
 *      ptcl    : input : particle name
 *      infi    : input : "initial" or "final"
 */
{
    node   n;


    /* find empty slot in the node table and store */
    for(n = 0; n < MAXNODES; n++) {
        if(ndtype(n) == NDNULL) {
            nnodes++;


            ndbasel(n) = cgraph->offsl + nlgtbl*sizeof(tleg);
            nlgtbl++;

            ndtype(n)   = NDEXTERN;
            ndlegs(n)   = 1;
            ndloop(n)   = 0;
            ndfree(n)   = 1;
            ndundef(n)  = 0;

            ndname(n)  = ptcl;

            /* normalize to FINAL */
            ndiorf(n)  = infi;
            if(infi == ELINITIAL)
                ndname(n) *= ptable[abs(ptcl)]->antip;

            return;
        }
    }

    /* the node table is full */
    error("elgene", "too many nodes");
    prdump();
    erfail();
}

/*==================================================================*/
Public Void   elgenv(lg, lp, n)
int lg;
int lp;
int n;
/*--------------------------------------------------------------------
 *  Generate nodes.
 *
 *  Arguments:
 *    lg : input : the number of legs of the nodes.
 *    lp : input : the number of loops of the nodes.
 *    n  : input : the number of nodes to be generated.
 */
{
    node v;
    int p;

    for(v = 0; v < MAXNODES; v++) {
        if(ndtype(v) == NDNULL) {
            nnodes++;

            ndbasel(v)  = cgraph->offsl + nlgtbl*sizeof(tleg);
            nlgtbl     += lg;

            ndtype(v)   = (lp == 0) ? NDTREEV : NDLOOPEDV;
            ndlegs(v)   = lg;
            ndfree(v)   = lg;
            ndloop(v)   = lp;
            ndvertex(v) = VNULL;
            ndundef(v)  = lg;
            for(p = 0; p < lg; p++)
                ndparticle(v, p) = PNULL;

            if(--n < 1)
                return;
        }
    }
    error("elgenv", "too many nodes.");
    prdump();
    erfail();
}

/*==================================================================*/
Public Void   eldelv(lg, lp, n)
int lg;
int lp;
int n;
/*--------------------------------------------------------------------
 *  Delete nodes.
 *
 *  Arguments:
 *    lg : input : the number of legs of the node.
 *    lp : input : the number of loops of the node.
 *    n  : input : the number of nodes to be deleted.
 */
{
    node m;
    int  type;

    type = (lp == 0) ? NDTREEV : NDLOOPEDV;
    for(m = MAXNODES-1; m >= 0; m--) {
        if(ndtype(m) == type && ndlegs(m) == lg && ndloop(m) == lp){
            ndtype(m)   = NDNULL;
            nnodes--;
            if((--n) < 1)
                return;
        }
    }
    error("eldelv", "inconsistent configuration.");
    prdump();
    erfail();
}

/*==================================================================*/
Public Void   elgnfin(NoArg)
/*--------------------------------------------------------------------
 *  End of the node generation
 */
{
    int linner, legs;
    int i, j;

    /* check consistency */
    legs    = 0;
    linner  = 0;
    for (i = 0; i < nnodes; i++) {
        legs    += ndlegs(i);
        linner  += ndloop(i);
    }

    j  = nnodes - legs/2 + loop - linner - 1;
    if(legs%2 != 0 || j != 0) {
        error("elgnfin", "inconsistent configuration:");
        fprintf(Stderr, " i2=%d:even, j=%d:0\n", legs, 0);
        fprintf(Stderr, " nextern = %d, legs = %d, nnodes = %d,",
                        nextern, legs, nnodes);
        fprintf(Stderr, " loop = %d, linner = %d\n", loop, linner);
        eldump();
        erfail();
    }
}

/*==================================================================*/
Public Void   eldump(NoArg)
/*--------------------------------------------------------------------
 *  Print the current configurations of graphs elements.
 *  Table of nodes are printed in "lcdump".
 */
{
    node        n1, n2;

    printf("--------------------------\n");
    printf(" The number of nodes = %d\n", nnodes);

    printf("---- external line table ----\n");
    for(n1 = 0; n1 < nnodes; n1++)
        if(ndtype(n1) == NDEXTERN) {
            printf("%2d %s %s\n", n1, eliorf(ndiorf(n1)),
                   pvname(ndname(n1)));
        }
    putchar(EOL);

    printf("---- connection table ----\n");
    printf("      ");
    for(n2 = 0; n2 < nnodes; n2++) {
        printf("%4d", n2);
    }
    putchar(EOL);

    for(n1 = 0; n1 < nnodes; n1++) {
        printf("%4d :", n1);
        for(n2 = 0; n2 < nnodes; n2++) {
            printf("%4d", edgetbl[n1][n2]);
        }
        putchar(EOL);
    }
}

/*==================================================================*/
Public int      eledge(nd1, nd2, ptcl, mult)
node   nd1, nd2;
pindex *ptcl;
int    *mult;
/*--------------------------------------------------------------------
 *  make table edges going from nd1 to nd2.
 *  ptcl[] : is the sorted list of particles.
 *  mult[] : is the multiplicity 
 *  the size of ptcl is returned.
 */
{
    int    legs;
    int    m, l, i;
    int    np;
    pindex p;

    if(ndtype(nd1) == NDEXTERN) {
        legs = 1;
    } else {
        legs = ndlegs(nd1);
    }

    np = 0;
    for(l = 0; l < legs; l++) {
        if(ndnextn(nd1, l) == nd2) {
            p = ndparticle(nd1, l);

            i = np;
            for(m = 0; m < np; m++) {
                if(ptcl[m] > p) {
                    i = m;
                    break;
                } else if(ptcl[m] == p) {
                    i = -1;
                    mult[m]++;
                    break;
                }
            }
            /* new particle */
            if(i >= 0) {
                for(m = np-1; m >= i; m--) {
                    ptcl[m+1] = ptcl[m];
                    mult[m+1] = mult[m];
                }
                ptcl[i] = p;
                mult[i] = 1;
                np++;
            }
        }
    }
    return np;
}
