/*********************************************************************
 * $id$
 *
 *  File : lclass.c
 *
 *  Manipulation of L-class
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include "grc.h"
#include "perm.h"


/*--------------------------------------------------------------------
 * Local functions
 */
Private Void    lcmkcn       ARG((NoArg));
Private bool    lcmklc       ARG((bool));
Private bool    lcdef        ARG((bool));
Private Void    lcsave       ARG((NoArg));
Private Void    lcsepc       ARG((node));
Private node    lcmorb       ARG((lclass*));

Private long    lcdtest      ARG((bool));
Private long    lcpmult      ARG((NoArg));
Private int     lccmpr       ARG((int*,bool));

Private int     lcccmp       ARG((node, node, int, bool));
Private int     lcedge       ARG((node, node, pindex*, int*));
Private long    lcmult       ARG((node, node));
Private int     lcncmp       ARG((node, node));
Private int     lccpcmp      ARG((int, pindex, int, pindex));
Private int     lcpcmp       ARG((node, node));

/*--------------------------------------------------------------------
 * Local variables
 */
/* node-class table. */
Local bool   lcexist = False;   /* class table is exist or not */
Local long   lcnsym;    /* symmetry factor */

/* stack */
Local int    lcstkpt;  /* stack pointer */
#define lcstpos(x)      ((x)*(nnodes+1))

Local int  iself, fself;   /* selection intermidate, final graph */
Local int  testf, prf, stopf;

/*==================================================================*/
Public  Void    lcdump(mode)
int     mode;
/*--------------------------------------------------------------------
    Dump tables.

    mode  = 0 : dump only ndclass.
         != 0 : dump also l-class table.
--------------------------------------------------------------------*/
{
    int    j, v;
    lclass c;

    if(!lcexist)
        lcmkcn();       /* restore table */

    if(mode != 0) {
        /* title */
        printf("\n--------- l-class --------\n");

        /* the number of classes */
        printf("%d l-classes\n", nlclass);

#if DBG
        printf("lcctop, lcpton\n");
        for(c = 0; c <= nnodes; c++) {
            printf(" %6d  %6d\n", lcctop[c], lcpton[c]);
        }
#endif

        /* l-classes as sets of nodes */
        for(c = 0; c < nlclass; c++) {
            printf("class %4d = {", c);
            for(j = lcctop[c]; j < lcctop[c+1]; j++)
                printf(" %2d", lcpton[j]);
            printf(" }\n");
        }
    }

    /* scale */
    printf("\n--------- node table --------\n");
    printf("node :  lvl : free : clss :  typ : legs : loop : ");
    for(j = 0; j < nnodes; j++)
        printf("%4d", j);
    printf("\n");

    /* nddeg */
    for(v = 0; v < nnodes; v++) {
        printf("%4d : %4d : %4d : ", v, ndlevel(v), ndfree(v));
        printf("%4d : %4d : %4d : %4d : ", ndclass(v), ndtype(v),
               ndlegs(v), ndloop(v));
        for(j = 0; j < nnodes; j++)
            printf("%4ld", nddeg(v,j));
        printf("\n");
    }
}

/*==================================================================*/
Public  Void    lcinit(NoArg)
/*--------------------------------------------------------------------
  Construct inital l-class.
  1. clear edge table.
  2. sort nodes
  3. initialize node table
  4. initialize stack of l-class.
--------------------------------------------------------------------*/
{
    int    v, tt, f;
    Uint   tb;
    node   v1, v2;
    tdnode *t;
    tpntbl *tp;
    bool   cont;

    /* clear edge table */
    for(v1 = 0; v1 < nnodes; v1++) {
        ndlevel(v1) = NDINF;            /* level is infinity */
        ndclass(v1) = 0;
        for(v2 = 0; v2 < nnodes; v2++) {
            edgetbl[v1][v2] = 0;
        }
    }

    /* sort node by (type, legs, loop) */
    for(v1 = 0; v1 < nnodes-1; v1++) {
        cont = True;
        while (cont) {
            cont = False;
            for(v2 = v1+1; v2 < nnodes; v2++) {
                if(lcncmp(v1, v2) > 0) {
                      /* exchange all elements */

                      tt          = ndtype(v1);
                      ndtype(v1)  = ndtype(v2);
                      ndtype(v2)  = tt;

                      tt          = ndvertex(v1);
                      ndvertex(v1)= ndvertex(v2);
                      ndvertex(v2)= tt;

                      tt          = ndlegs(v1);
                      ndlegs(v1)  = ndlegs(v2);
                      ndlegs(v2)  = tt;

                      tt          = ndloop(v1);
                      ndloop(v1)  = ndloop(v2);
                      ndloop(v2)  = tt;

                      tb          = ndbasel(v1);
                      ndbasel(v1) = ndbasel(v2);
                      ndbasel(v2) = tb;

                      tt          = ndclass(v1);
                      ndclass(v1) = ndclass(v2);
                      ndclass(v2) = tt;

                      t           = nodetbl[v1];
                      nodetbl[v1] = nodetbl[v2];
                      nodetbl[v2] = t;

                      tp          = pntbl[v1];
                      pntbl[v1]   = pntbl[v2];
                      pntbl[v2]   = tp;
                      cont = True;
                      break;
                 }
            }
        }
    }

    /* initial l-class of external lines */
    nlclass = 0;
    lcctop[nlclass] = 0;

    /* the first node is the root */
    lcpton[0] = 0;
    ndclass(0) = nlclass;

    /* initial l-class for other nodes */
    for(v = 1; v < nnodes; v++) {
        if(ndtype(v) == NDEXTERN ||
           lcncmp(v-1, v) != 0) {
            lcctop[++nlclass] = v;
        }
        lcpton[v]  = v;
        ndfree(v)  = ndlegs(v);
        ndclass(v) = nlclass;
    }
    lcctop[++nlclass] = nnodes;

    /* initailize stack of saving l-class */
    lcstkpt = 0;

    iself = cptest & 0x0f;
    fself = cptest >> 4;
    f       = fself;
    testf   = f & 0x0f;
    f       = f >> 4;
    prf     = f & 0x0f;
    f       = f >> 4;
    stopf   = f;
    if(stopf > 0)
        display = -1;

    if(cptest != 0) {
        printf("lcinit: cptest=%d, iself=%d, fself=%d, ", cptest,iself,fself);
        printf("testf=%d, prf=%d, stopf=%d\n", testf,prf,stopf);
    }
}

/*==================================================================*/
Public  bool    lcupdate(NoArg)
/*--------------------------------------------------------------------
  Update l-classes.
  Old class is pushed to the stack
  lcreset() is needed after calling this function.
--------------------------------------------------------------------*/
{
     lcsave();
     return lcdef(False);
}

/*==================================================================*/
Private Void    lcsave(NoArg)
/*--------------------------------------------------------------------
  Save the current classes to the stack.
--------------------------------------------------------------------*/
{
    int b, i;

#if DBG
    erdbg("lcsave", "begin", lcdbg);
#endif
    if(lcstpos(lcstkpt+1) >= LCSTACK) {
        error("lcsave", "stack overflow.");
        erfail();
    }

    b = lcstpos(lcstkpt);
    for(i = 0; i < nnodes; i++) {   /* memcpy */
        lcstack[b + i] = ndclass(i);
    }
    lcstack[b + nnodes] = nlclass;
    lcstkpt++;
}

/*==================================================================*/
Public  Void    lcreset(NoArg)
/*--------------------------------------------------------------------
  Restore the last class saved in the stack by "lcupdate".
--------------------------------------------------------------------*/
{
    int b, i;

    lcstkpt--;
    if(lcstkpt < 0) {
        error("lcreset", "stack underflow.");
        erfail();
    }
#if DBG
    erdbg("lcreset", "begin", lcdbg);
#endif

    b = lcstpos(lcstkpt);
    for(i = 0; i < nnodes; i++) {     /* memcpy */
        ndclass(i) = lcstack[b + i];
    }
    nlclass = lcstack[b + nnodes];
    lcexist = False;
}

/*==================================================================*/
Private Void    lcmkcn(NoArg)
/*--------------------------------------------------------------------
  Construct class table "lcctop" and "lcpton" from "ndclass"
--------------------------------------------------------------------*/
{
    lclass      c;
    node        n, np;

    for(c = 0; c <= nlclass; c++) {          /* clear pointer table */
        lcctop[c] = 0;
    }

    for(n = 0; n < nnodes; n++) {                    /* insert node */
        c = ndclass(n);                         /* class of the node */
        lcctop[c+1]++;
    }
    for(c = 0; c < nlclass; c++) {
        lcctop[c+1] += lcctop[c];
    }
    for(n = 0; n < nnodes; n++) {
        lcpton[n] = NDNULL;
    }
    for(n = 0; n < nnodes; n++) {
        c = ndclass(n);
        for(np = lcctop[c]; np < lcctop[c+1]; np++) {
            if(lcpton[np] == NDNULL) {
                lcpton[np] = n;
                break;
            }
        }
    }
    lcexist = True;
}

/*==================================================================*/
Private bool    lcmklc(assign)
bool    assign;
/*--------------------------------------------------------------------
  Construct class table "lcctop", "lcpton" and "ndclass" from "nodetbl"
  There must be better way to sort, since sorting is needed only
  within a class.
--------------------------------------------------------------------*/
{
    node        v, j;
    bool        cont;
    int         t, r, ncmp, nlc, k, lv;
    int         cmax, dmax;
    lclass      c;

    /* sort nodetbl[v] */
    for(v = 0; v < nnodes; v++) {
        lcpton[v] = v;
    }

    if(!lcexist)
        lcmkcn();

    ncmp = nlclass;

    for(c = 0; c < nlclass; c++) {
        if(lcctop[c+1] - lcctop[c] < 2)
            continue;
        for(v = lcctop[c]; v < lcctop[c+1]-1; v++) {
            cont = True;
            for(k = 0; k < nnodes*nnodes; k++) {
                cont = False;
                for(j = v + 1; j < lcctop[c+1]; j++) {
                    r = lcccmp(lcpton[v], lcpton[j], ncmp, assign);

                    /* if r = 0 then in ther order of node number */
                    if(r < 0 || (r == 0 && lcpton[v] > lcpton[j])) {
                        t = lcpton[v];
                        lcpton[v] = lcpton[j];
                        lcpton[j] = t;
                        cont = True;
                        break;
                    }
                }
                if(!cont)
                    break;
            }
            if(cont) {
                fprintf(Stderr,"lcmklc: system error\n");
                prdump();
                exit(1);
            }
        }
    }

    /* new class : as modification of "ndclass()" gives side effect
     *             for comparison in "lcccmp", new class number
     *             is stored in "ndword[]" first.
     */
    lv   = -1;
    nlc  = -1;
    for(c = 0; c < nlclass; c++) {
        v = -1;
        cmax = -1;
        dmax = -1;
        for(t = lcctop[c]; t < lcctop[c+1]; t++) {
            j = v;
            v = lcpton[t];
            if(j < 0 || (r = lcccmp(j, v, ncmp, assign)) != 0) {
                nlc++;
                ndwork[nlc] = t;
                cmax = max(cmax, dmax);
                dmax = -1;
            }
            if(cmax > v) {
                lv = v;
            }
            dmax = max(dmax, v);
            ndwork2[v] = nlc;
        }
    }

    nlc++;
    ndwork[nlc] = nnodes;
    nlclass = nlc;

    for(t = 0; t <= nlclass; t++) {
        lcctop[t] = ndwork[t];
    }
    for(v = 0; v < nnodes; v++) {
        ndclass(v) = ndwork2[v];
    }

    lcexist = True;

    if(lv >= 0) {
        return False;
    }

    return True;
}


/*==================================================================*/
Private bool    lcdef(assign)
bool assign;
/*--------------------------------------------------------------------
  Construct l-class from the current graph.
--------------------------------------------------------------------*/
{
    node v;
    int  j;
    int   c;

#if DBG
    erdbg("lcdef", "begin", lcdbg);
#endif
    do {    /* while sub-division is possible */
        c = nlclass;    /* nlclass is updated in lcmklc */
        for(v = 0; v < nnodes; v++) {
            for(j = 0; j < nnodes; j++)
                nddeg(v, j) = 0;
            for(j = 0; j < nnodes; j++)
                nddeg(v, ndclass(j)) += lcmult(v, j);
        }
        if(!lcmklc(assign)) {
            return False;
        }
    } while (c != nlclass);

    return True;
}

/*==================================================================*/
Public  node    lcfind(level, lnd)
int     level;
node    lnd;
/*--------------------------------------------------------------------
  Find a node with free legs in the specified level.
  This function returns found node.  If not found, it returns "NDNULL".

  Argument:
    level : intput : level to be searched.
--------------------------------------------------------------------*/
{
    node   nd;

    for(nd = lnd + 1; nd < nnodes; nd++) {
        if(ndlevel(nd) == level) {
            if(ndfree(nd) > 0) {
                return nd;
            }
        }
    }
    return NDNULL;
}

/*==================================================================*/
Public  node    lctrgt(snd, lnd)
node    snd, lnd;
/*--------------------------------------------------------------------
  Find a target node for the source node "snd".
  The last target node is "lnd".
  This function returns found node.  If not found, it returns "NDNULL".
--------------------------------------------------------------------*/
{
    node   nd;

    for(nd = lnd + 1; nd < nnodes; nd++) {
        if(edgetbl[nd][snd] == 0 && ndfree(nd) > 0) {
            return nd;
        }
    }
    return NDNULL;
}

/*==================================================================*/
Public  bool    lcftsto(assign)
bool    assign;
/*--------------------------------------------------------------------
--------------------------------------------------------------------*/
{
        return ((lcnsym = lcdtest(assign)) > 0);
}

/*==================================================================*/
Public  bool    lcftstf(assign)
bool    assign;
/*--------------------------------------------------------------------
--------------------------------------------------------------------*/
{
    lclass lc;
    node nd;

    if(lcmaxs() < 2)
        return True;

    lcsave();

    lc = 0;
    while((nd = lcmorb(&lc)) != NDNULL) {
        lcsepc(nd);
        if(!lcdef(assign)) {
            lcreset();
            return False;
        }
    }
    lcreset();
    return True;
}

/*==================================================================*/
Public  bool    lcftest(assign)
bool    assign;
/*--------------------------------------------------------------------
  Test duplication of graph
--------------------------------------------------------------------*/
{
    bool st, stf, sto;
    static int cnt = 0;

    lcnsym = 1;

    /* 
     * testf = 0 : del. by orderly algorithm
     * testf = 1 : del. by vertex fixing
     * testf = 2 : del. by both
     * testf = 3 : not delete
     */
    /* delete by orderly algorithm */
    if(testf == 0) {
        st = lcftsto(assign);
        if(prf && !st) {
            printf("Graph is deleted (orderly %d)\n", testf);
        }
#if DBG
        if(st)
            orda1++;
        else
            orda0++;
#endif

    /* delete by vertex fixing algorithm */
    } else if(testf == 1 || testf == 2){
        st = stf = lcftstf(assign);
        if(testf == 2) {
            if(!lcexist)
                lcmkcn();
            sto = lcftsto(assign);
            st  = sto && stf;
            if((!st && prf) || sto != stf) {
                if(sto != stf)
                    printf("%ld : ", nconfig);
                printf("Graph is deleted ");
                if(!stf)
                    printf("(vertex_fixing)");
                if(!sto)
                    printf("(orderly)");
                printf("(%d)\n", testf);
            }
            if(stopf && stf != sto) {
                fograph();
                prgraph();
                cnt++;
                if(cnt >= stopf) {
                    exit(1);
                }
            }
        } else {
            if(prf && !stf) {
                printf("Graph is deleted (vertex fixing %d)\n", testf);
            }
        }

    /* not delete */
    } else {
        return True;
    }

    return st;
}

/*==================================================================*/
Private Void    lcsepc(nd)
node    nd;
/*--------------------------------------------------------------------
  Take out node "nd" from its l-class and put to a new l-class
--------------------------------------------------------------------*/
{
    lclass lc;
    node   n;

    lc = ndclass(nd);
    for(n = 0; n < nnodes; n++) {
        if(n != nd && ndclass(n) >= lc)
            (ndclass(n))++;
    }
    nlclass++;
    lcexist = False;
}

/*==================================================================*/
Public  int     lcmaxs(NoArg)
/*--------------------------------------------------------------------
  Return the maximum size of l-class.
--------------------------------------------------------------------*/
{
    lclass c;
    int    t, m=0;

    if(!lcexist)
        lcmkcn();       /* restore table */

    for(c = 0; c < nlclass; c++) {
        t = lcctop[c+1] - lcctop[c];
        m = max(m, t);
    }
    return m;
}

/*==================================================================*/
Private node    lcmorb(lc)
lclass  *lc;
/*--------------------------------------------------------------------
  Find an orbit, whose number of elements > 1.
  This function returns the minimum node in such an orbit.
--------------------------------------------------------------------*/
{
    lclass c;
    node   n;
    int    p;

    for(c = *lc; c < nlclass; c++) {
        if(lcctop[c+1] - lcctop[c] > 1) {
            p = lcctop[c];
            n = lcpton[p++];
            for( ; p < lcctop[c+1]; p++)
                n = min(n, lcpton[p]);
            *lc = c;
            return n;
        }
    }
    return NDNULL;
}

/*==================================================================*/
Public  long    lcatest(NoArg)
/*--------------------------------------------------------------------
  Test after the particle assignment.
  lcreset() is needed after calling this function.
--------------------------------------------------------------------*/
{
    bool st;
    long nsym;

    nsym = 1;
    lcsave();

#if DBG
    if(dbg > padbg) {
        printf("lcatest : lcmaxs = %d\n", lcmaxs());
        lcdump(0);
    }
#endif

    if(lcmaxs() < 2)
        return nsym*lcpmult();

    st = lcdef(True);

#if DBG
    if(dbg > padbg) {
        printf("lcatest : result of lcdef: lcmaxs = %d\n", lcmaxs());
        lcdump(0);
    }
#endif

    if(st) {
        st = lcftest(True);
    }

    if(st) {
        nsym = lcnsym*lcpmult();  /* calculated in lcftest */
    } else {
        nsym = ((long) 0);
    }

    return nsym;
}


/*********************************************************************
 */


/*==================================================================*/
Private long    lcdtest(assign)
bool    assign;
/*--------------------------------------------------------------------
  Test duplication of graphs and calculate symmetry factor.
--------------------------------------------------------------------*/
{
    int  *a, r;
    long nsym;


    /* the first permutation is 1 */
    a = pminit(nlclass, lcctop);

    nsym = 1;
    while(pmnext()) {

        r = lccmpr(a, assign);   /* compare graphs */
        if(r < 0) {
            return 0;
        } else if(r > 0) {
        } else {
            nsym++;
        }
    }
    
    return nsym;
}
/*==================================================================*/
Private int     lccmpr(prm, assign)
int     prm[];
bool    assign;
/*--------------------------------------------------------------------
  Test duplication of graphs and calculate symmetry factor.
--------------------------------------------------------------------*/
{
    int  j, k, l, r;
    int  nj0, nj1, n0, n1;
    int  lg0[MAXTL+1], ln0[MAXTL+1];
    int  lg1[MAXTL+1], ln1[MAXTL+1];

    /* product of permutation  lcpton*prm */
    for(j = 0; j < nnodes; j++) {
        ndwork[j] = lcpton[prm[j]];
    }

#if DBG
    if(dbg >= lcdbg) {

        printf("lcpton,    prm\n");

        for(j = 0; j < nnodes; j++)
            printf("%3d ", lcpton[j]);
        printf("    ");
        for(j = 0; j < nnodes; j++)
            printf("%3d ", prm[j]);
        putchar(EOL);
        putchar(EOL);

        for(j = 0; j < nnodes; j++) {
            for(k = 0; k < nnodes; k++)
                printf("%3d ", edgetbl[lcpton[j]][lcpton[k]]);
            printf("    ");
            for(k = 0; k < nnodes; k++)
                printf("%3d ", edgetbl[ndwork[j]][ndwork[k]]);
            putchar(EOL);
        }
        putchar(EOL);
    }
#endif

    /* compare graphs */
    for(j = 0; j < nnodes; j++) {
        nj0 = lcpton[j];
        nj1 = ndwork[j];
        for(k = j; k < nnodes; k++) {
            r = edgetbl[nj0][lcpton[k]] 
              - edgetbl[nj1][ndwork[k]];
            if(r != 0)
                return r;
        }
    }

    if(!assign)
        return 0;

    /* compare assigned particle */
    for(j = 0; j < nnodes; j++) {
        nj0 = lcpton[j];
        nj1 = ndwork[j];
        for(k = j; k < nnodes; k++) {
            if(edgetbl[nj0][lcpton[k]] == 0)
                continue;
            n0 = lcedge(nj0, lcpton[k], lg0, ln0);
            n1 = lcedge(nj1, ndwork[k], lg1, ln1);
            if(n0 != n1) {
                error("lccmpr", "*** cmcmpr : system error");
                fprintf(Stderr, "n0(%d) = %d, n1(%d,%d) = %d\n",
                       nj0, lcpton[k], n0, nj1, ndwork[k]);
                exit(1);
            }
            for(l = 0; l < n0; l++) {
                if((r = lg0[l] - lg1[l]) != 0)
                    return r;
                if((r = ln0[l] - ln1[l]) != 0)
                    return r;
            }
        }
    }

    return 0;
}

/*==================================================================*/
Private long    lcpmult(NoArg)
/*--------------------------------------------------------------------
    calculate symmetry factor for multiple connections.
--------------------------------------------------------------------*/
{
    int  i, j, k, l, m, n;
    int  lg[MAXTL+1], ln[MAXTL+1];
    long sf;

    sf = 1;
    for(i = 0; i < ndnodes(); i++) {
        for(j = 0; j <= i; j++) {

            /* find multiple connection */
            if(edgetbl[i][j] > 1) {

                /* for looped vertex */
                if(ndlegs(i) > MAXTL) {
                    error("lcpmult", "illegal # legs.");
                    fprintf(Stderr, "node = %d, l = %d\n",
                            i, ndlegs(i));
                    erfail();
                }

                /* classify connections by particle */
                n = lcedge(i, j, lg, ln);
                l = edgetbl[i][j];

                /* check consistency between ndnextn and edgetbl */
                if(l != edgetbl[i][j]) {
                    error("lcpmult", "inconsistent connection");
                    fprintf(Stderr, "nodes: %d -- %d, ", i, j);
                    fprintf(Stderr, "edg = %d, l = %d\n", 
                            edgetbl[i][j], l);
                    erfail();
                }
                /* mulitply factorials */
                for(k = 0; k < n; k++) {
                    if(i == j) {
                        if(lg[k] < 0)
                            continue;
                        ln[k] /= 2;
                        if(ptable[lg[k]]->antip > 0) {
                            for(m = 1; m <= ln[k]; m++)
                                sf *= 2;
                        }
                    }
                    for(m = 2; m <= ln[k]; m++)
                        sf *= m;
                }
            }
        }
    }
#if DBG
    if(dbg >= sfdbg)
        printf("lcpmult: %ld\n", sf);
#endif

    return sf;
}

/*==================================================================*/
Private int    lcncmp(nd1, nd2)
node nd1, nd2;
/*--------------------------------------------------------------------
 * compare nodes "nd1" and "nd2" in the lexcographical ordering
 * of its table elements e for "s" <= e < "n".
 */
{
    int  r;

    if((r = ndtype(nd1) - ndtype(nd2)) != 0)
        return r;

    if((r = ndlegs(nd1) - ndlegs(nd2)) != 0)
        return r;

    if((r = ndloop(nd1) - ndloop(nd2)) != 0)
        return r;

    return 0;

}

/*==================================================================*/
Private int    lcccmp(nd1, nd2, n, assign)
node nd1, nd2;
int  n;
bool assign;
/*--------------------------------------------------------------------
 * compare nodes "nd1" and "nd2" in the lexcographical ordering
 * of its table elements : e < "n".
 */
{
    int  v;
    long r;

    for(v = 0; v < n; v++) {
        if((r = nddeg(nd1, v) - nddeg(nd2, v)) != 0) {
            return r;
        }
    }
    if(assign)
        return lcpcmp(nd1, nd2);
    else
        return 0;

}

/*==================================================================*/
Private int    lccpcmp(c1, p1, c2, p2)
int    c1;
pindex p1;
int    c2;
pindex p2;
/*--------------------------------------------------------------------
 */
{
    int r;

    if((r = c1 - c2) != 0)
        return r;
    return (p1 - p2);
}

/*==================================================================*/
Private int    lcpcmp(nd1, nd2)
node nd1, nd2;
/*--------------------------------------------------------------------
 * compare nodes "nd1" and "nd2" in the lexcographical ordering.
 * This function compares by assigned particles.
 * This is callable only after completing particle assignmnet.
 */
{
    node   nd[2];
    int     elcl[2][MAXTL];
    pindex  elpt[2][MAXTL];
    vindex n;
    int    l, i, cl, lg, r = 0;
    pindex pt;

    nd[0] = nd1;
    nd[1] = nd2;
    lg = ndlegs(nd1);
    if(lg != ndlegs(nd2))
        return (lg - ndlegs(nd2));

    /* for looped vertex */
    if(lg > MAXTL) {
        fprintf(Stderr,"*** lcpcmp : too many legs, lg = %d\n", lg);
        exit(16);
    }
    
    /* construct sorted table of particles and l-class
     * of opposit node for each leg
     */
    for(n = 0; n < 2; n++) {
        for(l = 0; l < lg; l++) {
            cl = ndclass(ndnextn(nd[n], l));
            pt = ndparticle(nd[n], l);

            /* find the position to be inserted */
            for(i = l - 1; i >= 0; i--) {
                if(lccpcmp(cl, pt, elcl[n][i], elpt[n][i]) < 0) {
                    elcl[n][i+1] = elcl[n][i];
                    elpt[n][i+1] = elpt[n][i];
                } else {
                    break;
                }
            }
            elcl[n][i+1] = cl;
            elpt[n][i+1] = pt;
        }
    }

#if DBG
    if(dbg > padbg) {
        for(n = 0; n < 2; n++) {
            printf("lcpcmp : nd%d=%d : ", n, nd[n]);
            for(l = 0; l < lg; l++)
                printf("(%d,%d) ", elcl[n][l], elpt[n][l]);
            putchar(EOL);
        }
    }
#endif

    for(l = 0; l < lg; l++) {
        r = lccpcmp(elcl[0][l], elpt[0][l], elcl[1][l], elpt[1][l]);
#if DBG
        if(dbg > padbg)
            printf("lcpcmp : r=%d \n",r);
#endif
        if(r != 0) {
            return r;
        }
    }
    return r;
}

/*==================================================================*/
Private long   lcmult(nd1, nd2)
node    nd1, nd2;
/*--------------------------------------------------------------------
 * returns multiplicity of the connection between "nd1" and "nd2".
 * The value is coded as ("MAXTL" + 1)**multiplicity.
 *
 * Check for overflow is necessary.
 */
{
    long        val;
    int         i;

    if(edgetbl[nd1][nd2] < 1)
        return 0;

    val = 1;
    for(i = 0; i < edgetbl[nd1][nd2] - 1; i++) {
        val *= MAXTL+1;
    }
    if(nd1 != nd2 && val > 1)
        val *= MAXTL;        /* assume max(deg(n1)) <= 4 */
    return val;
}

/*==================================================================*/
Private int      lcedge(nd1, nd2, ptcl, mult)
node   nd1, nd2;
pindex *ptcl;
int    *mult;
/*--------------------------------------------------------------------
 *  make table edges going from nd1 to nd2.
 *  ptcl[] : is the sorted list of particles.
 *  mult[] : is the multiplicity 
 *  the size of ptcl is returned.
 */
{
    int    legs;
    int    m, l, i;
    int    np;
    pindex p;

    if(ndtype(nd1) == NDEXTERN)
        legs = 1;
    else
        legs = ndlegs(nd1);

    np = 0;
    for(l = 0; l < legs; l++) {
        if(ndnextn(nd1, l) == nd2) {
            p = ndparticle(nd1, l);

            i = np;
            for(m = 0; m < np; m++) {
                if(ptcl[m] > p) {
                    i = m;
                    break;
                } else if(ptcl[m] == p) {
                    i = -1;
                    mult[m]++;
                    break;
                }
            }
            /* new particle */
            if(i >= 0) {
                for(m = np-1; m >= i; m--) {
                    ptcl[m+1] = ptcl[m];
                    mult[m+1] = mult[m];
                }
                ptcl[i] = p;
                mult[i] = 1;
                np++;
            }
        }
    }
    return np;
}
