/*******************************************************************
 * $id$
 *
 *    File : rdprocs.c
 *
 *    Read process
 *
 *    Calling sequence:
 *
 *      1. Initialization (read multiple process).
 *
 *              if(!rpinit(filename, False))
 *                  exit(FAIL);
 *
 *      2. Get next process.
 *
 *              while((nproc = rpread()) >= 0) {
 *                    ...
 *              }
 *
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "grc.h"
#include "pset.h"
#include "token.h"

/*--------------------------------------------------------------------
 * Private functions
 */
#define YorN(x)   ((x) ? "Yes": "No")

Private bool    rpget        ARG((NoArg));
Private bool    rpexternal   ARG((int,int,int));
Private bool    rppset       ARG((int));
Private Void    rpdup        ARG((int,int));
Private Void    rpchk        ARG((NoArg));
#if DBG
Private Void    rpdump       ARG((NoArg));
#endif

/*--------------------------------------------------------------------
 * Local variable
 */
Local   FILE    *rpfp;
Local   int     rpind;
Local   int     rpsord[MAXPROCS][MAXSORD]; 
                /* order of coupling constants */
Local   char    rpmdl[MAXPATHLEN];
Local   char    rpmtest[MAXPATHLEN];

Local   bool    rdset = False;
Local   bool    useset = False;

/* help message */
Local string syntax[] = {
    "Syntax of input process file",
    "",
    "\t <file>    ::= Model = <filename>;",
    "\t               <process-list>",
    "\t <process> ::= Process;",
    "\t                 <order-definition>",
    "\t                 <initial-definition>",
    "\t                 <final-definition>",
    "\t                 <flag-list>",
    "\t               Pend;",
    "\t <order-definition> ::= <coupling-name> = <int>;",
    "\t                      | <coupling-name> = { <int> [ , <int> ]* };",
    "\t <initial-definition> ::= { <particle-name> [ , <particle-name> ]* };",
    "\t   <final-definition> ::= { <particle-name> [ , <particle-name> ]* };",
    "\t <flag> ::= ",
    "\t\t | OPI       = <boolean>;",
    "\t\t | Block     = <boolean>;",
    "\t\t | Tadpole   = <boolean>;",
    "\t\t | SelfE     = <boolean>;",
    "\t\t | Extself   = <boolean>;",
    "\t\t | CounterT  = <boolean>;",
    "\t\t | AnyCT     = <boolean>;",
    "\t\t | ShrinkAUX = <boolean>;",
    "\t\t | Expand    = <boolean>;",
    "\t\t | ExpMin    = <int>;",
    "\t\t | ExpMax    = <int>;",
    "\t\t | Kinem     = <string>;",
    "\t\t | Prefix    = <string>;",
    "",
    "\t <boolean> ::= Yes | No"
};

/*==================================================================*/
Public  bool    rpinit(fn, rdst)
string  fn;
bool    rdst;
/*--------------------------------------------------------------------
    read processes

    rdset : enable set of particles
--------------------------------------------------------------------*/
{
    FILE*   echo;
    int     arg;
    trdsava sarea;

#if DBG
    erdbg("rpinit", "begin", rpdbg);
#endif

    /* echo : file pointer for echo of input. */
#if DBG
    echo = stdout;
#else
    echo = (FILE *) NULL;        /* no echo */
#endif

    rdset = rdst;

    /* allocate are for cprocess */
    if((cprocess = (tproc*) malloc(sizeof(tproc))) == NULL) {
        fprintf(Stderr, "\n*err* rpinit: memory allocation failed\n");
        exit(1);
    }

    /* open input file */
    if((rpfp = fopen(fn, "r")) == NULL) {
        fprintf(Stderr, "\n*err* rpinit: open error\n");
        fprintf(Stderr, "fname = %s\n", fn);
        return False;
    }

    /* initialize token package */
    rdinit(rpfp, RDNOCHANGE, echo);

    /* input : model="<model file>"; */
    rpind = rdstat();
    arg   = rdarg(rpind);
    fnmdl = rpmdl;
#if 1
    while(True) {
        /* dummy commands */
        if(rdchkid(rpind, "model")) {
            break;
        }
        fprintf(stderr, "*** ignore command \n");
        rpind = rdstat();
        arg   = rdarg(rpind);
    }
#endif

    if(rdchkid(rpind, "model") && rdtype(arg) == RDSTRING) {
        putchar('\n');
        pvfname(fnmdl, rdid(arg), MAXPATHLEN); /* model file name */
        rdsave(&sarea);      /* save file to use token for model */
        pvread(fnmdl);       /* read model file */
        rdrestore(&sarea);   /* recover file in the token package */
    } else {
        fprintf(Stderr, "\n*err* rpinit: model = \"<filename>\";");
        fprintf(Stderr, " is expected.\n");
        rdprint(rpind);
        rphelp(Stderr);
        return False;
    }

    /* the number of processes */
    rproc = 0;
    while (rpget()) /* read all processes into the table */
        ;

    rdend();
    fclose(rpfp);

    rpchk();

#if DBG
    if(dbg >= rpdbg) {
        rpdump();
        erdbg("rpinit", "end", rpdbg);
        for (arg = 0; arg < rproc; arg++) {
            rprin1(stdout, arg, True);
        }
    }
#endif

    if(useset) {
        psdump();
    }

    return True;
}

/*==================================================================*/
Private bool    rpget(NoArg)
/*--------------------------------------------------------------------
    read one process
--------------------------------------------------------------------*/
{
    bool err, value;
    int  arg, found, prc, noprc, np, i;

#if DBG
    erdbg("rpget", "begin", rpdbg);
#endif

    /* process; */
    err = False;
    if((rpind = rdstat()) == RDEOF) {
        return False;
    } else if(rdtype(rpind) != RDID) {
        fprintf(Stderr, "\n*err* rpget: \"process;\" expected (not id)\n");
        exit(1);

    /* input : model="<model file>"; */
    } else if(rdchkid(rpind, "model")) {
        arg   = rdarg(rpind);
        if(rdtype(arg) == RDSTRING) {
            putchar('\n');
            strncpy(rpmtest, rdid(arg), MAXPATHLEN); /* model file name */
            if(strcmp(rpmtest, fnmdl) != 0) {
                fprintf(Stderr, "\n*err* rpget: model should be same ");
                fprintf(Stderr, "as older one\n;");
                fprintf(Stderr, "old:\"%s\"\n;", fnmdl);
                fprintf(Stderr, "new:\"%s\"\n;", rpmtest);
                exit(1);
            }
        } else {
        fprintf(Stderr, "\n*err* rpinit: model = \"<filename>\";");
        fprintf(Stderr, " is expected.\n");
            rdprint(rpind);
            rphelp(Stderr);
            return False;
        }
        if((rpind = rdstat()) == RDEOF) {
            return False;
        } else if(rdtype(rpind) != RDID) {
            fprintf(Stderr, "\n*err* rpget: \"process;\" expected (not id)\n");
            exit(1);
        }

    /* pset; name=...; psend; */
    } else if(rdset && rdstrcmp(rdid(rpind), "pset") == 0) {
        useset = True;
        return rppset(rpind);
    } else if(rdset && rdstrcmp(rdid(rpind), "defaultpset") == 0) {
        useset = True;
        psinit();
        return True;
    }


    if(!rdchkid(rpind, "process")) {
        fprintf(Stderr, "\n*err* rpget: \"process;\" expected\n");
        rddump(rpind);
        exit(1);
    }

    /* check size of process table */
    if(rproc >= MAXPROCS - 1) {
        fprintf(Stderr, "\n*err* rpget: too many processes.\n");
        exit(1);
    }

    /* get space */
    np = rproc++;
    if((rpextn[np] = (rpextt*) malloc(sizeof(rpextt))) == NULL) {
        fprintf(Stderr, "\n*err* rpget: memory allocation error.\n");
        exit(1);
    }

    /* clear items and set by default values */
    rpextn[np]->nextern   = 0;
    rpextn[np]->expand    = True;
    rpextn[np]->expmin    = 0;
    rpextn[np]->expmax    = -1;           /* infinity */
    rpextn[np]->onepi     = False;
    rpextn[np]->block     = False;
    rpextn[np]->tadpole   = False;
    rpextn[np]->extself   = INMODEL;
    rpextn[np]->hep       = True;
    rpextn[np]->selfe     = False;
    rpextn[np]->countert  = True;
    rpextn[np]->anyct     = False;
    rpextn[np]->gm2       = False;
    rpextn[np]->undefp    = False;
    rpextn[np]->skipasp   = skipasp;
    rpextn[np]->kinem     = NULL;
    rpextn[np]->pprefix   = NULL;
    rpextn[np]->shrinkaux = True;
    rpextn[np]->cptest    = 0;

    /* prepare to accept more than one processes defined by the list
       of values of a coupling constant.    ex: qed={4, 6};
    */
    noprc = 0;   /* the number of processes */
    for(prc = 0; prc < MAXPROCS; prc++) {
        for(i = 0; i < MAXSORD; i++) {
            rpsord[prc][i] = 0;  /* buffer of coupling constants */
        }
    }

    /* read "pend;", "initial=...", "final=...", "opi=yes/no;", ... */
    while (True) {
        rpind = rdstat();
        arg = rdarg(rpind);

        if(rpind == RDEOF || rdtype(rpind) == RDEOF) {
            fprintf(Stderr, "\n*err* rpget: EOF before Pend\n");
            exit(1);

        } else if(rdtype(rpind) != RDID) {
            fprintf(Stderr, "\n*err* rpget: illegal statement\n");
            rdprint(rpind);
            err = True;

        /* pend; */
        } else if(rdstrcmp(rdid(rpind), "pend") == 0) {
            break;

        /* initial=...*/
        } else if(rdstrcmp(rdid(rpind), "initial") == 0) {
            err |= !rpexternal(arg, np, ELINITIAL);

        /* final=... */
        } else if(rdstrcmp(rdid(rpind), "final") == 0) {
            err |= !rpexternal(arg, np, ELFINAL);

        /* opi=yes/no; */
        } else if(rdstrcmp(rdid(rpind), "opi") == 0) {
            if(!rdbool(arg, &value)) {
                fprintf(Stderr, "\n*err* rpget: opi should yes/no\n");
                rdprint(rpind);
                err = True;
            }
            rpextn[np]->onepi = value;

        /* gm2=yes/no; */
        } else if(rdstrcmp(rdid(rpind), "gm2") == 0) {
            if(!rdbool(arg, &value)) {
                fprintf(Stderr, "\n*err* rpget: gm2 should yes/no\n");
                rdprint(rpind);
                err = True;
            }
            rpextn[np]->gm2 = value;

        /* expand=yes/no; */
        } else if(rdstrcmp(rdid(rpind), "expand") == 0) {
            if(!rdbool(arg, &value)) {
                fprintf(Stderr, "\n*err* rpget: expand should yes/no\n");
                rdprint(rpind);
                err = True;
            }
            rpextn[np]->expand = value;

        } else if(rdstrcmp(rdid(rpind), "expmin") == 0) {
            if(!rdintarg(arg, &value)) {
                fprintf(Stderr, "\n*err* rpget: expmin should integer\n");
                rdprint(rpind);
                err = True;
            }
            rpextn[np]->expmin = value;

        } else if(rdstrcmp(rdid(rpind), "expmax") == 0) {
            if(!rdintarg(arg, &value)) {
                fprintf(Stderr, "\n*err* rpget: expmax should integer\n");
                rdprint(rpind);
                err = True;
            }
            rpextn[np]->expmax = value;

        /* shrinkaux=yes/no; */
        } else if(rdstrcmp(rdid(rpind), "shrinkaux") == 0) {
            if(!rdbool(arg, &value)) {
                fprintf(Stderr, "\n*err* rpget: block should yes/no\n");
                rdprint(rpind);
                err = True;
            }
            rpextn[np]->shrinkaux = value;

        /* block=yes/no; */
        } else if(rdstrcmp(rdid(rpind), "block") == 0) {
            if(!rdbool(arg, &value)) {
                fprintf(Stderr, "\n*err* rpget: block should yes/no\n");
                rdprint(rpind);
                err = True;
            }
            rpextn[np]->block = value;

        /* tadpole=yes/no; */
        } else if(rdstrcmp(rdid(rpind), "tadpole") == 0) {
            if(!rdbool(arg, &value)) {
                fprintf(Stderr, "\n*err* rpget: tadpole should yes/no\n");
                rdprint(rpind);
                err = True;
            }
            rpextn[np]->tadpole = value;

        /* hep=yes/no; */
        } else if(rdstrcmp(rdid(rpind), "hep") == 0) {
            if(!rdbool(arg, &value)) {
                fprintf(Stderr, "\n*err* rpget: hep should yes/no\n");
                rdprint(rpind);
                err = True;
            }
            rpextn[np]->hep = value;

        /* extself=yes/no; */
        } else if(rdstrcmp(rdid(rpind), "extself") == 0) {
            if(!rdenum(arg, &value, nmdlbool, mdlbools, mdlboolv)) {
                fprintf(Stderr, "\n*err* rpget: extself should yes/no");
                for(i = 0; i < nmdlbool; i++) {
                    fprintf(Stderr, "/%s", mdlbools[i]);
                }
                fprintf(Stderr, "=\n");
                rdprint(rpind);
                err = True;
            }
            rpextn[np]->extself = value;

        /* selfe=yes/no; */
        } else if(rdstrcmp(rdid(rpind), "selfe") == 0) {
            if(!rdbool(arg, &value)) {
                fprintf(Stderr, "\n*err* rpget: selfe should yes/no\n");
                rdprint(rpind);
                err = True;
            }
            rpextn[np]->selfe = value;

        /* countert=yes/no; */
        } else if(rdstrcmp(rdid(rpind), "countert") == 0) {
            if(!rdbool(arg, &value)) {
                fprintf(Stderr, "\n*err* rpget: countert should yes/no\n");
                rdprint(rpind);
                err = True;
            }
            rpextn[np]->countert = value;

        /* anyct=yes/no; */
        } else if(rdstrcmp(rdid(rpind), "anyct") == 0) {
            if(!rdbool(arg, &value)) {
                fprintf(Stderr, "\n*err* rpget: anyct should yes/no\n");
                rdprint(rpind);
                err = True;
            }
            rpextn[np]->anyct = value;

        /* cptest=integer; */
        } else if(rdstrcmp(rdid(rpind), "cptest") == 0) {
            if(!rdintarg(arg, &value)) {
                fprintf(Stderr, "\n*err* rpget: cptest should integer\n");
                rdprint(rpind);
                err = True;
            }
            rpextn[np]->cptest = value;

        } else if(rdstrcmp(rdid(rpind), "kinem") == 0) {
            if(rdtype(arg) == RDSTRING){
                rpextn[np]->kinem =
                    (string) malloc((strlen(rdid(arg))+1)*sizeof(char));
                if(rpextn[np]->kinem == NULL) {
                    fprintf(Stderr, "\n*err* rpget: memory allocation error\n");
                    exit(1);
                }
                strcpy(rpextn[np]->kinem, rdid(arg));
            } else {
                fprintf(Stderr, "\n*err* rpget: kinem should string\n");
                rdprint(rpind);
                err = True;
            }

        } else if(rdstrcmp(rdid(rpind), "prefix") == 0) {
            if(rdtype(arg) == RDSTRING){
                rpextn[np]->pprefix =
                    (string) malloc((strlen(rdid(arg))+1)*sizeof(char));
                if(rpextn[np]->pprefix == NULL) {
                    fprintf(Stderr, "\n*err* rpget: memory allocation error\n");
                    exit(1);
                }
                strcpy(rpextn[np]->pprefix, rdid(arg));
            } else {
                fprintf(Stderr, "\n*err* rpget: prefix should string\n");
                rdprint(rpind);
                err = True;
            }

        /* coupling constants */
        /* xxx={number1, number2, ... }; */
        } else {
            found = False;
            arg = -1;
            for (i = 0; i < nsordtbl; i++) { /* find name of coupl. */
                if(rdstrcmp(sordtbl[i], rdid(rpind)) != 0) {
                    continue;
                }

                /* coupling constant is found*/
                found = True;

                /* for elements in list */
                for(arg = rdarg(rpind), prc = 0; arg >= 0;
                         arg = rdarg(arg), prc++) {
                    if(prc >= MAXPROCS) {
                        fprintf(Stderr, "\n*err* rpget: too many processes.\n");
                        rdprint(rpind);
                        err = True;
                        break;
                    }

                    if(rdtype(arg) != RDINT) {
                        fprintf(Stderr, "\n*err* rpget: illegal fname.\n");
                        rdprint(rpind);
                        err = True;
                        break;
                    } else if (err) {
                        break;
                    } else if (rpsord[prc][i] != 0) {
                        fprintf(Stderr, "\n*err* rpget: duplicate coupling.\n");
                        rdprint(rpind);
                        err = True;
                        break;
                    } else {
                        rpsord[prc][i] = rdint(arg);
                    }

                } /* for arg */

                if(!err && noprc > 0 && noprc != prc) {
                    fprintf(Stderr, "\n*err* rpget: different number");
                    fprintf(Stderr, " of processes.\n");
                    fprintf(Stderr, "Lengths of the list of coupling ");
                    fprintf(Stderr, "constant should be equal.\n");
                    rdprint(rpind);
                    err = True;
                }
                noprc = prc;

                break;
            } /* for coupling */

            if (!found) {
                fprintf(Stderr, "\n*err* rpget: illegal data.\n");
                rdprint(rpind);
                err = True;
            }
        } /* else : end of analysis of a statement */

    } /* while(True) : statements */

    /* check coupling constants */
    if (noprc < 1) {
        fprintf(Stderr, "\n*err* rpget: order of coupling constants ");
        fprintf(Stderr, " is not defined.\n");
        err = True;
    }

    /* check the number of coupling constants */
    if (rpextn[np]->nextern < 1) {
        fprintf(Stderr, "\n*err* rpget: no external particles\n");
        err = True;
    }

    if(err) {
        rdprint(rpind);
        rphelp(Stderr);
        exit(1);
    }

    /* copy order of coupling consts */
    rpdup(np, noprc);    

    /* Here values may be checked : see rpread */

#if DBG
    erdbg("rpget", "end", rpdbg);
#endif

    return True;
}

/*==================================================================*/
Private bool    rpexternal(arg, np, iorf)
int arg;   /* arg of the statement */
int np;    /* process number */
int iorf;  /* ELINITIAL or ELFINAL */
/*--------------------------------------------------------------------
    read external particle.
--------------------------------------------------------------------*/
{
    int  pt, opt, aopt;
    bool st;

#if DBG
    erdbg("rpexternal", "begin", rpdbg);
#endif

    pt = rpextn[np]->nextern;

    /* for elements in the argument list */
    st = True;
    for( ; arg >= 0; arg = rdarg(arg)) {

        /* particle name */
        if(rdtype(arg) != RDID) {
            fprintf(Stderr, "\n*err* rpexternal: illegal particle name.\n");
            rdprint(rpind);
            st = False;
            break;
        }

        if(pt >= MAXELINE - 1) {
            fprintf(Stderr, "\n*err* rpexternal:");
            fprintf(Stderr, " too many external particles.\n");
            st = False;
            break;
        }

        /* save 
           (if particle is not defined, then program stops in pvfind)
        */
        rpextn[np]->particle[pt] = pvfind(rdid(arg));
        rpextn[np]->iorf[pt] = iorf;
        if(rpextn[np]->particle[pt] == 0) {
            rpextn[np]->undefp = True;
        }

        /* option to the particle name */
        if((opt = rdopt(arg)) >= 0) {
            for( ; opt >= 0; opt = rdbrothr(opt)) {
                if(rdtype(opt) == RDID) {
                    if(rdstrcmp(rdid(opt),"pcode") == 0) {
                        aopt = rdarg(opt);
                        if(rdtype(aopt) != RDINT) {
                            fprintf(Stderr, "*err* rpexternal: ");
                            fprintf(Stderr, "illegal PCode\n");
                        } else {
                            rpextn[np]->spcode[pt] = (int)rdint(aopt);
                        }
                    } else if(rdstrcmp(rdid(opt),"kfcode") == 0) {
                        aopt = rdarg(opt);
                        if(rdtype(aopt) != RDINT) {
                            fprintf(Stderr, "*err* rpexternal: ");
                            fprintf(Stderr, "illegal KFCode\n");
                        } else {
                            rpextn[np]->skfcode[pt] = (int)rdint(aopt);
                        }
                    }
                }
            }
        }
            
        pt++;
    }

    /* the number of external particles */
    rpextn[np]->nextern = pt;

#if DBG
    erdbg("rpexternal", "end", rpdbg);
#endif

    return st;
}

/*==================================================================*/
Private Void    rpdup(np, nprc)
int np;
int nprc;
/*--------------------------------------------------------------------
    save table of coupling constants
    They are saved into from np-th to (np+nprc-1)-th processes.
    From (np+1)-th to (np+nprc-1)-th processes, other information
    is duplicated from np-th process.
--------------------------------------------------------------------*/
{
    int i, p, n;

#if DBG
    erdbg("rpdup", "begin", rpdbg);
#endif

    for(i = 0; i < nprc; i++) {

        /* from (np+1)-th to (np+nprc-1)-th process */
        if(i > 0) {

            /* check table size */
            if(rproc > MAXPROCS) {
                fprintf(Stderr, "\n*err* rpdup: too many processes.\n");
                exit(1);
            }

            /* get space */
            n = rproc++;
            if((rpextn[n] = (rpextt*) malloc(sizeof(rpextt)))
                     == NULL) {
                fprintf(Stderr, "\n*err* rpdup: memory allocation error\n");
                exit(1);
            }

            /* copy data */
            rpextn[n]->nextern   = rpextn[np]->nextern;
            rpextn[n]->expand    = rpextn[np]->expand;
            rpextn[n]->expmin    = rpextn[np]->expmin;
            rpextn[n]->expmax    = rpextn[np]->expmax;
            rpextn[n]->onepi     = rpextn[np]->onepi;
            rpextn[n]->block     = rpextn[np]->block;
            rpextn[n]->tadpole   = rpextn[np]->tadpole;
            rpextn[n]->extself   = rpextn[np]->extself;
            rpextn[n]->hep       = rpextn[np]->hep;
            rpextn[n]->selfe     = rpextn[np]->selfe;
            rpextn[n]->countert  = rpextn[np]->countert;
            rpextn[n]->anyct     = rpextn[np]->anyct;
            rpextn[n]->shrinkaux = rpextn[np]->shrinkaux;
            rpextn[n]->cptest    = rpextn[np]->cptest;
            rpextn[n]->gm2       = rpextn[np]->gm2;
            rpextn[n]->undefp    = rpextn[np]->undefp;
            rpextn[n]->skipasp   = rpextn[np]->skipasp;
            rpextn[n]->kinem     = rpextn[np]->kinem;
            rpextn[n]->pprefix   = rpextn[np]->pprefix;

            /* copy external particles */
            for(p = 0; p < rpextn[np]->nextern; p++) {
                rpextn[n]->particle[p] = rpextn[np]->particle[p];
                rpextn[n]->iorf[p]     = rpextn[np]->iorf[p];
                rpextn[n]->spcode[p]   = rpextn[np]->spcode[p];
                rpextn[n]->skfcode[p]  = rpextn[np]->skfcode[p];
            }

        /* np-th process */
        } else {
            n = np;
        }

        /* copy coupling constants */
        for(p = 0; p < nsordtbl; p++) {
            rpextn[n]->sorder[p] = rpsord[i][p];
        }
    }

#if DBG
    erdbg("rpdup", "end", rpdbg);
#endif

}

/*==================================================================*/
Public  int     rpread(NoArg)
/*--------------------------------------------------------------------
    read the next process from the process table.
--------------------------------------------------------------------*/
{
    static int nproc = 0, np;
    int lp2, i;

#if DBG
    erdbg("rpread", "begin", rpdbg);
#endif


    if(rproc <= nproc) { /* out of the process table */
#if DBG
        if(dbg >= rpdbg) {
            printf("*dbgr* pread :rproc = %d, nproc = %d\n", 
                   rproc, nproc);
        }
#endif

        return -1;
    }

    /* new process number */
    np = rprocc = nproc++;

    /* copy as the current process */
    nextern = rpextn[np]->nextern;   
             /* the number of external particles */

    order = 0;                       /* total order of coupling */
    for(i = 0; i < nsordtbl; i++) {
        sorder[i] = rpextn[np]->sorder[i];
        order    += sorder[i];
    }

    if(!osinit(sorder)) {
        osmessage();
        exit(1);
    }

    lp2 = order - nextern + 2;       /* 2*loop */
    if(lp2 % 2 != 0) {
        fprintf(Stderr, "\n*err* rpread: illegal order of coupling const.\n");
        rprin1(Stderr, np, True);
        exit(1);
    }
    loop = lp2 / 2;
    if(loop < 0) {
        fprintf(Stderr, "\n*err* rpread: negative number of loops.\n");
        rprin1(Stderr, np, True);
        exit(1);
    }

    /* flags */
    expand    = rpextn[np]->expand;
    expmin    = rpextn[np]->expmin;
    expmax    = rpextn[np]->expmax;
    onepi     = rpextn[np]->onepi;
    block     = rpextn[np]->block;
    tadpole   = rpextn[np]->tadpole;
    extself   = rpextn[np]->extself;
    hep       = rpextn[np]->hep;
    selfe     = rpextn[np]->selfe;
    countert  = rpextn[np]->countert;
    anyct     = rpextn[np]->anyct;
    shrinkaux = rpextn[np]->shrinkaux;
    cptest    = rpextn[np]->cptest;
    gm2       = rpextn[np]->gm2;
    undefp    = rpextn[np]->undefp;
    skipasp   = rpextn[np]->skipasp;
    kinem     = rpextn[np]->kinem;
    prefix    = rpextn[np]->pprefix;

    /* normalization */
    if(loop < 1 || !expand) {
        onepi    = False;
        block    = False;
        selfe    = False;
        countert = False;
        anyct    = False;
        hep      = True;
    } else {
        if(block) {
            onepi = True;
        }
        if(nextern < 2) {
            tadpole = True;
        } else if(nextern < 3) {
            extself = True;
            selfe   = True;
        }
    }

    if(!defaux) {
        shrinkaux = False;
    }
    if(shrinkaux) {
        sexpand  =        aexpand  = expand;
        sexpmin  = 0;     aexpmin  = expmin;
        sexpmax  = -1;    aexpmax  = expmax;
        stadpole = True;  atadpole = tadpole;
        sextself = True;  aextself = extself;
        sonepi   = False; aonepi   = onepi;
        sblock   = False; ablock   = block;
        sselfe   = True;   aselfe   = selfe;
    } else {
        sexpand  = aexpand  = expand;
        sexpmin  = aexpmin  = sexpmin;
        sexpmax  = aexpmax  = expmax;
        stadpole = atadpole = tadpole;
        sextself = aextself = extself;
        sonepi   = aonepi   = onepi;
        sblock   = ablock   = block;
        sselfe   = aselfe   = selfe;
    }

    /* initialize table of graph elements */
    elinit();

    /* enter external particles */
    for(i = 0; i < nextern; i++) {
        elgene(rpextn[np]->particle[i], rpextn[np]->iorf[i]);
    }

#if DBG
    erdbg("rpread", "end", rpdbg);
#endif

    /* construct current process data */
    cprocess->procid   = nproc;
    cprocess->graphs   = 0;        /* initial value */
    cprocess->nextern  = nextern;
    cprocess->loop     = loop;
    cprocess->order    = order;
    cprocess->bgraph   = 0;
    cprocess->maxnodes = MAXNODES;
    cprocess->maxlegs  = MAXLEGS;
    cprocess->msize    = sizeof(tgraph)
                + sizeof(tnode) * cprocess->maxnodes
                + sizeof(tleg)  * cprocess->maxlegs;
    cprocess->expand   = expand;
    cprocess->onepi    = onepi;
    cprocess->gm2      = gm2;
    cprocess->undefp   = undefp;
    cprocess->skipasp  = skipasp;
    cprocess->kinem    = kinem;
    cprocess->pprefix  = prefix;

    cprocess->mult     = PC_NORMAL;
    cprocess->proc1    = 0;
    cprocess->proc2    = 0;
    for(i = 0; i < nsordtbl; i++) {
        cprocess->sorder[i] = sorder[i];
    }

    return nproc;
}

/*==================================================================*/
Public  Void    rpropt(fp)
FILE   *fp;
/*--------------------------------------------------------------------
 *  Print the options of the process
 */
{
    bool pr;

    fprintf(fp, "%%\n");
    fprintf(fp, "%% Options of graph generation\n");
    fprintf(fp, "%%\n");
    fprintf(fp, "Expand=%s; ",      YorN(expand));
    fprintf(fp, "ExpMin=%d; ",      expmin);
    fprintf(fp, "ExpMax=%d;\n",     expmax);
    fprintf(fp, "OPI=%s; ",         YorN(onepi));
    fprintf(fp, "Block=%s; ",       YorN(block));
    fprintf(fp, "Tadpole=%s; ",     YorN(tadpole));
    fprintf(fp, "Extself=%s;\n",    YorN(extself));
    fprintf(fp, "Selfe=%s; ",       YorN(selfe));
    fprintf(fp, "Countert=%s; ",    YorN(countert));
    fprintf(fp, "AnyCT=%s;\n",      YorN(anyct));

    pr = False;
    if(kinem != NULL) {
        fprintf(fp, "Kinem=\"%s\"; ",   kinem);
        pr = True;
    }
    if(prefix != NULL) {
        fprintf(fp, "Prefix=\"%s\"; ",  prefix);
        pr = True;
    }
    if(pr) {
        fprintf(fp, "\n");
    }
    if(gm2) {
        fprintf(fp, "Gm2=%s;",        YorN(gm2));
    }
    fprintf(fp, "Undefp=%s; ",      YorN(undefp));
    fprintf(fp, "Assign=%s; ",      YorN(!skipasp));
    if(!hep) {
        fprintf(fp, "HEP=%s; ",        YorN(hep));
    }
    fprintf(fp, "\n");
}

/*==================================================================*/
Public  Void    rprint(fn)
string  fn;
/*--------------------------------------------------------------------
    Print the processes table in the form of process file.
--------------------------------------------------------------------*/
{
    FILE *fp;
    int  np;
    bool mode;

    if((fp = fopen(fn, "w")) == NULL) {
        fprintf(Stderr, "\n*err* rprint: open error to write ");
        fprintf(Stderr,"filename = %s\n", fn);
        exit(1);
    }

    mode = True;
    for(np = 0; np < rproc; np++) {
        rprin1(fp, np, mode);
        mode = False;
    }
    fclose(fp);
}

/*==================================================================*/
Public  Void    rprin1(fp, proc, mdl)
FILE *fp;
int proc;
bool mdl;  /* with model file name */
/*--------------------------------------------------------------------
    Print the processes.
--------------------------------------------------------------------*/
{
    int i, j;
    bool pr;

    /* name of model file */
    if(mdl) {
        fprintf(fp, "%%\n");
        fprintf(fp, "Model=\"%s\";\n", fnmdl);
    }

    /* process number */
    fprintf(fp, "%%\n");
    fprintf(fp, "Process; %% %d\n", proc);

    /* order of coupling constants */
    for(i = 0; i < nsordtbl; i++) {
        fprintf(fp, "%s=%d; ", sordtbl[i], rpextn[proc]->sorder[i]);
    }
    fprintf(fp,"\n");

    /* initail particles */
    j = 0;
    fprintf(fp, "Initial={");
    for(i = 0; i < rpextn[proc]->nextern; i++) {
        if(rpextn[proc]->iorf[i] == ELINITIAL) {
            if(j > 0) {
                fprintf(fp, ", ");
            }
            j++;
            fprintf(fp, "%s", pvname(rpextn[proc]->particle[i]));
            pr = False;
            if(rpextn[proc]->spcode[i] != 0) {
                if(pr) {
                    fprintf(fp, ";");
                } else {
                    fprintf(fp, "[");
                }
                fprintf(fp, "PCode=%d", rpextn[proc]->spcode[i]);
                pr = True;
            }
            if(rpextn[proc]->skfcode[i] != 0) {
                if(pr) {
                    fprintf(fp, ";");
                } else {
                    fprintf(fp, "[");
                }
                fprintf(fp, "KFCode=%d", rpextn[proc]->skfcode[i]);
                pr = True;
            }
            if(pr) {
                fprintf(fp, "]");
            }
        }
    }
    fprintf(fp,"};\n");

    /* final particles */
    j = 0;
    fprintf(fp, "Final  ={");
    for(i = 0; i < rpextn[proc]->nextern; i++) {
        if(rpextn[proc]->iorf[i] == ELFINAL) {
            if(j > 0) {
                fprintf(fp, ", ");
            }
            j++;
            fprintf(fp, "%s", pvname(rpextn[proc]->particle[i]));
            pr = False;
            if(rpextn[proc]->spcode[i] != 0) {
                if(pr) {
                    fprintf(fp, ";");
                } else {
                    fprintf(fp, "[");
                }
                fprintf(fp, "PCode=%d", rpextn[proc]->spcode[i]);
                pr = True;
            }
            if(rpextn[proc]->skfcode[i] != 0) {
                if(pr) {
                    fprintf(fp, ";");
                } else {
                    fprintf(fp, "[");
                }
                fprintf(fp, "KFCode=%d", rpextn[proc]->skfcode[i]);
                pr = True;
            }
            if(pr) {
                fprintf(fp, "]");
            }
        }
    }
    fprintf(fp,"};\n");

    /* flags */
    fprintf(fp,"Expand=%d; Opi=%d; Block=%d; HEP=%d; Gm2=%d\n",
            rpextn[proc]->expand,
            rpextn[proc]->onepi,
            rpextn[proc]->block, rpextn[proc]->hep,
            rpextn[proc]->gm2);
    fprintf(fp,"Tadpole=%d; Extself=%d; Selfe=%d; Countert=%d; AnyCT=%d\n",
            rpextn[proc]->tadpole, rpextn[proc]->extself,
            rpextn[proc]->selfe,   rpextn[proc]->countert,
            rpextn[proc]->anyct);
    if(rpextn[proc]->kinem != NULL) {
        fprintf(fp,"kinem=%s\n", rpextn[proc]->kinem);
    }
    if(rpextn[proc]->pprefix != NULL) {
        fprintf(fp,"prefix=%s\n", rpextn[proc]->pprefix);
    }

    /* end */
    fprintf(fp, "Pend;\n");
}

/*==================================================================*/
Public  Void    rphelp(fp)
FILE    *fp;
/*--------------------------------------------------------------------
    Print help
--------------------------------------------------------------------*/
{
    int i, synlen;

    synlen = sizeof(syntax)/sizeof(syntax[0]);
    for(i = 0; i < synlen; i++) {
        fprintf(fp, "%s\n", syntax[i]);
    }
}

/*==================================================================*/
Private Void    rpchk(NoArg)
/*--------------------------------------------------------------------
   Check consistency.
--------------------------------------------------------------------*/
{
    int np, i, lp2, lp, ordr, fn, ne, p, pt, a;
    bool err=False;

    for(np = 0; np < rproc; np++) {
        ne = rpextn[np]->nextern;   
    
        ordr = 0;                       /* total order of coupling */
        for(i = 0; i < nsordtbl; i++) {
            ordr += rpextn[np]->sorder[i];
        }
    
        lp2 = ordr - ne + 2;       /* 2*loop */
        if(lp2 % 2 != 0) {
            fprintf(Stderr, "** err **: illegal order of coupling const.\n");
            rprin1(Stderr, np, True);
            err = True;
        }
        lp = lp2 / 2;
        if(lp < 0) {
            fprintf(Stderr, "** err **: negative number of loops.\n");
            rprin1(Stderr, np, True);
            err = True;
        }

        if(!rpextn[np]->undefp && !useset) {
            fn = 0;
            for(p = 0; p < rpextn[np]->nextern; p++) {
                pt = rpextn[np]->particle[p];
                if(pt < 0) {
                    pt = - pt;
                    a  = -1;
                } else {
                    a  = 1;
                }
                switch(ptable[pt]->ptype) {
                    case P_DIRAC:
                    case P_GHOST:
                        if(majop) {
                            fn++;
                        } else if(rpextn[np]->iorf[p] == ELFINAL) {
                            fn -= a;
                        } else {
                            fn += a;
                        }
                        break;
                    case P_MAJORANA:
                        fn++;
                        break;
                    case P_SET:
                    case P_UNDEF:
                        break;
                }
            }
            if((fn != 0 && !majop) || (fn%2 != 0 && majop)) {
                fprintf(Stderr, "** err **: illegal fermion number\n");
                fprintf(Stderr, "fn=%d, majop=%d\n", fn, majop);
                rprin1(Stderr, np, True);
                err = True;
            }
        }
    }
    if(err) {
        exit(1);
    }
}

/*==================================================================*/
Private bool    rppset(arg)
int arg;
/*--------------------------------------------------------------------
 * Define a set of particles.
 */
{
    bool status = True;
    bool assign = False;
    int  set=-1, p, kfcode=0, pcode=0;

    while(True) {
        rpind = rdstat();

        if(rdtype(rpind) == RDEOF) {
            fprintf(Stderr, "\n*err* rppset : EOF\n");
            rddump(rpind);
            status = False;
            break;

        } else if(rdtype(rpind) != RDID) {
            fprintf(Stderr, "\n*err* rppset : illegal data\n");
            rddump(rpind);
            status = False;

        } else if(rdstrcmp(rdid(rpind), "psend") == 0) {
            break;

        /* pcode = integer; */
        } else if(rdstrcmp(rdid(rpind), "pcode") == 0) {
            arg = rdarg(rpind);
            if(rdtype(arg) != RDINT) {
                fprintf(Stderr, "\n*err* rppset :");
                fprintf(Stderr, " illgal pcode (not int)\n");
                rddump(rpind);
                status = False;
            } else {
                pcode = (int)rdint(arg);
            }

        /* kfcode = integer; */
        } else if(rdstrcmp(rdid(rpind), "kfcode") == 0) {
            arg = rdarg(rpind);
            if(rdtype(arg) != RDINT) {
                fprintf(Stderr, "\n*err* rppset :");
                fprintf(Stderr, " illgal kfcode (not int)\n");
                rddump(rpind);
                status = False;
            } else {
                kfcode = (int)rdint(arg);
            }

        /* not keyword : name_of_set={particle, ...}; */
        } else if(!assign) {
            set = psnew(rdid(rpind), 0);
            if(set <= 0) {
                fprintf(Stderr, "\n*err* rppset : psnew failed\n");
                status = False;
            }
            assign = status;

            for(arg = rdarg(rpind); arg >= 0; arg = rdarg(arg)) {
                if(rdtype(arg) != RDID) {
                    fprintf(Stderr, "\n*err* rppset : illegal particle\n");
                    rddump(rpind);
                    status = False;
                    break;
        
                } else {
                    p = pvfind(rdid(arg));
                    if(ptable[abs(p)]->ptype == P_SET) {
                        fprintf(Stderr, "\n*err* rppset : ");
                        fprintf(Stderr, "a set cannot be a member of");
                        fprintf(Stderr, "another set.\n");
                        rddump(rpind);
                        status = False;
                    }
                    if(status) {
                        psput(set, p, 1);
                    }
                }
            }

        } else {
            fprintf(Stderr, "\n*err* rppset : illegal data \n");
            rddump(rpind);
            status = False;
        }
    }
    if(!status) {
        exit(FAIL);
    }
    if(set < 0) {
        fprintf(Stderr, "\n*err* rppset : Set is not defined\n");
        rddump(rpind);
        status = False;
    }
    ptable[set]->pcode  = pcode;
    ptable[set]->kfcode = kfcode;
    
    return status;
}

/*==================================================================*/
Public  Void    rpflags(NoArg)
/*--------------------------------------------------------------------
 */
{
    printf("Name    = (Global, biconn, axconn)\n");
    printf("expand  = (%d, %d, %d)\n", expand,  sexpand,  aexpand);
    printf("expmin  = (%d, %d, %d)\n", expmin,  sexpmin,  aexpmin);
    printf("expmax  = (%d, %d, %d)\n", expmax,  sexpmax,  aexpmax);
    printf("tadpole = (%d, %d, %d)\n", tadpole, stadpole, atadpole);
    printf("extself = (%d, %d, %d)\n", extself, sextself, aextself);
    printf("onepi   = (%d, %d, %d)\n", onepi,   sonepi,   aonepi);
    printf("block   = (%d, %d, %d)\n", block,   sblock,   ablock);
    printf("selfe   = (%d, %d, %d)\n", selfe,   sselfe,   aselfe);
}

#if DBG
/*==================================================================*/
Private Void    rpdump(NoArg)
/*--------------------------------------------------------------------
   Dump the process table.
--------------------------------------------------------------------*/
{
    int p, i;

    printf("rpdump : dump process table: rproc = %d\n", rproc);

    for(p = 0; p < rproc; p++) {
        printf("%d-th process\n", p);

        printf("nextern=%d, expand=%d(%d--%d), ounepi=%d, block=%d, hep=%d",
            rpextn[p]->nextern,
            rpextn[p]->expand,
            rpextn[p]->expmin,
            rpextn[p]->expmax,
            rpextn[p]->onepi,
            rpextn[p]->hep,
            rpextn[p]->block);
        printf("tadpole=%d, extself=%d, selfe=%d, countert=%d, anyct=%d, \n",
            rpextn[p]->tadpole,
            rpextn[p]->extself,
            rpextn[p]->selfe,
            rpextn[p]->countert,
            rpextn[p]->anyct);
        printf("gm2=%d\n", rpextn[p]->gm2);
        printf("undefp=%d\n", rpextn[p]->undefp);
        if(rpextn[p]->kinem != NULL) {
            printf("kinem=%s\n", rpextn[p]->kinem);
        }
        if(rpextn[p]->pprefix != NULL) {
            printf("prefix=%s\n", rpextn[p]->pprefix);
        }

        for(i = 0; i < nsordtbl; i++) {
            printf("%s=%d ", sordtbl[i], rpextn[p]->sorder[i]);
        }
        putchar(EOL);

        for(i = 0; i < rpextn[p]->nextern; i++) {
            printf("%4d %s %s(%d,pc=%d,kf=%d)\n",
                   i, eliorf(rpextn[p]->iorf[i]),
                   pvname(rpextn[p]->particle[i]),
                   rpextn[p]->particle[i],
                   rpextn[p]->spcode[i],
                   rpextn[p]->skfcode[i]);
        }
    }
}
#endif
