/*********************************************************************
 * $id$
 *
 *  File : scalar.c
 *
 *  Driver of graph generation
 *
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include "grc.h"


/*--------------------------------------------------------------------
 * local functions
 */
Private Void    gslegs       ARG((int, node, int, node, int));
Private Void    gsconn       ARG((int, int, node));

Private bool    gscable         ARG((NoArg));
Private Void    gsnconn         ARG((node, node, int));
Private Void    gsndisc         ARG((node, node, int));
Private bool    gsleft          ARG((NoArg));
Private int     gsroot          ARG((NoArg));
Private int     gslevel         ARG((node));

/*--------------------------------------------------------------------
 * local variables
 */
Local   int     gsloop;

/* For testing performance of the program 
 *
 * iself : 0         : Class ordering test in each level.
 *       : 1         : Class ordering test only at the end.
 *       : otherwise : No class ordering test.
 */
Local   int     iself, fself;

/*==================================================================*/
Public  Void    gscalar(NoArg)
/*--------------------------------------------------------------------
 *  Entry point of scalar graph generation.
 */
{
    int root;

    lcinit();

    /* for testing performance of the program */
    iself = cptest & 0x0f;
    fself = cptest >> 4;

    /* the number of loops in the current graph */
    gsloop = 0;

    root = gsroot();   /* node number of the root is 0 */
    gsconn(root, -1, -1);
}

/*==================================================================*/
Private Void    gsconn(level, lastl, lnd)
int     level, lastl;
node    lnd;
/*--------------------------------------------------------------------
 *  Connect nodes.
 *
 *  Arguments :
 *      level : level of the source node.
 *      lastl : level of the last source node.
 *      lnd   : the last node.
 */
{
    node        ln;

#if DBG
    erenter("gsconn");
#endif

    /* find a node with free legs and a class */
    ln = lcfind(level, lnd);

    /* no more nodes in this level */
    if(ln == NDNULL) {

        /* no nodes in the last nore in the current level */
        if(lastl != level) {

        /* connected graph */
        } else if(gscable()) {

            /* update l-class : not satisfy class ordering convension */
            if(iself == 0 && !lcupdate()) {
#if DBG
                if(mon) {
                    ermon("gsconn", "not in order");
                }
#endif
            /* next level : with free legs of nodes */
            } else if(gsleft()) {

                /* the edge between level 0 and level 1 is always a bridge */
                if(level < 2 || biconn()) {
                    gsconn(level+1, level, -1);
                }

            /* end : no more free legs of nodes */
            } else {

                /* update l-class */
                if(iself == 1 && !lcupdate()) {
#if DBG
                    if(mon) {
                        ermon("gsconn", "not in order");
                    }
#endif
                } else if(biconn()) {
                    if(skipasp) {
                        if(lcftest(False)) {
                            ilsetup();
                            presult();
                        }
                    } else {
                        ilsetup();
                        prtype(1);
                        passign();
                    }
                }
                if(iself == 1)
                    lcreset();      /* reset l-class */
            }
            if(iself == 0)
                lcreset();      /* reset l-class */
#if DBG
        } else {
            if(mon) {
                ermon("gsconn", "disconnected graph");
            }
#endif
        }
    } else {            /* connect legs */
        gslegs(level, ln, ndfree(ln), -1, -1);
    }
#if DBG
    erreturn("gsconn");
#endif
}

/*==================================================================*/
Private Void    gslegs(level, sn, fl, tn, ll)
int     level;
node    sn;
int     fl;
node    tn;
int     ll;
/*--------------------------------------------------------------------
 *  Connect legs.
 *
 *  Arguments :
 *    level : level of the source node.
 *    sn    : source node.
 *    fl    : the number of free legs of the source node.
 *    tn    : the last target node.
 *    ll    : the multiplicity of the last connection.
 */
{
    int         maxfl, minfl; /* bound of multiplicity */
    int         lg, sfl;
    int         luup0, luup;
    node        nn;
/*
    int         selfl = tadpole;
*/
    int         selfl = True;  /* include self-loop */

#if DBG
    static int trgslevel=0;
    if(trace) {
        erenter("gslegs");
        printf("       %d (l:%d) %d %d %d %d\n",
               trgslevel, level, sn, fl, tn, ll);
    }
#endif

    /* select target node : including the last target node */
    for(nn = lctrgt(sn, tn); nn != NDNULL; nn = lctrgt(sn, nn)) {

        /* Avoid self-loop */
        if(!selfl && sn == nn) {
            continue;
        }

        /* Bound of multiplicity of the connection */
        if(expand) {
            minfl = 1;

            /* bound by the source node */
            if(sn == nn) {
                maxfl = ((int) fl/2);
            } else {
                maxfl = fl;
            }
            if(loop < 1) {        /* tree */
                maxfl = 1;
            }

            /* bound by the target node */
            maxfl = min(maxfl, ndfree(nn));
        } else {
            maxfl = minfl = 1;
        }

        /* save gsloop : the current number of the graph */
        luup0 = gsloop;
        for(lg = minfl; lg <= maxfl; lg++) {  /* vary multiplicity */

            /* lg-1 loops are created by the lg-connection */
            luup = gsloop + lg - 1;

            /* not first leg of the target node : 
             * this connection is a back edge.
             */
            if(ndfree(nn) < ndlegs(nn)) {
                luup++;
            }

            /* too many loops */
            if(luup > loop) {
                continue;
            }
            gsnconn(sn, nn, lg);         /* connect nodes */
            gsloop = luup;
            sfl = (sn == nn) ? fl - 2*lg : fl - lg;

            /* no free legs in the source node */
            if(sfl < 1) {

                /* the edge between level 0 and level 1 is always a bridge */
                if(level < 2 || biconn()) {
                    /* next node in the same level */
                    gsconn(level, level, sn);
                }

            /* free legs are left in the source node */
            } else {
                gslegs(level, sn, sfl, nn, lg);
            }

            gsloop = luup0;      /* restore gsloop */
            gsndisc(sn, nn, lg);  /* cut the last connection */
        }
    }
#if DBG
    if(trace)
        erreturn("gslegs");
#endif
}

/*==================================================================*/
Private Void   gsnconn(n1, n2, legs)
node n1, n2;
int  legs;
/*--------------------------------------------------------------------
 * connect two nodes with arg3 legs.
 * Configuration is saved.
 */
{
#if DBG
    if(edgetbl[n1][n2] != 0 || edgetbl[n2][n1] != 0) {
        error("gsnconn", "inconsistent connection.");
        fprintf(Stderr, "(%d,%d,%d) %d %d\n", n1, n2, legs,
                edgetbl[n1][n2], edgetbl[n2][n1]);
        prdump();
        erfail();
    }
#endif
    if(n1 == n2) {
        ndfree(n1)      -= 2*legs;
        edgetbl[n1][n1]  = 2*legs;
    } else {
        ndfree(n1)      -= legs;
        ndfree(n2)      -= legs;
        edgetbl[n1][n2]  = legs;
        edgetbl[n2][n1]  = legs;
        ndlevel(n1)      = min(ndlevel(n1), ndlevel(n2) + 1);
        ndlevel(n2)      = min(ndlevel(n2), ndlevel(n1) + 1);
    }
}

/*==================================================================*/
Private Void   gsndisc(n1, n2, legs)
node n1, n2;
int  legs;
/*--------------------------------------------------------------------
 * disconnect two nodes with arg3 legs.
 * The last configuration is restored.
 */
{
    if(n1 == n2) {
#if DBG
        if(edgetbl[n1][n1] != 2*legs) {
            error("gsndisc", "inconsistent connection.");
            fprintf(Stderr, "(%d,%d,%d)\n", n1, n2, legs);
            prdump();
            erfail();
        }
#endif
        edgetbl[n1][n1]  = 0;
        ndfree(n1)      += 2*legs;

    } else if(edgetbl[n1][n2] != legs || edgetbl[n2][n1] != legs) {
#if DBG
        error("gsndisc", "inconsistent connection.");
        fprintf(Stderr, "(%d,%d,%d)\n", n1, n2, legs);
        prdump();
        erfail();
#endif
    } else {
        edgetbl[n1][n2]  = 0;
        edgetbl[n2][n1]  = 0;
        ndfree(n1)      += legs;
        ndfree(n2)      += legs;
        ndlevel(n1)      = gslevel(n1);
        ndlevel(n2)      = gslevel(n2);
    }
}

/*==================================================================*/
Private bool   gsleft(NoArg)
/*--------------------------------------------------------------------
 * Test whether if nodes are left with free legs.
 * This routine should be changed to used table (switch with #define)
 */
{
    node n;

    for(n = 0; n < nnodes; n++) {
        if(ndfree(n) > 0) {
            return True;
        }
    }
    return False;
}

/*==================================================================*/
Private int     gslevel(n)
node n;
/*--------------------------------------------------------------------
 * calculate level of the node "n" from the information of 
 * the edge table.
 */
{
    int         lvl;
    node        m;

    lvl = NDINF;        /* level - 1 */

    if(n == 0)
        return 0;
    for(m = 0; m < nnodes; m++) {
        if(m != n && edgetbl[n][m] > 0) {
            lvl = min(lvl, ndlevel(m)+1);
        }
    }
    return lvl;
}

/*==================================================================*/
Private int   gsroot(NoArg)
/*--------------------------------------------------------------------
 * determines root and returns its level.
 */
{
    if(ndtype(0) != NDEXTERN) {
        error("gsroot", "no root is found.");
        prdump();
        erfail();
    }
    ndlevel(0) = 0;
    return 0;
}

/*==================================================================*/
Private bool    gscable(NoArg)
/*--------------------------------------------------------------------
 * test if there are free legs in the connected components
 * including the root.
 */
{
    node n;
    bool ans;

    ans = True;
    for(n = 0; n < nnodes; n++) {
        if(ndlevel(n) != NDINF) {
            if(ndfree(n) > 0)
                return True;
        } else {
            ans = False;
        }
    }
    return ans;
}
