/********************************************************************
 *  File :  window.c
 *
 *    Usage:
 *       int c;
 *       wdinit();
 *       wdopen(dest, opt);
 *       while (...)
 *       {
 *           wdwnew();
 *           --- draw graph ---
 *           wdwend();           "c" is a character read
 *       }
 *       wdcls(dest);
 *                                                 by Kaneko Toshiaki
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "vr.h"
#include "window.h"

/*--------------------------------------------------------------------
 * Local functions
 */
#define todx(x)  ((coord) (x)*wsize + ox)
#define tody(y)  ((coord) (y)*wsize + oy)
#define todr(r)  ((coord) (r)*wsize)
#define towx(x)  ((coord) (x - ox)*wdn)
#define towy(y)  ((coord) (y - oy)*wdn)
#define towr(r)  ((coord) (r)*wdn)

/*--------------------------------------------------------------------
 * Local variables
 */
static int   wdn;
static coord wsize;
static bool  drawf = False;
static coord xmin, xmak, ymak, ymin;
static coord ox, oy;
static int   wdblck;

static int dbgw = False;

/*==================================================================*/
Public  Void wdinit(NoArg)
/*--------------------------------------------------------------------
 *    initialize devices
 */
{
    if(dbgw)
        printf("wdinit\n");

    vrinit();
}

/*==================================================================*/
Public  Void wdopen(dest, opt)
int     dest;
int    *opt;
/*--------------------------------------------------------------------
 *    open window
 */
{
    if(dbgw)
        printf("wdopen(%d, %d)\n", dest, *opt);

    vropen(dest, opt);

    /* boundary color */
    wdblck = (vrback(VR_BLACK) == VR_BLACK) ? VR_WHITE : VR_BLACK;
}

/*==================================================================*/
Public  Void wdcls(dest)
int     dest;
/*--------------------------------------------------------------------
*/
{
    if(dbgw)
        printf("wdclose(%d): drawf = %d\n", dest, drawf);

    if(drawf)
        vrcls(dest);

    drawf = False;
}

/*==================================================================*/
Public  int wdsetn(n)
int     n;
/*--------------------------------------------------------------------
 *    set n x n windows
 */
{
    if(dbgw)
        printf("wdsetn\n");

    wdn   = (n > 0) ? n : vrgnum();
    wsize = ((coord) 1.0)/((coord) wdn);
    xmak  = 1.0 - 0.5*wsize;
    xmin  = 0.0;
    ymak  = 1.0 - wsize;
    ymin  = - 0.5*wsize;
    ox    = xmin;
    oy    = ymak;
    vrwvw((coord) (0.02*wsize));

    return wdn;
}

/*==================================================================*/
Public  bool wdclr(bool clr)
/*--------------------------------------------------------------------
*/
{
    bool st;

    if(dbgw)
        printf("wdclr\n");

    st = vrclr(clr);
    ox = xmin;
    oy = ymak;
    drawf = False;

    return st;
}

/*==================================================================*/
Public  Void    wddisp(NoArg)
/*--------------------------------------------------------------------
*/
{
    if(dbgw)
        printf("wddisp\n");

    vrdisp();
}

/*==================================================================*/
Public  Void wdnew(NoArg)
/*--------------------------------------------------------------------
*/
{
    coord px, py;
#ifdef TURBOC
    static int ngr = 0;
    static char ngrs[17];
    int n;
    coord  csz;
#endif

    if(dbgw)
        printf("wdwnew\n");

    vrnew();

/*
    if(oy == ymak && ox == xmin)
        vrclr();
*/

    px = ox + wsize;
    py = oy + wsize;
    vrline(ox, oy, px, oy, wdblck);
    vrline(px, oy, px, py, wdblck);
    vrline(px, py, ox, py, wdblck);
    vrline(ox, py, ox, oy, wdblck);
    drawf = True;

#ifdef TURBOC
    ngr++;
    if(ox < 0.5*wsize) {
        itoa(ngr, ngrs, 10);
        csz = 0.05;
        wdchrh(csz, &csz);
        wdchrw(ngrs, &csz);
        wdtext(- csz - 0.05, 0.05, ngrs, wdblck);
    }
#endif
}

/*==================================================================*/
Public  int     wdend(NoArg)
/*--------------------------------------------------------------------
*/
{
    int c = 0;

    if(dbgw)
        printf("wdwend\n");

    if (drawf) {
        ox += wsize;
        if (ox > xmak) {
            ox = xmin;
            oy -= wsize;
            if (oy < ymin) {
                vrend();
                vrdisp();
                oy = ymak;
            }
        }
    }

    return c;
}

/*==================================================================*/
Public  Void wdline(x0, y0, x1, y1, color)
coord x0, y0;
coord x1, y1;
int color;
/*--------------------------------------------------------------------
*/
{
    if(dbgw)
        printf("wdline\n");

    vrline(todx(x0), tody(y0), todx(x1), tody(y1), color);
    drawf = True;
}

/*==================================================================*/
Public  Void wdcircle(x, y, r, color)
coord x, y;
coord r;
int color;
/*--------------------------------------------------------------------
   draw circle. center = ("x", "y"); radius = "r"; color = "color"
*/
{
    if(dbgw)
        printf("wdcircle\n");

    vrcrcl(todx(x), tody(y), todr(r), color);
    drawf = True;
}

/*==================================================================*/
Public  Void wdarc(x0, y0, x1, y1, x2, y2, color)
coord x0, y0;
coord x1, y1;
coord x2, y2;
int color;
/*--------------------------------------------------------------------
   draw arc. It goes (x0, y0) - (x1, y1) - (x2, y2)
*/
{
    if(dbgw)
        printf("wdarc(%f, %f, %f, %f, %f, %f, %d)\n",
               x0, y0, x1, y1, x2, y2, color);

    vrarc(todx(x0), tody(y0), todx(x1), tody(y1),
          todx(x2), tody(y2), color);
    drawf = True;
}

/*==================================================================*/
Public  Void  wdtext(x, y, str, color)
coord x, y;
string str;
int color;
/*--------------------------------------------------------------------
   draw text string.
   (x, y) points lower left position of the string.
*/
{
    if(dbgw)
        printf("wdtext\n");

    vrtext(todx(x), tody(y), str, color);
}

/*==================================================================*/
Public  int wdcolor(NoArg)
/*--------------------------------------------------------------------
*/
{
    if(dbgw)
        printf("wdcolor\n");

    return vrcolr();
}

/*==================================================================*/
Public  int wdblack(NoArg)
/*--------------------------------------------------------------------
*/
{
    if(dbgw)
        printf("wdblack\n");

    return wdblck;
}

/*==================================================================*/
Public  Void  wdchrh(r, ans)
coord r;
coord *ans;
/*--------------------------------------------------------------------
   set character hight. "r" is desierable size. It returns size.
*/
{
    coord b, c;

    if(dbgw)
        printf("wdchrh\n");

    c = todr(r);
    vrchrh(c, &b);
    *ans = towr(b);
}

/*==================================================================*/
Public  Void  wdchrw(str, ans)
string str;
coord  *ans;
/*--------------------------------------------------------------------
   get length of string 
*/
{
    coord a;

    if(dbgw)
        printf("wdchrw\n");

    vrchrw(str, &a);
    *ans = towr(a);
}

/*==================================================================*/
Public  int wdgnum(NoArg)
/*--------------------------------------------------------------------
*/
{
    if(dbgw)
        printf("wdgnum\n");

    return vrgnum();
}
